/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                            Find/Replace dialog                                *
*********************************************************************************
*   $Id: FXFindDialog.h,v 1.2 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXFindDialog.h,v $
* Revision 1.2  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.1.1.1  2000/10/29 14:20:25  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXFINDDIALOG_H
#define FXFINDDIALOG_H


class FXPacker;
class FXMatrix;
class FXHorizontalFrame;
class FXTextField;
class FXCheckButton;
class FXRadioButton;
class FXWindow;
class FXTextSrc;


/// Struct passed as data with FXPLText::ID_FINDSTR/ID_REPLSTR etc. messages.
struct FXFindData {
  FXchar * find;	/// String to find
  FXint flen;		/// Length of search string
  FXchar * replace;	/// Replacement string
  FXint rlen;		/// Length of replacement string
  FXbool case_sens;	/// Whether case-sensitive search
  FXbool fwd;		/// Whether to search forward
  FXbool restart;	/// Whether to restart search from top (fwd) or bottom (!fwd)
  FXbool whole_word;	/// Whether to find whole words only (word delimiters defined by text src).
  FXbool multi_line;	/// Whether multi-line match is allowed (this is implied if find string
  			/// not a regex, and has \n chars.
  enum{
    FD_LITERAL,		/// Literal match
    FD_LITTERAL=FD_LITERAL, /// Alternative spelling for us illit(t)erates -- one 'T' is correct.
    FD_C_ESC,		/// C-escapes e.g. \n for newline, \t for tab.
    FD_REGEX		/// Regular expression - POSIX recommended, but determined by text src.
    } search_type, replace_type; /// Currently, both types are set the same.
  
  int limit;		/// Maximum number of finds/replaces allowed in one hit.
  FXbool more;		/// Set by find/replace engine if limit halted search prematurely
  			/// i.e. there is at least one more match available.
  FXbool move;		/// Move cursor to last of 'Find All' or 'Replace All', else these
  			/// operations don't move the cursor.
  FXbool extend;	/// For operations involving secondary selections, if this is true
  			/// then the current sec sels are not removed before the first
  			/// successful operation.  Otherwise, the first success causes previous
  			/// sec sels to be removed.
  };


/// Find/replace dialogue for FXPLText widgets.
class FXFindReplaceDialog : public FXDialogBox {
  FXDECLARE(FXFindReplaceDialog)
protected:
  FXTextSrc * textsrc;
  FXPacker* contents;
  FXMatrix* fields;
  FXHorizontalFrame* buttons2;
  FXMatrix* buttons;
  FXTextField* find;
  FXTextField* replace;
  FXTextField* limit;
  FXCheckButton* case_sens;
  FXCheckButton* whole_word;
  FXCheckButton* multi_line; 
  FXCheckButton* move_cursor; 
  FXRadioButton* literal;
  FXRadioButton* c_esc;
  FXRadioButton* regex;
  FXFindData fd;
private:
  FXFindReplaceDialog(){}
  void common_init(FXTextSrc * ts);
public:
  long onCmdAction(FXObject*,FXSelector,void*);
  enum{
    ID_FIND=FXDialogBox::ID_LAST,/// Find next
    ID_FIND_PREV,		/// Find previous
    ID_FIND_FIRST,		/// Find first occurrence
    ID_FIND_LAST,		/// Find last
    ID_FIND_ALL,		/// Secondary select all occurrences
    
    ID_REPLACE,			/// Replace current selection if owner has it
    ID_REPLFIND,		/// Replace then find
    ID_FINDREPL,		/// Find then replace
    ID_REPLALL_SEL,		/// Replace all secondary selections
    ID_REPLALL,			/// Replace all (no selection needed)
    
    ID_LAST
    };
  
  /// Construct dialogue box.  Non-modal.  ts is queried to determine its capabilities,
  /// since it actually implements the find/replace functionality.  If ts is null, then
  /// a default capability is assumed.  Floats over owner.
  FXFindReplaceDialog(FXWindow * owner, FXTextSrc * ts);
  
  /// Construct as above, except free-floating.  Application will need to explicitly
  /// setTarget() to the required FXPLText (or delegator) object.
  FXFindReplaceDialog(FXApp * a, FXTextSrc * ts);
  
  /// Set search string.
  void setFindText(FXuchar *, FXint len);
  
  /// Set replace string.
  void setReplaceText(FXuchar *, FXint len); 
  
  /// Return information about current find parameters.  Returns null if specification error.
  FXFindData * getFindData();
  
  /// Return information about current replace parameters.  Returns null if specification error.
  FXFindData * getReplaceData(); 
  
  /// Get target text source
  FXTextSrc * getTextSrc() const { return textsrc; }
  
  /// Set target text source
  void setTextSrc(FXTextSrc * ts) { textsrc = ts; }
  
  /// Destructor
  virtual ~FXFindReplaceDialog();
  };



#endif