/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                              Text Widget                                      *
*********************************************************************************
*   $Id: FXPLText.cpp,v 1.4 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Jeroen van der Zijp
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXPLText.cpp,v $
* Revision 1.4  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.3  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.2  2000/10/30 12:43:24  tekno
* fixed minor problems noted in initial release
*
* Revision 1.1.1.1  2000/10/29 14:17:36  tekno
* Initial release
*
*
*
********************************************************************************/
#ifdef __INSURE__
#define __attribute__(x)
#endif

#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXCharset.h"
#include "FXDict.h"
#include "FXSettings.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXSize.h"
#include "FXPoint.h"
#include "FXRectangle.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXDCWindow.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXScrollbar.h"
#include "FXPLText.h"
#include "FXShell.h"
#include "FXTextSrc.h"
#include "FXPageFormat.h"
#include "FXTextAnchor.h" 
#include "FXTopWindow.h" 
#include "FXDialogBox.h" 
#include "FXFindDialog.h"
#include "FXMessageBox.h"
#ifdef POSIX_REGEX
#include <regex.h>
#else
#include "fxregex.h"
#endif

/*
  Notes:
  - Suggested additional bindings:
    Ctl-F   Find dialogue popup
    Ctl-G   Find next matching selection (Shift-Ctl-G backward)
    Ctl-E   Goto selected line number
    Ctl-I   Indent 1 character right (Shift-Ctl-I 1 char left)
    Ctl-Z   undo (Shift-Ctl-Z redo)
    Ctl-Y   move cursor to start location of next undo (Shift: redo)
*/

#define AUTOSCROLL_FUDGE  10       // Proximity to wall at which we start autoscrolling
#define SCROLLER_MASK     0x000F8000



#define MOUSE_NONE    0                // None in effect
#define MOUSE_CHARS   1                // Character drag select
#define MOUSE_WORDS   2                // Word drag select
#define MOUSE_LINES   3                // Line drag select
#define MOUSE_SCROLL  4                // Scrolling mode
#define MOUSE_DRAG    5                // DND-drag
#define MOUSE_PASTE   6                // Text paste
#define MOUSE_TRYDRAG 7                // Tentative DND-drag

//FIXME: add readonly and prepend flags
// The options that can be changed by calling setTextStyle().
#define TEXT_MASK   (PLTEXT_SEL_BY_ROW|PLTEXT_OVERSTRIKE|PLTEXT_AUTO_INDENT| \
  PLTEXT_SHOW_AUX|PLTEXT_SHOW_LINE_NUMBERS|PLTEXT_SHOW_TABS)
 
// The options (from above) which require recalc/update
#define REQ_RECALC (PLTEXT_SHOW_AUX|PLTEXT_SHOW_LINE_NUMBERS|PLTEXT_SHOW_TABS)

// Scroll acceleration near edge
static const FXint acceleration[AUTOSCROLL_FUDGE+1]={1,1,1,2,3,4,6,8,12,16,24};


/*******************************************************************************/

  
// Map
FXDEFMAP(FXPLText) FXPLTextMap[]={
  FXMAPFUNC(SEL_PAINT,0,FXPLText::onPaint),
  FXMAPFUNC(SEL_MOTION,0,FXPLText::onMotion),
  FXMAPFUNC(SEL_DRAGGED,0,FXPLText::onDragged),
  FXMAPFUNC(SEL_TIMEOUT,FXWindow::ID_AUTOSCROLL,FXPLText::onAutoScroll),
  FXMAPFUNC(SEL_FOCUSIN,0,FXPLText::onFocusIn),
  FXMAPFUNC(SEL_FOCUSOUT,0,FXPLText::onFocusOut),
  FXMAPFUNC(SEL_BEGINDRAG,0,FXPLText::onBeginDrag),
  FXMAPFUNC(SEL_ENDDRAG,0,FXPLText::onEndDrag),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,0,FXPLText::onLeftBtnPress),
  FXMAPFUNC(SEL_LEFTBUTTONRELEASE,0,FXPLText::onLeftBtnRelease),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,0,FXPLText::onMiddleBtnPress),
  FXMAPFUNC(SEL_MIDDLEBUTTONRELEASE,0,FXPLText::onMiddleBtnRelease),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS,0,FXPLText::onRightBtnPress),
  FXMAPFUNC(SEL_RIGHTBUTTONRELEASE,0,FXPLText::onRightBtnRelease),
  FXMAPFUNC(SEL_UNGRABBED,0,FXPLText::onUngrabbed),
  FXMAPFUNC(SEL_CHANGED,0,FXPLText::onChanged),
  FXMAPFUNC(SEL_CHANGED,FXPLText::ID_TRACK_TOP,FXPLText::onTrackTop),
  FXMAPFUNC(SEL_DND_ENTER,0,FXPLText::onDNDEnter),
  FXMAPFUNC(SEL_DND_LEAVE,0,FXPLText::onDNDLeave),
  FXMAPFUNC(SEL_DND_DROP,0,FXPLText::onDNDDrop),
  FXMAPFUNC(SEL_DND_MOTION,0,FXPLText::onDNDMotion),
  FXMAPFUNC(SEL_DND_REQUEST,0,FXPLText::onDNDRequest),
  FXMAPFUNC(SEL_SELECTION_LOST,0,FXPLText::onSelectionLost),
  FXMAPFUNC(SEL_SELECTION_GAINED,0,FXPLText::onSelectionGained),
  FXMAPFUNC(SEL_SELECTION_REQUEST,0,FXPLText::onSelectionRequest),
  FXMAPFUNC(SEL_CLIPBOARD_LOST,0,FXPLText::onClipboardLost),
  FXMAPFUNC(SEL_CLIPBOARD_GAINED,0,FXPLText::onClipboardGained),
  FXMAPFUNC(SEL_CLIPBOARD_REQUEST,0,FXPLText::onClipboardRequest),
  FXMAPFUNC(SEL_KEYPRESS,0,FXPLText::onKeyPress),
  FXMAPFUNC(SEL_KEYRELEASE,0,FXPLText::onKeyRelease),
  FXMAPFUNC(SEL_SELECTED,0,FXPLText::onSelected),
  FXMAPFUNC(SEL_DESELECTED,0,FXPLText::onDeselected),
  FXMAPFUNC(SEL_INSERTED,0,FXPLText::onInserted),
  FXMAPFUNC(SEL_DELETED,0,FXPLText::onDeleted),
  FXMAPFUNC(SEL_UPDATE,FXWindow::ID_QUERY_HELP,FXPLText::onQueryHelp),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TOGGLE_ESCAPE,FXPLText::onUpdToggleEscape),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TOGGLE_EDITABLE,FXPLText::onUpdToggleEditable),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TOGGLE_OVERSTRIKE,FXPLText::onUpdToggleOverstrike),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TOGGLE_FLASH,FXPLText::onUpdToggleFlash),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_CURSOR_ROW,FXPLText::onUpdCursorRow),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_CURSOR_LINE,FXPLText::onUpdCursorLine),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_CURSOR_COLUMN,FXPLText::onUpdCursorColumn),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_CURSOR_OFFSET,FXPLText::onUpdCursorOffset),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_CUT_SEL,FXPLText::onUpdHaveSelection),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_COPY_SEL,FXPLText::onUpdHaveSelection),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_PASTE_SEL,FXPLText::onUpdYes),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_DELETE_SEL,FXPLText::onUpdHaveSelection), 
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_BACK_COLOR,FXPLText::onUpdBackColor),
  FXMAPFUNCS(SEL_UPDATE,FXPLText::ID_FIRST_COLOR,FXPLText::ID_LAST_COLOR,
    FXPLText::onUpdColor),
  FXMAPFUNCS(SEL_COMMAND,FXPLText::ID_FIRST_COLOR,FXPLText::ID_LAST_COLOR,
    FXPLText::onCmdColor),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TAB_COLUMNS,FXPLText::onUpdTabColumns),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_WRAP_COLUMNS,FXPLText::onUpdWrapColumns),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_TOP,FXPLText::onCmdCursorTop),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_BOTTOM,FXPLText::onCmdCursorBottom),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_HOME,FXPLText::onCmdCursorHome),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_END,FXPLText::onCmdCursorEnd),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_RIGHT,FXPLText::onCmdCursorRight),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_LEFT,FXPLText::onCmdCursorLeft),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_UP,FXPLText::onCmdCursorUp),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_DOWN,FXPLText::onCmdCursorDown),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_WORD_LEFT,FXPLText::onCmdCursorWordLeft),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_WORD_RIGHT,FXPLText::onCmdCursorWordRight),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_PAGEDOWN,FXPLText::onCmdCursorPageDown),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_PAGEUP,FXPLText::onCmdCursorPageUp),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_SCRNTOP,FXPLText::onCmdCursorScreenTop),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_SCRNBTM,FXPLText::onCmdCursorScreenBottom),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_SCRNCTR,FXPLText::onCmdCursorScreenCenter),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_PAR_HOME,FXPLText::onCmdCursorParHome),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_PAR_END,FXPLText::onCmdCursorParEnd),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_MARK,FXPLText::onCmdMark),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_EXTEND,FXPLText::onCmdExtend),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_INSERT_CHAR,FXPLText::onCmdInsertChar),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_INSERT_STRING,FXPLText::onCmdInsertString),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_INSERT_NEWLINE,FXPLText::onCmdInsertNewline),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CUT_SEL,FXPLText::onCmdCutSel),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_COPY_SEL,FXPLText::onCmdCopySel),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_PASTE_SEL,FXPLText::onCmdPasteSel),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_DELETE_SEL,FXPLText::onCmdDeleteSel),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_SELECT_CHAR,FXPLText::onCmdSelectChar),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_SELECT_WORD,FXPLText::onCmdSelectWord),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_SELECT_LINE,FXPLText::onCmdSelectLine),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_SELECT_ALL,FXPLText::onCmdSelectAll),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_DESELECT_ALL,FXPLText::onCmdDeselectAll),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_BACKSPACE,FXPLText::onCmdBackspace),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_BACKSPACE_WORD,FXPLText::onCmdBackspaceWord),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_BACKSPACE_BOL,FXPLText::onCmdBackspaceBol),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_DELETE,FXPLText::onCmdDelete),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_DELETE_WORD,FXPLText::onCmdDeleteWord),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_DELETE_EOL,FXPLText::onCmdDeleteEol),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_DELETE_LINE,FXPLText::onCmdDeleteLine),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_ESCAPE,FXPLText::onCmdToggleEscape),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_EDITABLE,FXPLText::onCmdToggleEditable),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_OVERSTRIKE,FXPLText::onCmdToggleOverstrike),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_FLASH,FXPLText::onCmdToggleFlash),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_ROW,FXPLText::onCmdCursorRow),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_LINE,FXPLText::onCmdCursorLine),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_COLUMN,FXPLText::onCmdCursorColumn),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CURSOR_OFFSET,FXPLText::onCmdCursorOffset),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_BACK_COLOR,FXPLText::onCmdBackColor),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_SETSTRINGVALUE,FXPLText::onCmdSetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_GETSTRINGVALUE,FXPLText::onCmdGetStringValue),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TAB_COLUMNS,FXPLText::onCmdTabColumns),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_WRAP_COLUMNS,FXPLText::onCmdWrapColumns),
  
  FXMAPFUNC(SEL_COMMAND,FXWindow::ID_VSCROLLED,FXPLText::onVScrollerChanged),
  FXMAPFUNC(SEL_CHANGED,FXWindow::ID_VSCROLLED,FXPLText::onVScrollerDragged),
  FXMAPFUNC(SEL_TIMEOUT,FXWindow::ID_AUTOSCROLL,FXPLText::onAutoScroll),
  
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_FIND,		FXPLText::onCmdFind),
  FXMAPFUNCS(SEL_COMMAND,FXPLText::ID_FINDNEXT,
                         FXPLText::ID_FINDPREV,		FXPLText::onCmdFindNext),
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_FINDSTR,		FXPLText::onCmdFindStr),
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_FINDALL,		FXPLText::onCmdFindAll),
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_REPLSTR,		FXPLText::onCmdReplStr),
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_REPLALLSEL,	FXPLText::onCmdReplAllSel),
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_REPLALL,		FXPLText::onCmdReplAll),
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_GOTO_SEL_LINE,	FXPLText::onCmdGotoSelLine),
  FXMAPFUNC(SEL_COMMAND, FXPLText::ID_INDENT,		FXPLText::onCmdIndent),
  FXMAPFUNCS(SEL_COMMAND,FXPLText::ID_ROLLBACK,
                         FXPLText::ID_ROLLFORWARD,	FXPLText::onCmdRoll),
  FXMAPFUNCS(SEL_COMMAND,FXPLText::ID_CURSOR_TO_ROLLBACK,
                         FXPLText::ID_CURSOR_TO_ROLLFORWARD,	FXPLText::onCmdCursorToRoll),

  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_SHOWTABS,FXPLText::onCmdToggleShowtabs),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_AUTOINDENT,FXPLText::onCmdToggleAutoindent),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_SHOWAUX,FXPLText::onCmdToggleShowaux),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TOGGLE_SHOWTABS,FXPLText::onUpdToggleShowtabs),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TOGGLE_AUTOINDENT,FXPLText::onUpdToggleAutoindent),
  FXMAPFUNC(SEL_UPDATE,FXPLText::ID_TOGGLE_SHOWAUX,FXPLText::onUpdToggleShowaux),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_TOGGLE_SECSEL,FXPLText::onCmdToggleSecSel),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CLEAR_SECSEL,FXPLText::onCmdClearSecSel),
  FXMAPFUNC(SEL_COMMAND,FXPLText::ID_CLEAR_ALL_SECSEL,FXPLText::onCmdClearAllSecSel),
  };




// Object implementation
FXIMPLEMENT(FXPLText,FXComposite,FXPLTextMap,ARRAYNUMBER(FXPLTextMap))

   

/*******************************************************************************/


// Helper
static inline FXint fxabs(FXint a){ return a<0?-a:a; }


// For deserialization
FXPLText::FXPLText(){
  ok = TRUE;
  flags|=FLAG_ENABLED|FLAG_SHOWN;
  horizontal=NULL;
  vertical=NULL;
  corner=NULL;
  scrolltimer=NULL;
  viewport_w=viewport_w_pix=1;
  viewport_h=viewport_h_pix=1;
  content_w=1;
  content_h=1;
  pos_x=0;
  pos_y=0;
  nvisrows=0;
  toprow=0;
  charWidth=0;
  charHeight=0;
  dragmode=MOUSE_NONE;
  grabx=0;
  graby=0;
  anydirty = FALSE;
  rds = NULL;
  frd = NULL;
  frd_private = TRUE;
  }

void FXPLText::setDefaults() {
  FXint i;
  
  ok = TRUE;
  flags|=FLAG_ENABLED|FLAG_SHOWN;
  options &= ~PLTEXT_HSCROLLER_ALWAYS;
  options |= PLTEXT_HSCROLLER_NEVER;	// Ensure no horizontal scrollbar
  horizontal=NULL;
  vertical=new FXScrollbar(this,this,FXWindow::ID_VSCROLLED,SCROLLBAR_VERTICAL);
  corner=NULL;
  backColor=getApp()->getBackColor();
  scrolltimer=NULL;
  viewport_w=viewport_w_pix=1;
  viewport_h=viewport_h_pix=1;
  content_w=1;
  content_h=1;
  pos_x=0;
  pos_y=0;
  want_visible = TRUE;
  nvisrows=0;
  toprow=0;
  charWidth=0;
  charHeight=0;
  dragmode=MOUSE_NONE;
  grabx=0;
  graby=0;
  anydirty = FALSE;
  rds = NULL;
  frd = NULL;
  frd_private = TRUE;
  prefcol = 0;
  clipbuffer = NULL; 
  cliplength = 0;
  cursorActive = FALSE;
  
  margintop = marginleft = 3;
  marginbottom = marginright = 0;
  
  // Set up a reasonable default font.
  FXFontDesc fd;
  strcpy(fd.face, "fixed");	 // Don't care 
  fd.size = getApp()->getNormalFont()->getSize();
  fd.weight = FONTWEIGHT_DONTCARE;
  fd.slant = FONTSLANT_REGULAR;
  fd.encoding = FONTENCODING_DEFAULT;
  fd.setwidth = FONTSETWIDTH_DONTCARE;
  fd.flags = FONTPITCH_FIXED;
  FXTRACE((100,"new font\n"));
  font = new FXFont(getApp(), fd);
  auxFont = new FXFont(getApp(), fd);
  FXASSERT(font);
  FXASSERT(auxFont);
  greek_x = 2;
  greek_y = 3;
  track = NULL;
  
  // See PLCOLOR_* enum in FXPLText.h for correct ordering.  Let's try to strike
  // a balance between being garish and being too subtle.
#define pl_black FXRGB(0,0,0)
#define pl_white FXRGB(240,240,240)
#define pl_lt_gray FXRGB(192,192,192)
#define pl_mid_gray FXRGB(128,128,128)
#define pl_dk_gray FXRGB(64,64,64)
#define pl_lt_blue FXRGB(192,192,255)
#define pl_yellow FXRGB(220,220,0)
#define pl_orange FXRGB(255,120,0)
#define pl_lt_green FXRGB(192,255,192)
#define pl_pink FXRGB(255,192,192)
#define pl_blue FXRGB(0,0,192)
#define pl_tan FXRGB(212,140,0)
  static const FXColor pltext_scheme[] = {
    pl_dk_gray,		// Aux BG
    pl_black,		// Aux sep
    pl_lt_blue,		// Line nums
    pl_yellow,		// Aux highlight
    pl_tan,		// Greek window
    pl_black,		// resv2
    pl_black,		// resv3
    pl_black,		// resv4
    pl_black,		// Text fore
    pl_black,		// Selection back
    pl_white,		// Selection fore
    pl_orange,		// Error back
    pl_black,		// Error fore
    pl_lt_gray,		// Secondary back
    pl_black,		// Secondary fore
    pl_lt_green,	// Retentive back
    pl_black,		// Retentive fore
    pl_pink,		// Marker back
    pl_black,		// Marker fore
    pl_white,		// Input back
    pl_blue,		// Input fore
    pl_white,		// Output back
    pl_black,		// Output fore
    pl_white,		// Special back
    pl_mid_gray,	// Special fore
    pl_tan		// Cursor
    };
  
  // Default colour scheme
  for (i = 0; i < PLCOLOR_LAST; i++)
    colors[i] = pltext_scheme[i];
  
  bg_mask = FXPFATTR_SELECTED | FXPFATTR_ERRORS | FXPFATTR_SECONDARY | FXPFATTR_RETENTIVE |
            FXPFATTR_MARKER;
  fg_mask = FXPFATTR_SELECTED | FXPFATTR_INPUT | FXPFATTR_OUTPUT | FXPFATTR_SPECIAL;
  
  // Cursor styles
  cur_cstyle = 0;
  for (i = 0; i < FXPLTEXT_MAX_CURSOR_STYLES; i++) {
    // Set all to non-flashing caret
    cstyles[i].onstyle = 1;
    cstyles[i].offstyle = 4;
    cstyles[i].ontime = 0;
    cstyles[i].offtime = 0;
    cstyles[i].flash = FALSE;
    }
  // Style 1 is block for overstrike mode.
  cstyles[1].onstyle = 2;
  cstyles[1].offstyle = 3;
  
  // Set aux area on if specified
  set_auxwidth = 40;
  if (options & PLTEXT_SHOW_AUX)
    auxwidth = set_auxwidth;
  else
    auxwidth = 0;
  sepwidth = 1;
  
  rolines = 0;
  }

// Text widget
FXPLText::FXPLText(FXComposite *p,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXComposite(p,opts,x,y,w,h) {
  setDefaults();
  target=tgt;
  message=sel;
  
  // Set up default page format and text source
  pager = new FXWrapPageFormat(this);
  if (!pager || !pager->isOk())
    ok = FALSE;
  else {
    textsrc = pager->getTextSrc();
    cursor = new FXTextCursor(textsrc, this, 0, NULL, 0, FXTextCursor::TEXTCURSOR_IBEAM);
    if (cursor) {
      cursor->setColor(colors[PLCOLOR_CURSOR]);
      if ((options & PLTEXT_EDITMASK) == PLTEXT_APPEND ||
          (options & PLTEXT_EDITMASK) == PLTEXT_PREPEND ) {
        rolines = textsrc->lines() - 1;
        }
      }
    else
      ok = FALSE;
    }
  }

// Text widget with shared text source
FXPLText::FXPLText(FXComposite *p,FXTextSrc* ts,FXObject* tgt,FXSelector sel,FXuint opts,
                   FXint x,FXint y,FXint w,FXint h):
  FXComposite(p,opts,x,y,w,h) {
  setDefaults();
  target=tgt;
  message=sel;
    
  // Set up non-default page format and text source
  FXASSERT(ts);
  pager = new FXWrapPageFormat(this, ts);
  if (!pager || !pager->isOk())
    ok = FALSE;
  else {
    textsrc = pager->getTextSrc();
    cursor = new FXTextCursor(ts, this, 0, NULL, 0, FXTextCursor::TEXTCURSOR_CARET);
    if (cursor) {
      cursor->setColor(colors[PLCOLOR_CURSOR]);
      if ((options & PLTEXT_EDITMASK) == PLTEXT_APPEND ||
          (options & PLTEXT_EDITMASK) == PLTEXT_PREPEND ) {
        rolines = textsrc->lines() - 1;
        }
      }
    else
      ok = FALSE;
    }
  }


void FXPLText::makeLike(FXPLText * model) {
  setTextColor(model->getTextColor());
  setBackColor(model->getBackColor());
  setSelBackColor(model->getSelBackColor());
  setSelTextColor(model->getSelTextColor());
  // Options
  FXuint mask = PLTEXT_AUTO_INDENT|PLTEXT_SHOW_AUX|PLTEXT_SHOW_TABS|PLTEXT_SHOW_LINE_NUMBERS;
  FXuint s = getTextStyle();
  s &= ~mask;
  s |= model->getTextStyle() & mask;
  setTextStyle(s);
  getPageFormat()->setTabColumns(model->getPageFormat()->getTabColumns());
  setFont(model->getFont(), model->getAuxFont());
  }


// Create window
void FXPLText::create(){
  FXComposite::create();
  FXASSERT(font);
  FXTRACE((100,"create font\n"));
  font->create();
  FXASSERT(font->isFontMono());
  auxFont->create();
  if(!deleteType){deleteType=getApp()->registerDragType(deleteTypeName);}
  if(!textType){textType=getApp()->registerDragType(textTypeName);}
  setCharMetrics();
  dropEnable();
  }


// Detach window
void FXPLText::detach(){
  FXComposite::detach();
  font->detach();
  auxFont->detach();
  deleteType=0;
  textType=0;
  }


/* FXScrollArea-compatible functions... */
/* ------------------------------------ */
// This should really add the scroll bar size only when required; however, 
// that depends on the actual size.  We are potentially being called at
// a moment when this is not known yet, so we return a size which reflects
// the situation when the scrollbars have been placed; this way, we should
// at least have enough space to fully see the contents, and a bit extra
// when the scrollbars turn out to have been unnecessary.

// Get default width
FXint FXPLText::getDefaultWidth(){
  FXint w=charwidth2pix(getContentWidth());
  if(!(options&PLTEXT_VSCROLLER_NEVER)) w+=vertical->getDefaultWidth();
  return FXMAX(w,1); 
  }


// Get default height
FXint FXPLText::getDefaultHeight(){
  FXint h=0;
  if((options&PLTEXT_VSCROLLER_NEVER)&&(options&PLTEXT_VSCROLLER_ALWAYS)) 
    h=charheight2pix(getContentHeight());
  if(!(options&PLTEXT_VSCROLLER_NEVER)) h+=vertical->getDefaultHeight();
  return FXMAX(h,1); 
  }


// Move content (units of characters)
void FXPLText::moveContents(FXint x,FXint y){
  pos_x=x; // Dummy
  pos_y=y;
  pager->setTopRow(-pos_y);
  }



// Changed
long FXPLText::onVScrollerChanged(FXObject*,FXSelector,void* ptr){
  FXint new_y=-(FXint)(long)ptr;
  want_visible = FALSE;
  if(new_y!=pos_y){
    moveContents(0,new_y);
    }
  flags&=~FLAG_TIP;
  return 1;
  }



// Dragged
long FXPLText::onVScrollerDragged(FXObject*,FXSelector,void* ptr){
  want_visible = FALSE;
  if(!(options&PLTEXT_SCROLLERS_DONT_TRACK)){
    FXint new_y=-(FXint)(long)ptr;
    if(new_y!=pos_y){
      moveContents(0,new_y);
      }
    }
  flags&=~FLAG_TIP;
  return 1;
  }

// Start automatic scrolling 
FXbool FXPLText::startAutoScroll(FXint x,FXint y,FXbool onlywheninside){
  FXbool autoscrolling=FALSE;
  flags&=~FLAG_SCROLLINSIDE;
  if(onlywheninside) flags|=FLAG_SCROLLINSIDE;

  if(vertical->getPage()<vertical->getRange()){
    if((y<AUTOSCROLL_FUDGE) && (0<vertical->getPosition())) autoscrolling=TRUE;
    else if((viewport_h_pix-AUTOSCROLL_FUDGE<=y) && (vertical->getPosition()<vertical->getRange()-vertical->getPage())) autoscrolling=TRUE;
    }
  if(onlywheninside && (x<0 || y<0 || viewport_w_pix<=x || viewport_h_pix<=y)) autoscrolling=FALSE;
  if(autoscrolling){
    if(!scrolltimer){ scrolltimer=getApp()->addTimeout(getApp()->getScrollSpeed(),this,FXWindow::ID_AUTOSCROLL); }
    }
  else{
    if(scrolltimer){ scrolltimer=getApp()->removeTimeout(scrolltimer); }
    }
  return autoscrolling;
  }


// Stop automatic scrolling
void FXPLText::stopAutoScroll(){
  if(scrolltimer){scrolltimer=getApp()->removeTimeout(scrolltimer);}
  flags&=~FLAG_SCROLLINSIDE;
  }


// Set scroll style
void FXPLText::setScrollStyle(FXuint style){
  FXuint opts=(options&~SCROLLER_MASK) | (style&SCROLLER_MASK);
  if(options&SCROLLER_MASK!=opts){	// The & was omitted from original code
    options=opts;
    recalc();
    }
  }


// Get scroll style
FXuint FXPLText::getScrollStyle() const {
  return (options&SCROLLER_MASK); 
  }


// True if horizontally scrollable enabled
FXbool FXPLText::isHorizontalScrollable() const {
  return FALSE;
  }


// True if vertically scrollable enabled
FXbool FXPLText::isVerticalScrollable() const {
  return !((options&PLTEXT_VSCROLLER_NEVER) && (options&PLTEXT_VSCROLLER_ALWAYS));
  }


  
// Default viewport width
FXint FXPLText::getViewportWidth(){
  return width;
  }


// Default viewport height
FXint FXPLText::getViewportHeight(){
  return height;
  }


// Determine content width of scroll area (char units)
FXint FXPLText::getContentWidth(){ 
  if (flags&FLAG_RECALC) recompute();
  return pager->getCols(); 
  }


// Determine content height of scroll area (char units)
FXint FXPLText::getContentHeight(){ 
  if (flags&FLAG_RECALC) recompute();
  return pager->numRows();
  }


// Set position - units of characters
void FXPLText::setPosition(FXint x,FXint y){
  FXint new_y;

  // Set scroll bars
  vertical->setPosition(-y);

  // Then read back valid position from scroll bar
  new_y=-vertical->getPosition();

  // Move content if there's a change
  if(new_y!=pos_y){
    moveContents(0,new_y);
    }
  }
 
/* ------------------------------------ */


  // Convert pixel units to character units (rounding -ve).  If no font, uses 1:1.
  // May be same vars in as out.
void FXPLText::pix2char(FXint w_pix, FXint h_pix, FXint & w_char, FXint & h_char) const {
  if (w_pix >= 0)
    w_char = w_pix / charWidth;
  else
    w_char = -((-w_pix + charWidth - 1) / charWidth);
  if (h_pix >= 0)
    h_char = h_pix / charHeight;
  else
    h_char = -((-h_pix + charHeight - 1) / charHeight);
  }
  
  // Convert pixel units to character units (rounding +ve).  If no font, uses 1:1
  // May be same vars in as out.
void FXPLText::pix2charUp(FXint w_pix, FXint h_pix, FXint & w_char, FXint & h_char) const {
  if (w_pix >= 0)
    w_char = (w_pix + charWidth - 1) / charWidth;
  else
    w_char = -(-w_pix / charWidth);
  if (h_pix >= 0)
    h_char = (h_pix + charHeight - 1) / charHeight;
  else
    h_char = -(-h_pix / charHeight);
  }
  


/* ------------------------------------ */

// If window can have focus
FXbool FXPLText::canFocus() const { return 1; }


// Make a valid position
FXint FXPLText::validPos(FXint pos) const {
  if (pos < 0) return 0;
  // Subtract 1 in the following, because the trailing newline is not counted.
  if (pos >= getLength()) return getLength()-1;
  return pos;
  }


// Enable the window
void FXPLText::enable(){
  if(!(flags&FLAG_ENABLED)){
    FXComposite::enable();
    update();
    }
  }


// Disable the window
void FXPLText::disable(){
  if(flags&FLAG_ENABLED){
    FXComposite::disable();
    update();
    }
  }


// Propagate size change
void FXPLText::recalc(){ 
  FXComposite::recalc();
  flags|=FLAG_RECALC;
  }


/*******************************************************************************/


// Get character
FXchar FXPLText::getChar(FXint pos) const {
  return textsrc->getChar(pos);
  }


// Set character
void FXPLText::setChar(FXint pos,FXchar ch){
  replaceText(pos, 1, &ch, 1);
  }


/*******************************************************************************/



 

// Find end of previous word
FXint FXPLText::leftWord(FXint pos) const {
  return textsrc->leftWord(pos);
  }


// Find begin of next word
FXint FXPLText::rightWord(FXint pos) const {
  return textsrc->rightWord(pos);
  }


// Find begin of a word
FXint FXPLText::wordStart(FXint pos) const {
  return textsrc->wordStart(pos);
  }


// Find end of word
FXint FXPLText::wordEnd(FXint pos) const {
  return textsrc->wordEnd(pos);
  }


// Return position of begin of paragraph 
FXint FXPLText::lineStart(FXint pos) const {
  return textsrc->lineStart(pos);
  }


// Return position of end of paragraph
FXint FXPLText::lineEnd(FXint pos) const {
  return textsrc->lineEnd(pos);
  }


// Return start of next line
FXint FXPLText::nextLine(FXint pos,FXint nl) const {
  if (nl <= 0) return pos;
  return textsrc->getPosOfLine(textsrc->getLineOfPos(pos) + nl);
  }


// Return start of previous line
FXint FXPLText::prevLine(FXint pos,FXint nl) const {
  if (nl <= 0) return pos;
  return textsrc->getPosOfLine(textsrc->getLineOfPos(pos) - nl);
  }


// Return row start
FXint FXPLText::rowStart(FXint pos) const {
  return pager->rowStart(pos);
  }


// Return row end
FXint FXPLText::rowEnd(FXint pos) const {
  return pager->rowEnd(pos);
  }


// Move to next row given start of line
FXint FXPLText::nextRow(FXint pos,FXint nr) const {
  if (nr <= 0) return pos;
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  return pager->getPosOfRow(pager->getRowOfPos(pos, atend) + nr);
  }


// Move to previous row given start of line
FXint FXPLText::prevRow(FXint pos,FXint nr) const {
  if (nr <= 0) return pos;
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  return pager->getPosOfRow(pager->getRowOfPos(pos, atend) - nr);
  }



/*******************************************************************************/

FXuint FXPLText::getCursorLine() const {
  return textsrc->getLineOfPos(getCursorPos());
  }

  // Return window X pixel coord of position
FXint FXPLText::getXOfPos(FXint pos) const { 
  FXint startpos;
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  FXint base = pager->getRowOfPos(pos, startpos, atend);
  FXASSERT(startpos <= pos);
  return auxwidth + marginleft + charwidth2pix(pager->rowWidth(base, pos - startpos));
  }

  // Return window Y pixel coord of position
FXint FXPLText::getYOfPos(FXint pos) const {
  FXint startpos;
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  FXint base = pager->getRowOfPos(pos, startpos, atend);
  return margintop + charheight2pix(base - pager->getTopRow());
  }

  // Return selection starting position
FXint FXPLText::getSelStartPos() const {
  return textsrc->getSelectionStart();
  }

  // Return selection ending position
FXint FXPLText::getSelEndPos() const {
  return textsrc->getSelectionEnd();
  }

// See if pos is a visible position
FXbool FXPLText::posVisible(FXint pos) const {
  FXint startpos;
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  FXint rr = pager->getRowOfPos(pos, startpos, atend) - pager->getTopRow();
  return rr >= 0 && rr < nvisrows; // Don't count any partially visible row
  }


// See if position is 'in' the selection, which includes the position right after the
// end of the selection.
FXbool FXPLText::isPosSelected(FXint pos) const {
  return hasSelection() && getSelStartPos() <= pos && pos <= getSelEndPos();
  }

FXTextAnchor * FXPLText::getPosSecSel(FXint pos) const {
  FXTextAnchor * ta;
  if (textsrc->findTextAnchors(pos, pos, 1, &ta, FXPFATTR_SECONDARY, FXPFATTR_SECONDARY, TRUE))
    return ta;
  return NULL;
  }



// Localize position at x,y.  y positions outside visible area return pos of first displayed
// char or last displayed char (+1) as appropriate.  x out-of-range returns first or last char
// of line (not counting newline char).
FXint FXPLText::getPosAt(FXint x,FXint y) const {
  /* There is a slight complication here: when a line is split into more than 1 row, then
     there are two valid display positions for the cursor at the character position corresponding
     to the split point.  The cursor can be displayed either at the start of the 2nd row,
     or at the end of the 1st.  The cursor text anchor itself maintains the
     preference, depending on where the user pressed the mouse button.  atend is set true
     iff the x press position was beyond the first half-char of the row.  moveCursorToRow()
     applies similar logic.
  */
  FXint tr = (FXint)pager->getTopRow();
  x -= auxwidth + marginleft;
  y -= margintop;
  FXuint a = cursor->getAttrs();
  if (x > (charWidth>>1))
    a |= FXTextAnchor::ANCHOR_S_ATEND;
  else
    a &= ~FXTextAnchor::ANCHOR_S_ATEND;
  cursor->setAttrs(a);
  pix2char(x+(charWidth>>1), y, x, y);
  y += tr;	// y is now row number
  if (y < tr)	// Above top, return first char of top line
    return pager->getPosOfRow(tr);
  else if (y >= tr + nvisrows)	// below bottom, return first char of next line
    return pager->getPosOfRow(tr + nvisrows);
  if (x < 0)
    return pager->getPosOfRow(y);
  return pager->rowPos(y, x, FALSE);	// Return pos not counting trailing newline
  }



// Force position to become fully visible (and not on partial line).  Note that this is
// assumed to be the cursor position, which it usually is when called from this class.  Thus,
// we use the cursor's ATEND information to disambiguate positions at line split points.
void FXPLText::makePositionVisible(FXint pos){
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  FXint r = (FXint)pager->getRowOfPos(pos, atend);
  FXint tr = (FXint)pager->getTopRow();
  want_visible = TRUE;
#define MORE_VISIBLE 3
#ifdef MORE_VISIBLE
  // In "more visible" mode we try to move the line of interest a bit more towards the
  // centre of the display that right at the top or bottom line.  This is only done if
  // a scrolling movement is needed -- if the position is within the current display, then
  // no scrolling occurs.  For a display of MORE_VISIBLE*2+1 full lines or less, we move 
  // to the middle line.  Otherwise, the position is set to MORE_VISIBLE lines from the
  // top or bottom.  Thus the guaranteed "context" around the position is at least 
  // MORE_VISIBLE-1 lines.
  if (r < tr)
    if (nvisrows <= MORE_VISIBLE*2+1)
      setPosition(0, -(r - (nvisrows>>1)));
    else
      setPosition(0, -(r - MORE_VISIBLE));
  else if (r >= tr+nvisrows)
    if (nvisrows <= MORE_VISIBLE*2+1)
      setPosition(0, -(r - (nvisrows>>1)));
    else
      setPosition(0, -(r - nvisrows + MORE_VISIBLE + 1));
#else
  if (r < tr)
    setPosition(0, -r);
  else if (r >= tr + nvisrows)
    setPosition(0, -(r - nvisrows + 1));
#endif
  }


// Make line containing pos the top visible line
void FXPLText::setTopLine(FXint pos){
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  setPosition(0, -(FXint)pager->getRowOfPos(pos, atend));
  }


// Make line containing pos the bottom (fully) visible line
void FXPLText::setBottomLine(FXint pos){
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  setPosition(0, -((FXint)pager->getRowOfPos(pos, atend) - nvisrows + 1));
  }


// Center line of pos in the middle of the screen
void FXPLText::setCenterLine(FXint pos){ 
  setPosition(0, -((FXint)pager->getRowOfPos(pos) - nvisrows/2 + 1));
  }


// Get top line (position of 1st char thereof).  This is a misnomer: is actually
// first char of top _row_.
FXint FXPLText::getTopLine() const {
  return pager->getPosOfRow(pager->getTopRow());
  }


// Get bottom line
FXint FXPLText::getBottomLine() const { 
  return pager->getPosOfRow(pager->getTopRow()+nvisrows-1);
  }



/*******************************************************************************/



// Replace m characters at pos by n characters
void FXPLText::replaceText(FXint pos,FXint m,const FXchar *text,FXint n){
  textsrc->replace(pos, pos+m, text, n, FALSE, this);
  ok = textsrc->isOk() && pager->isOk();
  if (!ok)
    fxwarning("%s::replaceText: error in text replacement\n",getClassName());
  }


// Add text at the end
void FXPLText::appendText(const FXchar *text,FXint n){
  replaceText(getLength()-1,0,text,n);
  }


// Insert some text at pos
void FXPLText::insertText(FXint pos,const FXchar *text,FXint n){
  replaceText(pos,0,text,n);
  }
    

// Remove some text at pos
void FXPLText::removeText(FXint pos,FXint n){
  replaceText(pos,n,NULL,0);
  }


// Replace text at pos with spaces.  Existing newlines are preserved.  This is
// used for 'backspace' in overstrike mode.
void FXPLText::blankText(FXint pos,FXint n){
  FXchar * p, * s, * t;
  FXMALLOC(&p, FXchar, n);
  extractText(p, pos, n);
  for (s = p, t = p+n; s < t; s++)
    if (!isspace(*s)) *s = ' ';
  replaceText(pos,n,p,n);
  FXFREE(&p);
  }


// Grab range of text
void FXPLText::extractText(FXchar *text,FXint pos,FXint n) const {
  while (n--)
    *text++ = getChar(pos++);
  }


// Completely reflow the text, because font or display size may have changed.
void FXPLText::recompute(){
  FXint xchars, ychars;
  FXint xcharsup, ycharsup;
  // See explanation in layout() as to why we subtract the scrollbar width.
  FXint w = getViewportWidth()-auxwidth-marginleft-marginright - vertical->getDefaultWidth();
  FXint h = getViewportHeight()-margintop-marginbottom;
  
  // Convert to char coordinates, rounding down and up.
  pix2char(w, h, xchars, ychars);
  pix2charUp(w, h, xcharsup, ycharsup);
  if (ychars > 0) {
    nvisrows = ychars;
    partialRow = (FXint)(ycharsup > ychars);
    }
  else {
    // First row not completely visible.  Treat this case as if it's fully visible,
    // and no partial row.
    nvisrows = 1;
    partialRow = 0;
    }
  
  
  if (xchars < 2)	// FIXME: this could be an option parameter.
    xchars = 2;		// Always at least this many chars across
  
  // Ensure row update array is allocated to correct dimension
  FXRESIZE(&rds, RowDirty, nvisrows+partialRow);
  
  flags&=~(FLAG_RECALC|FLAG_DIRTY);
  // This eventually causes redraw.
  pager->resize(nvisrows + partialRow, xchars);
  }


// Change the text in the buffer to new text
void FXPLText::setText(const FXchar* text,FXint n){
  FXASSERT(n >= 0);
  replaceText(0, getLength()-1, text, n);
  }


// Retrieve text into buffer
void FXPLText::getText(FXchar* text,FXint n) const {
  extractText(text, 0, n);
  }

// Recompute character metrics.  Line spacing not smaller than aux font, unless greek.
void FXPLText::setCharMetrics() {
  if (options & PLTEXT_GREEK) {
    charWidth = greek_x;
    charHeight = greek_y;
    charBase = greek_y > 2 ? 1 : 0;
  }
  else if (font->id() && auxFont->id()) {
    FXASSERT(font->isFontMono());
    auxCharWidth = auxFont->getTextWidth("0",1);
    auxCharHeight = auxFont->getFontSpacing();
    auxCharBase = auxFont->getFontAscent();
    charWidth = font->getTextWidth("0",1);
    charHeight = font->getFontSpacing();
    charBase = font->getFontAscent();
    if (charHeight < auxCharHeight) {
      charHeight = auxCharHeight;
      }
    }
  }

// Change the font(s)
void FXPLText::setFont(FXFont* fnt, FXFont * auxf){
  FXbool changed = FALSE;
  
  if (fnt && font!=fnt){
    font = fnt;
    changed = TRUE;
    }
  if (auxf && auxf != auxFont) {
    auxFont = auxf;
    changed = TRUE;
    }
  if (changed) {
    setCharMetrics();
    recalc();
    if (id()) { layout(); update(); }
    }

  }

void FXPLText::setGreek(FXbool on, FXint xpix, FXint ypix, FXPLText * gtgt) {
  FXbool gk = !!(options & PLTEXT_GREEK);
  FXbool changed;
  FXASSERT(xpix >= 0 && xpix <= 2);
  FXASSERT(ypix >= 0 && xpix <= 3);
  track = gtgt;
  changed = on ^ gk || on && (xpix && xpix != greek_x || ypix && ypix != greek_y);
  if (xpix) greek_x = xpix;
  if (ypix) greek_y = ypix;
  if (changed) {
    if (on) {
      options |= PLTEXT_GREEK;
      auxwidth = 0;
      }
    else {
      options &= ~PLTEXT_GREEK;
      auxwidth = set_auxwidth;
      }
    setCharMetrics();
    recalc();
    if (id()) { layout(); update(); }
    }
  }


/*******************************************************************************/




// Recalculate layout
void FXPLText::layout(){
  /* ScrollArea stuff...
     With word-wrapping, layout is a bit tricky since the content height may depend on
     whether or not the scrollbar is shown.  We vastly simplify the problem by assuming
     the scrollbar is always shown, for the purpose of computing content width.  This
     may leave unused space when the scrollbar is hidden, but has the advantage of not
     requiring us to try both (potentially expensive) re-wrappings of the entire text.  */
  register FXint new_y;
  register FXint sv_w=0;
  
  //ungrab(); // for debugging
  
  // Inviolate
  FXASSERT(pos_y<=0);

  // Initial viewport size (pixels) - initially assume no scrollbar
  viewport_w_pix = width;
  viewport_h_pix = height;
  //printf("FXPLText::layout w=%d h=%d\n", width, height);
  content_w = viewport_w = getContentWidth();	// Number of columns visible.
  content_h = getContentHeight(); // Calls recalc() which calls pager->resize() which
  				// calls this->pagerResize() which updates display.
  viewport_h = nvisrows;	// Number of visible rows (not counting partial).
  FXTRACE((100,"vwp=%d vhp=%d vw=%d vh=%d cw=%d ch=%d\n", 
           viewport_w_pix, viewport_h_pix, viewport_w, viewport_h, content_w, content_h));

  // Get dimensions of the scroll bar (pixels)
  if(!(options&PLTEXT_VSCROLLER_NEVER)) sv_w=vertical->getDefaultWidth();
  

  // Should we disable the scroll bar?
  if(!(options&PLTEXT_VSCROLLER_ALWAYS) && (content_h<=viewport_h)) sv_w=0;

  // Viewport size with scroll bars taken into account
  viewport_w_pix -= sv_w;

  // Adjust content size, now that we know about scroll bar
  if((options&PLTEXT_VSCROLLER_NEVER)&&(options&PLTEXT_VSCROLLER_ALWAYS)) content_h=viewport_h;

  // Furthermore, content size won't be smaller than the viewport
  if(content_h<viewport_h) content_h=viewport_h;

  // Content size
  vertical->setRange(content_h);

  // Page size may have changed.  We subtract 1 if greater than 1, so that one line from
  // the previous 'page' is still visible.
  vertical->setPage(viewport_h - (viewport_h > 1));
  
  // Position may have changed
  vertical->setPosition(-pos_y);
 
  // Get back the adjusted position
  new_y=-vertical->getPosition();
  
  // If no scrollbar visible, and neither 'never' nor 'always', ensure top line is first line.
  if (!sv_w && !(options&PLTEXT_VSCROLLER_ALWAYS) && !(options&PLTEXT_VSCROLLER_NEVER))
    new_y = 0; 

  // Scroll to force position back into range
  if(new_y!=pos_y){
    moveContents(0,new_y);
    }

  // Read back validated position
  pos_y=-vertical->getPosition();

  // Hide or show vertical scroll bar
  if(sv_w){
    vertical->position(width-sv_w,0,sv_w,height);
    vertical->show();
    vertical->raise();
    }
  else{
    vertical->hide();
    }

  flags&=~FLAG_DIRTY;
  }


// The widget is resized
void FXPLText::resize(FXint w,FXint h){
  printf("FXPLText::resize w=%d h=%d\n", w,h);
  flags |= FLAG_DIRTY | FLAG_RECALC;
  FXComposite::resize(w, h);
  }


// The widget is moved and possibly resized
void FXPLText::position(FXint x,FXint y,FXint w,FXint h){
  flags |= FLAG_DIRTY | FLAG_RECALC;
  FXComposite::position(x, y, w, h); 
  }


/*******************************************************************************/


// Gained focus
long FXPLText::onFocusIn(FXObject* sender,FXSelector sel,void* ptr){
  if ((options & PLTEXT_EDITMASK) != PLTEXT_READONLY) {
    cursor->setVisible(TRUE);
    cursorActive = TRUE;
    }
  return 1;
  }


// Lost focus
long FXPLText::onFocusOut(FXObject* sender,FXSelector sel,void* ptr){
  cursor->setVisible(FALSE);
  cursorActive = FALSE;
  return 1;
  }

FXbool FXPLText::isActive() const {
  return cursor->isVisible(this); 
  }

// We were asked about status text
long FXPLText::onQueryHelp(FXObject* sender,FXSelector,void*){
  if(!help.empty() && (flags&FLAG_HELP)){
    sender->handle(this,MKUINT(ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&help);
    return 1;
    }
  return 0;
  }



// Pressed left button
long FXPLText::onLeftBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  flags&=~FLAG_TIP;
  if(isEnabled()){
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONPRESS),ptr)) return 1;
    flags|=FLAG_PRESSED;
    flags&=~FLAG_UPDATE;
    if(ev->click_count==1){
      pos=getPosAt(ev->win_x,ev->win_y);
      FXTRACE((150,"getPosAt(%d,%d) = %d getYOfPos(%d) = %d getXOfPos(%d)=%d\n",ev->win_x,ev->win_y,pos,pos,getYOfPos(pos),pos,getXOfPos(pos)));
      if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
      if(ev->state&SHIFTMASK){
        extendSelection(pos,PLTEXT_SELECT_CHARS);
        }
      else{
        killSelection();
        setAnchorPos(pos);
        }
      dragmode=MOUSE_CHARS;
      }
    
    // Select word by double click
    else if(ev->click_count==2){
      //if (options & PLTEXT_PREPEND)
        pos = getPosAt(ev->win_x, ev->win_y);
      //else
      //  pos = getCursorPos();
      setAnchorPos(pos);
      extendSelection(pos,PLTEXT_SELECT_WORDS);
      dragmode=MOUSE_WORDS;
      }
    
    // Select line by triple click
    else{
      //if (options & PLTEXT_PREPEND)
        pos = getPosAt(ev->win_x, ev->win_y);
      //else
      //  pos = getCursorPos();
      setAnchorPos(pos);
      extendSelection(pos,PLTEXT_SELECT_LINES);
      dragmode=MOUSE_LINES;
      }
    return 1;
    }
  return 0;
  }


// Released left button
long FXPLText::onLeftBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){ 
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    stopAutoScroll();
    if(target && target->handle(this,MKUINT(message,SEL_LEFTBUTTONRELEASE),ptr)) return 1;
    dragmode=MOUSE_NONE;
    return 1;
    }
  return 0;
  }


// Pressed middle button
long FXPLText::onMiddleBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  flags&=~FLAG_TIP;
  if(isEnabled()){
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONPRESS),ptr)) return 1;
    pos=getPosAt(ev->win_x,ev->win_y);
    
    // Move over
    if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
    
    // If greek with track target, tell the target to reposition its cursor.
    if (options & PLTEXT_GREEK && track) {
      track->handle(this, MKUINT(0, SEL_CHANGED), (void *)pos);
      track->setFocus();
      }
      
    // Start text drag
    else if(isPosSelected(pos)){
      dragmode=MOUSE_TRYDRAG;
      flags|=FLAG_PRESSED;
      flags&=~FLAG_UPDATE;
      }
 
    // Place cursor
    else if(isEditable()){
      dragmode=MOUSE_PASTE;
      flags|=FLAG_PRESSED;
      flags&=~FLAG_UPDATE;
      }
    
    // Beep
    else{
      getApp()->beep();
      }
    return 1;
    }
  return 0;
  }


// Released middle button
long FXPLText::onMiddleBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    stopAutoScroll();
    if(target && target->handle(this,MKUINT(message,SEL_MIDDLEBUTTONRELEASE),ptr)) return 1;

    // Drop text somewhere
    if(dragmode==MOUSE_DRAG){
      handle(this,MKUINT(0,SEL_ENDDRAG),ptr);
      }

    // Paste text from selection (X-Windows only)
#ifndef WIN32
    else if(dragmode==MOUSE_PASTE){
      FXuchar *data;
      FXuint  size;
      if(getDNDData(FROM_SELECTION,stringType,data,size)){
        if (FXRESIZE(&data, FXuchar, size+1)) {
          data[size] = 0;
          handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
          }
        FXFREE(&data);
        }
      }
#endif
    dragmode=MOUSE_NONE;
    return 1;
    }
  return 0;
  }


// Pressed right button
long FXPLText::onRightBtnPress(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  flags&=~FLAG_TIP;
  if(isEnabled()){
    handle(this,MKUINT(0,SEL_FOCUS_SELF),ptr);
    grab();
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONPRESS),ptr)) return 1;
    flags&=~FLAG_UPDATE;
    flags|=FLAG_PRESSED;
    grabx=ev->win_x-pos_x;
    graby=ev->win_y-pos_y;
    dragmode=MOUSE_SCROLL;
    return 1; 
    }
  return 0;
  }


// Released right button
long FXPLText::onRightBtnRelease(FXObject*,FXSelector,void* ptr){
  if(isEnabled()){
    ungrab();
    flags&=~FLAG_PRESSED;
    flags|=FLAG_UPDATE;
    if(target && target->handle(this,MKUINT(message,SEL_RIGHTBUTTONRELEASE),ptr)) return 1;
    dragmode=MOUSE_NONE;
    return 1;
    }
  return 0;
  }


// The widget lost the grab for some reason
long FXPLText::onUngrabbed(FXObject* sender,FXSelector sel,void* ptr){
  FXComposite::onUngrabbed(sender,sel,ptr);
  flags&=~FLAG_PRESSED;
  flags|=FLAG_UPDATE;
  stopAutoScroll();
  return 1;
  }


// Autoscroll timer fired 
long FXPLText::onAutoScroll(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint pos;
  FXint dx,dy;
  FXint xx,yy;
  FXuint state;
  
  scrolltimer=NULL;
  
  // Autoscroll while close to the wall
  dx=0;
  dy=0;
  
  // Where's the (mouse) cursor?
  getCursorPosition(xx,yy,state);
 
  // If scrolling only while inside, and not inside, we stop scrolling
  if((flags&FLAG_SCROLLINSIDE) && !(0<=xx && 0<=yy && xx<viewport_w_pix && yy<viewport_h_pix)) return 0;
 
  // Figure scroll amount x
  /* Not implemented...
  if(xx<AUTOSCROLL_FUDGE) dx=AUTOSCROLL_FUDGE-xx;
  else if(viewport_w_pix-AUTOSCROLL_FUDGE<=xx) dx=viewport_w_pix-AUTOSCROLL_FUDGE-xx;
  */
  
  // Figure scroll amount y
  if(yy<AUTOSCROLL_FUDGE) dy=AUTOSCROLL_FUDGE-yy;
  else if(viewport_h_pix-AUTOSCROLL_FUDGE<=yy) dy=viewport_h_pix-AUTOSCROLL_FUDGE-yy;
  
  // Keep autoscrolling
  if(dx || dy){
    FXint oldposx=pos_x;
    FXint oldposy=pos_y;
    if(flags&FLAG_SCROLLINSIDE){ 
      FXASSERT(FXABS(dx)<=AUTOSCROLL_FUDGE);
      FXASSERT(FXABS(dy)<=AUTOSCROLL_FUDGE);
      dx*=acceleration[FXABS(dx)]; 
      dy*=acceleration[FXABS(dy)]; 
      }
    
    // Scroll a bit
    setPosition(pos_x+dx,pos_y+dy); 
    
    // Setup next timer if we can still scroll some more
    if((pos_x!=oldposx) || (pos_y!=oldposy)){
      scrolltimer=getApp()->addTimeout(getApp()->getScrollSpeed(),this,SELID(sel));
      }
    }
  
  // Kill tip
  flags&=~FLAG_TIP;


  switch(dragmode){
    case MOUSE_CHARS:
      if((fxabs(ev->win_x-ev->click_x)>getApp()->getDragDelta())||(fxabs(ev->win_y-ev->click_y)>getApp()->getDragDelta())){
        pos=getPosAt(ev->win_x,ev->win_y);
        extendSelection(pos,PLTEXT_SELECT_CHARS);
        if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
        }
      return 1;
    case MOUSE_WORDS:
      if((fxabs(ev->win_x-ev->click_x)>getApp()->getDragDelta())||(fxabs(ev->win_y-ev->click_y)>getApp()->getDragDelta())){
        pos=getPosAt(ev->win_x,ev->win_y);
        extendSelection(pos,PLTEXT_SELECT_WORDS);
        if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
        }
      return 1;
    case MOUSE_LINES:
      if((fxabs(ev->win_x-ev->click_x)>getApp()->getDragDelta())||(fxabs(ev->win_y-ev->click_y)>getApp()->getDragDelta())){
        pos=getPosAt(ev->win_x,ev->win_y);
        extendSelection(pos,PLTEXT_SELECT_LINES);
        if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
        }
      return 1;
    }
  return 0;
  }


// Handle real or simulated mouse motion
long FXPLText::onMotion(FXObject*,FXSelector,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXint delta = options & PLTEXT_GREEK ? 2 : getApp()->getDragDelta();
  FXint pos;
  if(flags&FLAG_PRESSED){
    switch(dragmode){
      case MOUSE_CHARS:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>delta)||(fxabs(ev->win_y-ev->click_y)>delta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          extendSelection(pos,PLTEXT_SELECT_CHARS);
          if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      case MOUSE_WORDS:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>delta)||(fxabs(ev->win_y-ev->click_y)>delta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          extendSelection(pos,PLTEXT_SELECT_WORDS);
          if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      case MOUSE_LINES:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>delta)||(fxabs(ev->win_y-ev->click_y)>delta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          extendSelection(pos,PLTEXT_SELECT_LINES);
          if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      case MOUSE_SCROLL:
        //FXint cx, cy;
        setPosition(ev->win_x-grabx,ev->win_y-graby);
        break;
      case MOUSE_DRAG:
        handle(this,MKUINT(0,SEL_DRAGGED),ptr);
        break;
      case MOUSE_TRYDRAG:
        if(ev->moved){
          dragmode=MOUSE_NONE;
          if(handle(this,MKUINT(0,SEL_BEGINDRAG),ptr)){
            dragmode=MOUSE_DRAG;
            }
          }
        break;
      case MOUSE_PASTE:
        if(startAutoScroll(ev->win_x,ev->win_y,FALSE)) return 1;
        if((fxabs(ev->win_x-ev->click_x)>delta)||(fxabs(ev->win_y-ev->click_y)>delta)){
          pos=getPosAt(ev->win_x,ev->win_y);
          if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
          }
        break;
      }
    return 1;
    }
  return 0;
  }


// Cursor location changed
long FXPLText::onChanged(FXObject*,FXSelector,void* ptr) {
  FXuint eopts = options & PLTEXT_EDITMASK;
  FXint pos = (long)ptr;
  if (eopts == PLTEXT_APPEND) {
    if ((FXint)textsrc->getLineOfPos(pos) < rolines) return 1;
    }
  else if (eopts == PLTEXT_PREPEND) {
    if ((FXint)textsrc->getLineOfPos(pos) >= (FXint)textsrc->lines() - rolines) return 1;
    }
  makePositionVisible(pos);
  setCursorPos(pos);
  if (setpref) {
    setpref++;
    if (setpref > 2)
      setpref = 0;
    }
  if(target) target->handle(this,MKUINT(message,SEL_CHANGED),ptr);
  return 1;
  }


// Selected text
long FXPLText::onSelected(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_SELECTED),ptr);
  }


// Deselected text
long FXPLText::onDeselected(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_DESELECTED),ptr);
  }


// Inserted text
long FXPLText::onInserted(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_INSERTED),ptr);
  }


// Deleted text
long FXPLText::onDeleted(FXObject*,FXSelector,void* ptr){
  return target && target->handle(this,MKUINT(message,SEL_DELETED),ptr);
  }


/*******************************************************************************/


// Start a drag operation
long FXPLText::onBeginDrag(FXObject* sender,FXSelector sel,void* ptr){
  if(FXComposite::onBeginDrag(sender,sel,ptr)) return 1;
  beginDrag(&textType,1);
  setDragCursor(getApp()->getDefaultCursor(DEF_DNDSTOP_CURSOR));
  return 1;
  }


// End drag operation
long FXPLText::onEndDrag(FXObject* sender,FXSelector sel,void* ptr){
  if(FXComposite::onEndDrag(sender,sel,ptr)) return 1;
  endDrag((didAccept()!=DRAG_REJECT));
  setDragCursor(getApp()->getDefaultCursor(DEF_TEXT_CURSOR));
  return 1;
  }


// Dragged stuff around
long FXPLText::onDragged(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXDragAction action;
  if(FXComposite::onDragged(sender,sel,ptr)) return 1;
  action=DRAG_COPY;
  if(isEditable()){
    if(isDropTarget()) action=DRAG_MOVE;
    if(ev->state&CONTROLMASK) action=DRAG_COPY;
    if(ev->state&SHIFTMASK) action=DRAG_MOVE;
    }
  handleDrag(ev->root_x,ev->root_y,action);
  if(didAccept()!=DRAG_REJECT){
    if(action==DRAG_MOVE)
      setDragCursor(getApp()->getDefaultCursor(DEF_DNDMOVE_CURSOR));
    else
      setDragCursor(getApp()->getDefaultCursor(DEF_DNDCOPY_CURSOR));
    }
  else{
    setDragCursor(getApp()->getDefaultCursor(DEF_DNDSTOP_CURSOR));
    }
  return 1;
  }


// Handle drag-and-drop enter
long FXPLText::onDNDEnter(FXObject* sender,FXSelector sel,void* ptr){
  FXComposite::onDNDEnter(sender,sel,ptr);
  return 1;
  }


// Handle drag-and-drop leave
long FXPLText::onDNDLeave(FXObject* sender,FXSelector sel,void* ptr){
  stopAutoScroll();
  FXComposite::onDNDLeave(sender,sel,ptr);
  return 1;
  }


// Handle drag-and-drop motion
long FXPLText::onDNDMotion(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent* ev=(FXEvent*)ptr;
  FXDragAction action;
  FXint pos;
    
  // Scroll into view  
  if(startAutoScroll(ev->win_x,ev->win_y,TRUE)) return 1;

  // Handled elsewhere
  if(FXComposite::onDNDMotion(sender,sel,ptr)) return 1;

  // Correct drop type
  if(offeredDNDType(FROM_DRAGNDROP,textType)){
    
    // Is target editable?
    if(isEditable()){
      action=inquireDNDAction();
 
      // Check for legal DND action
      if(action==DRAG_COPY || action==DRAG_MOVE){
 
        // Get the suggested drop position
        pos=getPosAt(ev->win_x,ev->win_y);
 
        // Move cursor to new position
        if(getCursorPos()!=pos) handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
 
        // We don't accept a drop on the selection
        if(!isPosSelected(pos)){
          acceptDrop(DRAG_ACCEPT);
          }
        }
      }
    return 1;
    }
  
  // Didn't handle it here
  return 0;
  }


// Handle drag-and-drop drop
long FXPLText::onDNDDrop(FXObject* sender,FXSelector sel,void* ptr){
  FXuchar *data;
  FXuchar *junk; 
  FXuint   dum;
  FXuint   len;

  // Stop scrolling
  stopAutoScroll();

  // Try handling it in base class first
  if(FXComposite::onDNDDrop(sender,sel,ptr)) return 1;
  
  // Should really not have gotten this if non-editable
  if(isEditable()){
    
    // Try handle here
    if(getDNDData(FROM_DRAGNDROP,textType,data,len)){
 
      // Need to ask the source to delete his copy
      if(inquireDNDAction()==DRAG_MOVE){
        getDNDData(FROM_DRAGNDROP,deleteType,junk,dum);
        FXASSERT(!junk);
        }

      // Insert the new text
      if (FXRESIZE(&data, FXuchar, len+1)) {
        data[len] = 0;
        handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
        }
      FXFREE(&data);
      }
    return 1;
    }
  return 0;
  }


// Service requested DND data
long FXPLText::onDNDRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data;
  
  // Perhaps the target wants to supply its own data
  if(FXComposite::onDNDRequest(sender,sel,ptr)) return 1;
  
  // Return dragged text
  if(event->target==textType){
    FXASSERT(getSelStartPos()<=getSelEndPos());
    FXint len=getSelEndPos()-getSelStartPos();
    FXMALLOC(&data,FXuchar,len);
    if(!data){ 
      fxwarning("%s::onDNDRequest: out of memory\n",getClassName()); 
      return 0;
      }
    extractText((FXchar*)data,getSelStartPos(),len);
    setDNDData(FROM_DRAGNDROP,stringType,data,len);
    return 1;
    }

  // Delete dragged text
  if(event->target==deleteType){
    if(isEditable()){
      handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
      }
    return 1;
    }

  return 0;
  }


/*******************************************************************************/


// We now really do have the selection
long FXPLText::onSelectionGained(FXObject* sender,FXSelector sel,void* ptr){
  FXComposite::onSelectionGained(sender,sel,ptr);
  return 1;
  }


// We lost the selection somehow 
long FXPLText::onSelectionLost(FXObject* sender,FXSelector sel,void* ptr){
  FXbool hadselection=hasSelection();
  FXComposite::onSelectionLost(sender,sel,ptr);
  if(hadselection) textsrc->killSelection();
  return 1;
  }


// Somebody wants our selection
long FXPLText::onSelectionRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;
  FXuchar *data;

  // Perhaps the target wants to supply its own data for the selection
  if(FXComposite::onSelectionRequest(sender,sel,ptr)) return 1;
  
  // Return text of the selection
  if(event->target==stringType){
    FXASSERT(getSelStartPos() <= getSelEndPos());
    FXint len=getSelEndPos()-getSelStartPos();
    FXMALLOC(&data,FXuchar,len);
    if(!data){ 
      fxwarning("%s::onSelectionRequest: out of memory\n",getClassName()); 
      return 0;
      }
    extractText((FXchar*)data,getSelStartPos(),len);
    setDNDData(FROM_SELECTION,stringType,data,len);
    return 1;
    }

  return 0;
  }


/*******************************************************************************/
   
   
// We now really do have the selection
long FXPLText::onClipboardGained(FXObject* sender,FXSelector sel,void* ptr){
  FXComposite::onClipboardGained(sender,sel,ptr);
  return 1;
  }


// We lost the selection somehow 
long FXPLText::onClipboardLost(FXObject* sender,FXSelector sel,void* ptr){
  FXComposite::onClipboardLost(sender,sel,ptr);
  FXFREE(&clipbuffer);
  clipbuffer=NULL;
  cliplength=0;
  return 1;
  }


// Somebody wants our selection
long FXPLText::onClipboardRequest(FXObject* sender,FXSelector sel,void* ptr){
  FXEvent *event=(FXEvent*)ptr;

  // Try handling it in base class first
  if(FXComposite::onClipboardRequest(sender,sel,ptr)) return 1;

  // Requested data from clipboard
  if(event->target==stringType){
#ifndef WIN32
    FXMALLOC(&clipcopy,FXuchar,cliplength);
    if(!clipcopy){ 
      fxwarning("%s::onClipboardRequest: out of memory\n",getClassName()); 
      return 0;
      }
    memcpy(clipcopy,clipbuffer,cliplength);
    setDNDData(FROM_CLIPBOARD,stringType,clipcopy,cliplength); 
#else
    FXint i,j,c;
    FXMALLOC(&clipcopy,FXuchar,2*cliplength+1);
    if(!clipcopy){ 
      fxwarning("%s::onClipboardRequest: out of memory\n",getClassName()); 
      return 0;
      }
    for(i=j=0; j<cliplength; j++){
      c=clipbuffer[j]; 
      if(c=='\n'){clipcopy[i++]='\r';clipcopy[i++]='\n';}else{clipcopy[i++]=c;}
      }
    clipcopy[i] = 0;
    setDNDData(FROM_CLIPBOARD,stringType,clipcopy,i+1);
#endif
    return 1;
    }

  return 0;
  }

/*******************************************************************************/


// Keyboard press
long FXPLText::onKeyPress(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  char entry;
  FXbool edable, shift, cntl, esc;
  
  flags&=~FLAG_TIP;
  if(!isEnabled()) return 0;
  FXTRACE((200,"%s::onKeyPress keysym=0x%04x state=%04x\n",getClassName(),event->code,event->state));
  if(target && target->handle(this,MKUINT(message,SEL_KEYPRESS),ptr)) return 1;
  edable = isEditable();
  shift = !!(event->state & SHIFTMASK);
  cntl = !!(event->state & CONTROLMASK);
  if (event->code != KEY_Shift_L &&
      event->code != KEY_Shift_R &&
      event->code != KEY_Control_L &&
      event->code != KEY_Control_R)
    esc = setEscape(FALSE);	// Reset ESC but remember old
  else
    esc = isEscape();
  
  if(!(event->state&ALTMASK)){
    switch(event->code){
      case KEY_Shift_L:
      case KEY_Shift_R:
      case KEY_Control_L:
      case KEY_Control_R:
        if(dragmode==MOUSE_DRAG) handle(this,MKUINT(0,SEL_DRAGGED),ptr);
        return 1;
      case KEY_Up:
      case KEY_KP_Up:
        if (!edable) { setPosition(0, pos_y+1); return 1; }
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        handle(this,MKUINT(ID_CURSOR_UP,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_Down:
      case KEY_KP_Down:
        if (!edable) { setPosition(0, pos_y-1); return 1; }
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        handle(this,MKUINT(ID_CURSOR_DOWN,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_Left:
      case KEY_KP_Left:
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        if(cntl) handle(this,MKUINT(ID_CURSOR_WORD_LEFT,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_CURSOR_LEFT,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_Right:
      case KEY_KP_Right:
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        if(cntl) handle(this,MKUINT(ID_CURSOR_WORD_RIGHT,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_CURSOR_RIGHT,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_Home:
      case KEY_KP_Home:
        if (!edable) { setPosition(0, 0); return 1; }
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        if(cntl) handle(this,MKUINT(ID_CURSOR_TOP,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_CURSOR_HOME,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_End:
      case KEY_KP_End:
        if (!edable) { setPosition(0, nvisrows - getContentHeight()); return 1; }
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        if(cntl) handle(this,MKUINT(ID_CURSOR_BOTTOM,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_CURSOR_END,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_Page_Up:
        if (!edable) { setPosition(0, pos_y + (nvisrows > 1 ? nvisrows-1 : 1)); return 1; }
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        handle(this,MKUINT(ID_CURSOR_PAGEUP,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_Page_Down:
        if (!edable) { setPosition(0, pos_y - (nvisrows > 1 ? nvisrows-1 : 1)); return 1; }
        if(!shift) handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
        handle(this,MKUINT(ID_CURSOR_PAGEDOWN,SEL_COMMAND),NULL);
        if(shift) handle(this,MKUINT(ID_EXTEND,SEL_COMMAND),NULL);
        else handle(this,MKUINT(ID_MARK,SEL_COMMAND),NULL);
        return 1;
      case KEY_Insert:
      case KEY_KP_Insert:
        if(cntl) handle(this,MKUINT(ID_COPY_SEL,SEL_COMMAND),NULL);
        else if(shift){
          if(edable) handle(this,MKUINT(ID_PASTE_SEL,SEL_COMMAND),NULL);
          else getApp()->beep();
          }
        else handle(this,MKUINT(ID_TOGGLE_OVERSTRIKE,SEL_COMMAND),NULL);
        return 1;
      case KEY_Delete:
      case KEY_KP_Delete:
        if(edable){
          if(isPosSelected(getCursorPos()))
            handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
          else{
            handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
            if(event->state&CONTROLMASK)
              handle(this,MKUINT(ID_DELETE_WORD,SEL_COMMAND),NULL);
            else if(event->state&SHIFTMASK)
              handle(this,MKUINT(ID_DELETE_EOL,SEL_COMMAND),NULL);
            else
              handle(this,MKUINT(ID_DELETE,SEL_COMMAND),NULL);
            }
          }
        else getApp()->beep();
        return 1;
      case KEY_BackSpace:
        if(edable){
          if(isPosSelected(getCursorPos()))
            handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
          else{
           handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
           if(event->state&CONTROLMASK)
              handle(this,MKUINT(ID_BACKSPACE_WORD,SEL_COMMAND),NULL);
            else if(event->state&SHIFTMASK)
              handle(this,MKUINT(ID_BACKSPACE_BOL,SEL_COMMAND),NULL);
            else
              handle(this,MKUINT(ID_BACKSPACE,SEL_COMMAND),NULL);
            }
          }
        else getApp()->beep();
        return 1;
      case KEY_Return:
      case KEY_KP_Enter:
      case KEY_ISO_Enter:
        if(edable){
          if(isPosSelected(getCursorPos()))
            handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),(void *)1); // Defer update
          else
            handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
          // The parm is 1 to indicate that autoindent should be done if option is set.
          handle(this,MKUINT(ID_INSERT_NEWLINE,SEL_COMMAND),(void *)(FXint)1);
          }
        else getApp()->beep();
        return 1;
        
      // Some non-standard bindings
      case KEY_F7 : 
      case KEY_F19 : 
        return handle(this, 
          MKUINT(event->state & SHIFTMASK ? ID_FINDPREV : ID_FINDNEXT,SEL_COMMAND), ptr);
      case KEY_F8:
      case KEY_F20:
        return handle(this, 
          MKUINT(event->state & SHIFTMASK ? ID_ROLLFORWARD : ID_ROLLBACK,SEL_COMMAND), (void *)(FXint)1);
      case KEY_F10 : return handle(this, MKUINT(FXPLText::ID_COPY_SEL,SEL_COMMAND), ptr); break;
      case KEY_F11 : return handle(this, MKUINT(FXPLText::ID_PASTE_SEL,SEL_COMMAND), ptr); break;
      case KEY_F12 : return handle(this, MKUINT(FXPLText::ID_CUT_SEL,SEL_COMMAND), ptr); break;
      
      case KEY_Escape:
        return esc ? 1 : handle(this, MKUINT(ID_TOGGLE_ESCAPE,SEL_COMMAND), 0);
      
      default:
        if(event->state&CONTROLMASK){   
          // Control key.  Note that we ignore the caps lock key - shift must be
          // depressed to get the alternate function.
          entry = event->code;
          if (islower(entry))
            entry = toupper(entry);
          if (esc) {
            entry -= 'A'-1;	// Convert to control char
            goto char_insert;
            }
          if (!shift)
            entry = tolower(entry);
          switch(entry){
            case KEY_a:
              handle(this,MKUINT(ID_SELECT_ALL,SEL_COMMAND),NULL);
              return 1;
            case KEY_x:
              if(edable) handle(this,MKUINT(ID_CUT_SEL,SEL_COMMAND),NULL);
              else getApp()->beep();
              return 1;
            case KEY_c:
              handle(this,MKUINT(ID_COPY_SEL,SEL_COMMAND),NULL);
              return 1;
            case KEY_v:
              if(edable) handle(this,MKUINT(ID_PASTE_SEL,SEL_COMMAND),NULL);
              else getApp()->beep();
              return 1;
            case KEY_k:
              if(edable) handle(this,MKUINT(ID_DELETE_LINE,SEL_COMMAND),NULL);
              else getApp()->beep();
              return 1;
            case KEY_h:
              handle(this,MKUINT(ID_CURSOR_SCRNTOP,SEL_COMMAND),NULL);
              return 1;
            case KEY_m:
              handle(this,MKUINT(ID_CURSOR_SCRNCTR,SEL_COMMAND),NULL);
              return 1;
            case KEY_l:
              handle(this,MKUINT(ID_CURSOR_SCRNBTM,SEL_COMMAND),NULL);
              return 1;
            case KEY_g:
              handle(this, MKUINT(ID_FINDNEXT,SEL_COMMAND), NULL);
              return 1;
            case KEY_G:
              handle(this, MKUINT(ID_FINDPREV,SEL_COMMAND), NULL);
              return 1;
            case KEY_f:
              handle(this, MKUINT(ID_FIND,SEL_COMMAND), NULL);
              return 1;
            case KEY_e:
              handle(this, MKUINT(ID_GOTO_SEL_LINE,SEL_COMMAND), NULL);
              return 1;
            case KEY_i:
              if (edable) handle(this, MKUINT(ID_INDENT,SEL_COMMAND), (void *)(FXint)1); else getApp()->beep();
              return 1;
            case KEY_I:
              if (edable) handle(this, MKUINT(ID_INDENT,SEL_COMMAND), (void *)(FXint)-1); else getApp()->beep();
              return 1;
            case KEY_z:
              if (edable) handle(this, MKUINT(ID_ROLLBACK,SEL_COMMAND), (void *)(FXint)1); else getApp()->beep();
              return 1;
            case KEY_Z:
              if (edable) handle(this, MKUINT(ID_ROLLFORWARD,SEL_COMMAND), (void *)(FXint)1); else getApp()->beep();
              return 1;
            case KEY_y:
              handle(this, MKUINT(ID_CURSOR_TO_ROLLBACK,SEL_COMMAND), NULL);
              return 1;
            case KEY_Y:
              handle(this, MKUINT(ID_CURSOR_TO_ROLLFORWARD,SEL_COMMAND), NULL);
              return 1;
            case KEY_t:
              handle(this, MKUINT(ID_TOGGLE_SHOWTABS,SEL_COMMAND), NULL);
              return 1;
            case KEY_n:
              handle(this, MKUINT(ID_TOGGLE_SHOWAUX,SEL_COMMAND), NULL);
              return 1;
            case KEY_slash:
              handle(this, MKUINT(ID_TOGGLE_SECSEL,SEL_COMMAND), NULL);
              return 1;
            case KEY_period:
              handle(this, MKUINT(ID_CLEAR_SECSEL,SEL_COMMAND), NULL);
              return 1;
            case KEY_comma:
              handle(this, MKUINT(ID_CLEAR_ALL_SECSEL,SEL_COMMAND), NULL);
              return 1;
            }
          }
        else{                           // Normal key
          entry = event->text[0];
          if (!entry)
            return 0;	// unhandled F-key
          if((' '<=(FXuchar)entry)||(entry=='\t')){
     char_insert:
            if(edable){
              if(isPosSelected(getCursorPos())){
                handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),(void *)1L); // deferred update
                }
              else{
                handle(this,MKUINT(ID_DESELECT_ALL,SEL_COMMAND),NULL);
                }
              handle(this,MKUINT(ID_INSERT_CHAR,SEL_COMMAND),(void*)(long)entry);
              }
            else{
              getApp()->beep();
              }
            }
          return 1;
          }
        break;
      }
    }
  return 0;
  }
  
  
// Keyboard release
long FXPLText::onKeyRelease(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
  if(!isEnabled()) return 0;
  FXTRACE((200,"%s::onKeyRelease keysym=0x%04x state=%04x\n",getClassName(),event->code,event->state));
  if(target && target->handle(this,MKUINT(message,SEL_KEYRELEASE),ptr)) return 1;
  if(!(event->state&ALTMASK)){
    switch(event->code){
      case KEY_Shift_L:
      case KEY_Shift_R:
      case KEY_Control_L:
      case KEY_Control_R:
        if(dragmode==MOUSE_DRAG){
          handle(this,MKUINT(0,SEL_DRAGGED),ptr);
          }
        return 1;
      case KEY_Up:
      case KEY_KP_Up:
      case KEY_Down:
      case KEY_KP_Down:
      case KEY_Left:
      case KEY_KP_Left:
      case KEY_Right:
      case KEY_KP_Right:
      case KEY_Home:
      case KEY_KP_Home:
      case KEY_End:
      case KEY_KP_End:
      case KEY_Page_Up:
      case KEY_Page_Down:
      case KEY_Insert:
      case KEY_KP_Insert:
      case KEY_Delete:
      case KEY_KP_Delete:
      case KEY_BackSpace:
      case KEY_Return:
      case KEY_KP_Enter:
      case KEY_ISO_Enter:
        return 1;
      default:
        if(event->state&CONTROLMASK){   // Control key
          switch(event->code){
            case KEY_a:
            case KEY_x:
            case KEY_c:
            case KEY_v:
            case KEY_k:
            case KEY_h:
            case KEY_m:
            case KEY_l:
            case KEY_g:
            case KEY_G:
            case KEY_f:
            case KEY_e:
            case KEY_i:
            case KEY_I:
            case KEY_z:
            case KEY_Z:
            case KEY_y:
            case KEY_Y:
              return 1;
            }
          }
        else{                           // Normal key
          return 1;
          }
        break;
      }
    }
  return 0;
  }


/*******************************************************************************/

FXchar * FXPLText::getStringSel(FXint & len)
{
  FXuchar * str;
  FXuint size;
  if (fxTraceLevel > 99){
    FXDragType * types; FXuint ntypes;
    FXTRACE((100,"%s:getStringSel\n",getClassName()));
    if (inquireDNDTypes(FROM_SELECTION, types, ntypes)){
      for (FXuint i = 0; i < ntypes; i++)
        FXTRACE((102, "Got type %u (%s)\n", types[i], getApp()->getDragTypeName(types[i]).text()));
      FXFREE(&types);
    }
  }

  if (offeredDNDType(FROM_SELECTION, stringType) && 
      getDNDData(FROM_SELECTION, stringType, str, size)){
    FXTRACE((101, "Selection is `%*.*s' length %u\n", (int)size, (int)size, str, size));
    len = size;
    return len ? (FXchar *)str : NULL;
    }
  return 0;
}



/*******************************************************************************/
// Find/replace dialogue
long FXPLText::onCmdFind(FXObject* sender,FXSelector,void* ptr){
  FXchar * str;
  FXint len;
  FXTRACE((100,"%s:onCmdFind\n",getClassName()));
  if (!frd)
    frd = new FXFindReplaceDialog(this, textsrc);
  else
    frd->setTextSrc(textsrc);	// Point to our text
  if (!frd->id())
    frd->create();
  if ((str = getStringSel(len))){
    frd->setFindText((FXuchar *)str, len); 
    FXFREE(&str);
  }
  frd->show();
  return 1;
  }

void FXPLText::setFindReplaceDialog(FXFindReplaceDialog * dlg) {
  if (frd_private && frd)
    delete frd;		// Get rid of any private one we may have constructed
  frd_private = FALSE;
  frd = dlg;
  frd->setTextSrc(textsrc);
  }


FXint FXPLText::startSearchPos(FXbool fwd, FXbool restart) const {
  if (restart)
    return fwd ? 0 : textsrc->length()-1;
  if (hasSelection())
    return fwd ? getSelEndPos() : getSelStartPos()-1;
  else
    return getCursorPos() - (fwd ? 0 : 1); 
  }

FXint FXPLText::doSearchResults(FXint pos, FXint len) {
  if (pos >= 0) {
    setAnchorPos(pos);
    extendSelection(pos+len);
    // In append or prepend modes, just make the position visible.  Otherwise, move
    // the cursor (which makes visible).
    if (options & PLTEXT_PREPEND)
      makePositionVisible(pos);
    else
      if (getCursorPos() != pos) handle(this, MKUINT(0,SEL_CHANGED), (void*)pos);
    return 1;
    }
  else getApp()->beep();
  return 0;
  }

// Find next/previous matching current selection.  Selects the matching text.
// This is not used for the dialogue buttons, only for the hotkeys.  If nothing
// selected, uses last text selected in the text source.
long FXPLText::onCmdFindNext(FXObject* sender,FXSelector sel,void* ptr){
  FXchar * str;	// String to find
  FXint len;	// Length of above
  FXint sp;	// Starting position
  FXint fp;	// Position of next occurrence
  FXFindData fd;// Constructed search parameters
  FXbool freestr = TRUE;
  
  FXTRACE((100,"%s:onCmdFindNext\n",getClassName()));
  if ((str = getStringSel(len)) || (freestr = FALSE, str = textsrc->lastSelection(len))) {
    fd.find = str;
    fd.flen = len;
    fd.search_type = FXFindData::FD_LITERAL;
    fd.case_sens = TRUE;
    fd.fwd = SELID(sel) == ID_FINDNEXT;
    fd.restart = FALSE;
    fd.whole_word = FALSE;
    fd.multi_line = TRUE;
    
    sp = startSearchPos(fd.fwd, fd.restart);
    fp = textsrc->find(&fd, sp, sp, len, this);
    if (freestr)
      FXFREE(&str);
    doSearchResults(fp, len);
    return 1;
    }
  return 0;
  }

// Find next/previous string.  ptr points to FXFindData struct.
long FXPLText::onCmdFindStr(FXObject* sender,FXSelector sel,void* ptr){
  FXFindData * fd = (FXFindData *)ptr;
  FXint len;	// Length of found string
  FXint sp;	// Starting position
  FXint fp;	// Position of next occurrence
  FXTRACE((100,"%s:onCmdFindStr\n",getClassName()));

  sp = startSearchPos(fd->fwd, fd->restart);
  fp = textsrc->find(fd, sp, sp, len, this);
  return doSearchResults(fp, len);
  }

// Find all matching strings.  ptr points to FXFindData struct.
long FXPLText::onCmdFindAll(FXObject* sender,FXSelector sel,void* ptr){
  FXFindData * fd = (FXFindData *)ptr;
  FXint sp, ip;
  FXint fp;
  FXint len;
  FXint found = 0;
  FXint i;
  FXbool wrap = FALSE;
  FXTextAnchor * xta;
  
  fd->fwd = TRUE;	// Force forward search
  fd->more = FALSE;
  sp = ip = startSearchPos(fd->fwd, fd->restart);
  for (i = 0; i <= fd->limit; i++) {
    //printf("FindAll: sp=%d, end=%d\n", sp, ip);
    fp = textsrc->find(fd, sp, ip, len, this);
    //printf("  found at %d, len=%d\n", fp, len);
    wrap |= fp < sp;
    if (fp >= 0) {
      if (wrap && fp >= ip) break;	// wrapped around
      if (i == fd->limit) {
        // At least one more found.  Set flag in fd.
        fd->more = TRUE;
        break;
        }
      sp = fp + len;
      if (!found && !fd->extend)
        textsrc->removeMatchingTextAnchors(FXPFATTR_SECONDARY, FXPFATTR_SECONDARY);
      if (!fd->extend || !textsrc->findTextAnchors(fp, sp, 1, &xta, FXPFATTR_SECONDARY, FXPFATTR_SECONDARY)) {
        //printf("  setting sec sel %d - %d\n", fp, sp);
        setSecSel(fp, len);
        }
      found++;
      }
    else break;
    }
  if (!found)
    getApp()->beep();
  else {
    //printf("Found %d\n", found);
    if (fd->move) {
      setCursorPos(sp);
      makePositionVisible(sp);
      }
    }
  return 1;
  }

long FXPLText::onCmdReplStr(FXObject* sender,FXSelector sel,void* ptr){
  FXFindData * fd = (FXFindData *)ptr;
  FXint sp;
  FXint oldlen;
  if (hasSelection()) {
    sp = getSelStartPos();
    oldlen = getSelEndPos() - sp;
    if (oldlen) {
      replaceText(sp, oldlen, fd->replace, fd->rlen);
      //setModified();
      return 1;
      }
    else getApp()->beep();
    }
  else getApp()->beep();
  return 0;
  }

long FXPLText::onCmdReplAllSel(FXObject* sender,FXSelector sel,void* ptr){
  FXFindData * fd = (FXFindData *)ptr;
  FXint nta = textsrc->getNumTextAnchors();
  FXint i, sp;
  FXint replaced = 0;
  FXTextAnchor ** tas;
  FXTextAnchor * ta;
  
  FXMALLOC(&tas, FXTextAnchor *, nta);
  nta = textsrc->findTextAnchors(-1, -1, nta, tas, FXPFATTR_SECONDARY, FXPFATTR_SECONDARY);
  for (i = 0; i < nta; i++) {
    ta = tas[i];
    textsrc->replace(ta->getPos(), sp=ta->getEpos(), fd->replace, fd->rlen, TRUE, this);
    ok = textsrc->isOk() && pager->isOk();
    if (!ok) {
      fxwarning("%s::onCmdReplAllSel: error in text replacement\n",getClassName());
      FXFREE(&tas);
      return 1;	// Caller needs to check isOk()
      }
    replaced++;
    }
  FXFREE(&tas);
  if (!replaced)
    getApp()->beep();
  else {
    // A dummy replace to end the update transaction.
    textsrc->replace(0, 0, 0, 0, FALSE, this);
    ok = textsrc->isOk() && pager->isOk();
    if (!ok) {
      fxwarning("%s::onCmdReplAllSel: error in text replacement\n",getClassName());
      return 1;	// Caller needs to check isOk()
      }
    if (fd->move) {
      setCursorPos(sp);
      makePositionVisible(sp);
      }
    }
  return 1;
  }
  
long FXPLText::onCmdReplAll(FXObject* sender,FXSelector sel,void* ptr){
  /* Replace all occurrences.  Does not require selection.  Replacements are done in
     chunks of up to 64k, however if a gap of more than 16k occurs between successive
     replacements then the replacements before the gap are actioned before continuing.
     This results in a tradeoff between buffer usage and CPU time.  Dense changes are
     done with a maximum of <filesize>/64k calls to replace(), and sparse changes a max
     of <filesize>/16k.  This is much more efficient than simply calling replace() for
     each find (even with deferred graphics).
     
     FIXME: the above u-beaut scheme needs to be implemented.
  */
  FXFindData * fd = (FXFindData *)ptr;
  FXint sp, ip;
  FXint fp;
  FXint len;
  FXint replaced = 0;
  FXint i;
  FXbool wrap = FALSE;
    
  fd->fwd = TRUE;	// Force forward search
  fd->more = FALSE;
  sp = ip = startSearchPos(fd->fwd, fd->restart);
  for (i = 0; i <= fd->limit; i++) {
    fp = textsrc->find(fd, sp, ip, len, this);
    wrap |= fp < sp;
    if (fp >= 0) {
      if (wrap && fp >= ip) break;	// wrapped around
      if (i == fd->limit) {
        // At least one more found.  Set flag in fd.
        fd->more = TRUE;
        break;
        }
      // Deferred graphic update...
      textsrc->replace(fp, fp+len, fd->replace, fd->rlen, TRUE, this);
      ok = textsrc->isOk() && pager->isOk();
      if (!ok) {
        fxwarning("%s::onCmdReplAll: error in text replacement\n",getClassName());
        return 1;	// Caller needs to check isOk()
        }
      if (wrap) ip += fd->rlen - len;
      sp = fp + fd->rlen;
      replaced++;
      }
    else break;
    }
  if (!replaced)
    getApp()->beep();
  else {
    // A dummy replace to end the update transaction.
    textsrc->replace(0, 0, 0, 0, FALSE, this);
    ok = textsrc->isOk() && pager->isOk();
    if (!ok) {
      fxwarning("%s::onCmdReplAll: error in text replacement\n",getClassName());
      return 1;	// Caller needs to check isOk()
      }
    if (fd->move) {
      setCursorPos(sp);
      makePositionVisible(sp);
      }
    }
  return 1;
  }


long FXPLText::onCmdGotoSelLine(FXObject* sender,FXSelector sel,void* ptr) {
  // Move cursor to highlit line number.  We allow the target line number to be one more
  // than the actual number of lines, since some brain-dead compilers etc. print bad line
  // numbers if they reach EOF unexpectedly.  Leading characters in the selection are skipped
  // until a digit is found.
  FXint len;
  FXchar buf[20];
  FXchar * str = getStringSel(len);
  FXchar * s;
  if (!str)
    return 1;
  s = str;
  while (len && !isdigit(*s)) s++, len--;
  memcpy(buf, s, FXMIN(19, len));
  FXFREE(&str);
  buf[19] = 0;
  FXint line;
  sscanf(buf, "%d", &line);
  if (line >= 1 && line <= (FXint)textsrc->lines()+1)
    moveCursorToLine(line-1);
  else
    getApp()->beep();
  return 1;
  }

long FXPLText::onCmdIndent(FXObject* sender,FXSelector sel,void* ptr) {
  if (!hasSelection() || !isEditable())
    return 1;
  FXint indent = (FXint)ptr;  // +ve for indent, -ve for outdent.
  if (!indent)
    return 1;
  FXint ssp = getSelStartPos();
  FXint sep = getSelEndPos();
  FXint slp = lineStart(ssp);
  if (slp < ssp)
    // Selection start is not start of line.  Do not touch first partial line.
    slp = nextLine(ssp);
  if (slp >= sep)
    // Selection too small to indent
    return 1;
  // At least one line to indent.  We don't touch lines which are only whitespace.
  FXint lines = getLineOfPos(sep-1) - getLineOfPos(slp) + 1;
  FXchar * str;
  FXchar * p, * q;
  FXint i, ls, le, len;
  FXint oldlen = sep-slp;
  ls = slp;
  if (indent > 0) {
    FXMALLOC(&str, FXchar, oldlen+lines*indent);
    if (!str)
      return 1;
    p = str;
    for (i = 0; i < lines; i++) {
      le = nextLine(ls);
      if (le > sep) le = sep;
      len = le-ls;
      extractText(p, ls, len);
      // We add spaces after the initial line whitespace, otherwise tabs may cause no
      // apparent indentation.
      for (q = p; q < p+len && isspace(*q); q++);
      if (q < p+len) {
        memmove(q+indent, q, p+len-q);
        memset(q, ' ', indent);
        p += indent;
        }
      p += len;
      ls = le;
      }
    FXASSERT(p <= str+oldlen+lines*indent);
    }
  else {
    FXMALLOC(&str, FXchar, oldlen);
    if (!str)
      return 1;
    p = str;
    for (i = 0; i < lines; i++) {
      le = nextLine(ls);
      if (le > sep) le = sep;
      len = le-ls;
      extractText(p, ls, len);
      // We only delete spaces at the end of initial whitespace.  If the last whitespace is
      // a tab, we don't remove it, which will leave a ragged appearance.  This needs to be
      // done better (i.e. by the page formatter since it knows about tab appearance; we don't).
      for (q = p; q < p+len && isspace(*q); q++);
      if (q < p+len) {
        FXchar * r = q;
        FXint ind = -indent;
        while (q > p && ind-- && *--q == ' ');
        if (isspace(*q) && *q != ' ') q++;
        memmove(q, r, p+len-r);
        p -= r-q;
        }
      p += len;
      ls = le;
      }
    FXASSERT(p <= str+oldlen);
    }
  replaceText(slp, oldlen, str, p-str);
  FXFREE(&str);
  return 1; 
  }


long FXPLText::onCmdRoll(FXObject* sender,FXSelector sel,void* ptr) {
  FXbool redo = SELID(sel) == ID_ROLLFORWARD;
  FXint num = (FXint)ptr;
  if (redo)
    textsrc->rollforward(num);
  else
    textsrc->rollback(num);
  return 1;
  }

long FXPLText::onCmdCursorToRoll(FXObject* sender,FXSelector sel,void* ptr) {
  FXbool redo = SELID(sel) == ID_CURSOR_TO_ROLLFORWARD;
  FXint newpos;
  if (redo)
    newpos = textsrc->rollforwardPos();
  else
    newpos = textsrc->rollbackPos();
  if (newpos >= 0)
    handle(this, MKUINT(0,SEL_CHANGED),(void*)newpos);
  return 1;
  }

/*******************************************************************************/

// Move cursor to top of buffer
long FXPLText::onCmdCursorTop(FXObject*,FXSelector,void*){
  FXint pos = (options & PLTEXT_EDITMASK) == PLTEXT_APPEND ?
                textsrc->getPosOfLine(rolines) :
                0;
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
  }


// Move cursor to bottom of buffer
long FXPLText::onCmdCursorBottom(FXObject*,FXSelector,void*){
  FXint pos = (options & PLTEXT_EDITMASK) == PLTEXT_PREPEND ?
                textsrc->getPosOfLine(textsrc->lines() - rolines) - 1 :
                getLength();
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)pos);
  }


// Move cursor to begin of line or row
long FXPLText::onCmdCursorHome(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),
    options & PLTEXT_SEL_BY_ROW ? 
      (void*)rowStart(getCursorPos()) :
      (void*)lineStart(getCursorPos())
    );
  }


// Move cursor to end of line or row
long FXPLText::onCmdCursorEnd(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),
    options & PLTEXT_SEL_BY_ROW ? 
      (void*)rowEnd(getCursorPos()) :
      (void*)lineEnd(getCursorPos())
    );
  }


// Move cursor right
long FXPLText::onCmdCursorRight(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)(getCursorPos()+1));
  }


// Move cursor left
long FXPLText::onCmdCursorLeft(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)(getCursorPos()-1));
  }


// Move cursor to previous row
long FXPLText::onCmdCursorUp(FXObject*,FXSelector,void*){
  ungrab();
  setPrefcol();                                                                                           
  moveCursorToRow(getCursorRow()-1);
  return 1;
  }


// Move cursor to next row
long FXPLText::onCmdCursorDown(FXObject*,FXSelector,void*){
  setPrefcol();
  moveCursorToRow(getCursorRow()+1);
  return 1;
  }


// Page down
long FXPLText::onCmdCursorPageDown(FXObject*,FXSelector,void*){
  setPrefcol();
  moveCursorToRow(getCursorRow()+nvisrows);
  return 1;
  }


// Page up
long FXPLText::onCmdCursorPageUp(FXObject*,FXSelector,void*){
  setPrefcol();
  moveCursorToRow(getCursorRow()-nvisrows);
  return 1;
  }


// Word Left
long FXPLText::onCmdCursorWordLeft(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)leftWord(getCursorPos()));
  }


// Word Right
long FXPLText::onCmdCursorWordRight(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)rightWord(getCursorPos()));
  }


// Cursor pos to top of screen
long FXPLText::onCmdCursorScreenTop(FXObject*,FXSelector,void*){
  setTopLine(getCursorPos());
  return 1;
  }


// Cursor pos to bottom of screen
long FXPLText::onCmdCursorScreenBottom(FXObject*,FXSelector,void*){
  setBottomLine(getCursorPos());
  return 1;
  }


// Cursor pos to center of screen
long FXPLText::onCmdCursorScreenCenter(FXObject*,FXSelector,void*){
  setCenterLine(getCursorPos());
  return 1;
  }


// Move cursor to begin of paragraph
long FXPLText::onCmdCursorParHome(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)lineStart(getCursorPos()));
  }


// Move cursor to end of paragraph
long FXPLText::onCmdCursorParEnd(FXObject*,FXSelector,void*){
  return handle(this,MKUINT(0,SEL_CHANGED),(void*)lineEnd(getCursorPos()));
  }


// Mark
long FXPLText::onCmdMark(FXObject*,FXSelector,void*){
  setAnchorPos(getCursorPos());
  return 1;
  }


// Extend
long FXPLText::onCmdExtend(FXObject*,FXSelector,void*){
  extendSelection(getCursorPos());
  return 1;
  }


// Insert a character
long FXPLText::onCmdInsertChar(FXObject*,FXSelector,void* ptr){
  FXchar ch=(FXchar)(long)ptr;
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=getCursorPos();
  what[1]=1;
  // Changed overstrike to not obliterate newlines.
  if(options&PLTEXT_OVERSTRIKE && getChar(what[0]) != '\n'){
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    replaceText(what[0],1,&ch,1);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  else{
    insertText(what[0],&ch,1);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  //handle(this,MKUINT(0,SEL_CHANGED),(void*)getCursorPos());
  makePositionVisible(getCursorPos());
  return 1;
  }


// Insert a string
long FXPLText::onCmdInsertString(FXObject*,FXSelector,void* ptr){
  FXchar* string=(FXchar*)ptr;
  FXint what[2];
  if(!isEditable()) return 1;
  FXASSERT(string);
  what[0]=getCursorPos();
  what[1]=strlen(string);
  if(options&PLTEXT_OVERSTRIKE){
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    replaceText(what[0],what[1],string,what[1]);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  else{
    insertText(what[0],string,what[1]);
    handle(this,MKUINT(0,SEL_INSERTED),(void*)what);
    }
  //handle(this,MKUINT(0,SEL_CHANGED),(void*)getCursorPos());
  return 1;
  }


// Insert a character
long FXPLText::onCmdInsertNewline(FXObject*o,FXSelector sel,void * ptr){
  FXint what[2];
  FXbool doAutoIndent = (FXint)ptr && options & PLTEXT_AUTO_INDENT;
  FXint ls;  // Start pos of current line
  FXint len; // Number of space chars at start of line
  FXchar * buf; // Insertion buffer
  FXint p = getCursorPos();

  if(!isEditable()) return 1;
  ls = lineStart(p);
  len = p - ls;
  if (doAutoIndent && len) {
    FXMALLOC(&buf, FXchar, len+1);
    if (!buf)
      return 1;
    extractText(buf+1, ls, len);
    buf[0] = '\n';
    for (ls = 1; ls < len+1; ls++)
      if (!isspace(buf[ls]))
        break;
    len = ls;
    }
  else len = 1, buf = "\n";
  what[0] = p;
  what[1] = len;
  insertText(what[0], buf, len);
  handle(this, MKUINT(0,SEL_INSERTED), (void*)what);
  //handle(this, MKUINT(0,SEL_CHANGED), (void*)getCursorPos());
  makePositionVisible(getCursorPos());
  if (len > 1) FXFREE(&buf);
  return 1;
  }


// Cut 
long FXPLText::onCmdCutSel(FXObject*,FXSelector,void*){
  FXDragType types[1];
  if(hasSelection()){
    if(isEditable()){
      types[0]=stringType;
      if(acquireClipboard(types,1)){
        FXFREE(&clipbuffer);
        FXASSERT(getSelStartPos()<=getSelEndPos());
        cliplength=getSelEndPos()-getSelStartPos();
        FXCALLOC(&clipbuffer,FXchar,cliplength+1);
        if(!clipbuffer){ 
          fxwarning("%s::onCmdCutSel: out of memory\n",getClassName()); 
          cliplength=0;
          }
        else{
          extractText(clipbuffer,getSelStartPos(),cliplength);
          handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),NULL);
          }
        }
      }
    else{
      getApp()->beep();
      }
    }
  return 1;
  }


// Copy
long FXPLText::onCmdCopySel(FXObject*,FXSelector,void*){
  FXDragType types[1];
  if(hasSelection()){
    types[0]=stringType;
    if(acquireClipboard(types,1)){
      FXFREE(&clipbuffer);
      FXASSERT(getSelStartPos()<=getSelEndPos());
      cliplength=getSelEndPos()-getSelStartPos();
      FXCALLOC(&clipbuffer,FXchar,cliplength+1);
      if(!clipbuffer){ 
        fxwarning("%s::onCmdCopySel: out of memory\n",getClassName()); 
        cliplength=0;
        }
      else{
        extractText(clipbuffer,getSelStartPos(),cliplength);
        }
      }
    }
  return 1;
  }


// Paste
long FXPLText::onCmdPasteSel(FXObject*,FXSelector,void*){
  FXuchar *data;
  FXuint   size;
  if(isEditable()){
    if(getDNDData(FROM_CLIPBOARD,stringType,data,size)){
      if (FXRESIZE(&data, FXchar, size+1)) { 
        data[size]='\0'; 
#ifdef WIN32
        FXuint i,j;
        FXuchar c;
        for(i=j=0; j<size; j++){if((c=data[j])!='\r'){data[i++]=c;}}
        if(i<size) data[i]='\0';
#endif
        if(hasSelection())
          handle(this,MKUINT(ID_DELETE_SEL,SEL_COMMAND),(void *)1); // deferred update
        handle(this,MKUINT(ID_INSERT_STRING,SEL_COMMAND),(void*)data);
      }
      FXFREE(&data);
      return 1;
      }
    }
  else{
    getApp()->beep();
    }
  return 1;
  }


// Delete selection.  Parameter is flag indicating whether to defer the screen update - this
// would be set if the deletion is to be immediately followed by an insertion.
long FXPLText::onCmdDeleteSel(FXObject*,FXSelector,void*ptr){
  FXint what[2];
  if(!hasSelection()) return 1;
  FXbool defer = (FXint)ptr & 0x01;
  what[0]=getSelStartPos();
  what[1]=getSelEndPos()-getSelStartPos();
  if (isDeletable(what)) {
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    textsrc->replace(what[0],what[0]+what[1], 0, 0, defer, this);
    ok = textsrc->isOk() && pager->isOk();
    if (!ok)
      fxwarning("%s::onCmdDeleteSel: error in text replacement\n",getClassName());
    }
  killSelection();
  return 1;
  }


// Select character
long FXPLText::onCmdSelectChar(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(getCursorPos());
  extendSelection(getCursorPos()+1,PLTEXT_SELECT_CHARS);
  what[0]=getSelStartPos();
  what[1]=1;
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Select Word
long FXPLText::onCmdSelectWord(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(getCursorPos());
  extendSelection(getCursorPos(),PLTEXT_SELECT_WORDS);
  what[0]=getSelStartPos();
  what[1]=getSelEndPos()-what[0];
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Select Line
long FXPLText::onCmdSelectLine(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(getCursorPos());
  extendSelection(getCursorPos(),PLTEXT_SELECT_LINES);
  what[0]=getSelStartPos();
  what[1]=getSelEndPos()-what[0];
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Select All
long FXPLText::onCmdSelectAll(FXObject*,FXSelector,void*){
  FXint what[2];
  setAnchorPos(0);
  extendSelection(getLength(),PLTEXT_SELECT_CHARS);
  what[0]=getSelStartPos();
  what[1]=getSelEndPos()-what[0];
  handle(this,MKUINT(0,SEL_SELECTED),(void*)what);
  return 1;
  }


// Deselect All
long FXPLText::onCmdDeselectAll(FXObject*,FXSelector,void*){
  FXint what[2];
  what[0]=getSelStartPos();
  what[1]=getSelEndPos()-what[0];
  handle(this,MKUINT(0,SEL_DESELECTED),(void*)what);
  killSelection();
  return 1;
  }

// Test for deletability of range of text.  If partially deletable, alters *what
// appropriately and returns true.  If completely undeletable, returns false.
FXbool FXPLText::isDeletable(FXint what[2]) const {
  FXint bpos;
  switch (options & PLTEXT_EDITMASK) {
    case PLTEXT_APPEND:
      bpos = textsrc->getPosOfLine(rolines);	// 1st replaceable
      if (what[0]+what[1] < bpos) return FALSE;
      if (what[0] < bpos) {
        what[1] -= bpos - what[0];
        what[0] = bpos;
        }
      break;
    case PLTEXT_PREPEND:
      // 1st irreplaceable (-1 to preserve boundary newline)
      bpos = textsrc->getPosOfLine(textsrc->lines() - rolines) - 1;
      if (what[0] >= bpos) return FALSE;
      if (what[0]+what[1] > bpos)
        what[1] = bpos - what[0];
      break;
    }
  return TRUE;
  }


// Backspace character
long FXPLText::onCmdBackspace(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  if(getCursorPos()==0){ getApp()->beep(); return 1; }
  what[0]=getCursorPos()-1;
  what[1]=1;
  if (isDeletable(what)) {
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    if (options&PLTEXT_OVERSTRIKE)
      blankText(what[0],what[1]), setCursorPos(what[0]);
    else
      removeText(what[0],what[1]);
    }
  makePositionVisible(getCursorPos());
  return 1;
  }


// Backspace word
long FXPLText::onCmdBackspaceWord(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=leftWord(getCursorPos());
  what[1]=getCursorPos()-what[0];
  if (isDeletable(what)) {
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    if (options&PLTEXT_OVERSTRIKE)
      blankText(what[0],what[1]), setCursorPos(what[0]);
    else
      removeText(what[0],what[1]);
    }
  makePositionVisible(getCursorPos());
  return 1;
  }


// Backspace bol
long FXPLText::onCmdBackspaceBol(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=rowStart(getCursorPos());
  what[1]=getCursorPos()-what[0];
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  if (options&PLTEXT_OVERSTRIKE)
    blankText(what[0],what[1]), setCursorPos(what[0]);
  else
    removeText(what[0],what[1]);
  makePositionVisible(getCursorPos());
  return 1;
  }


// Delete character
long FXPLText::onCmdDelete(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  if(getCursorPos()==getLength()){ getApp()->beep(); return 1; }
  what[0]=getCursorPos();
  what[1]=1;
  if (isDeletable(what)) {
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    removeText(what[0],what[1]);
    }
  makePositionVisible(getCursorPos());
  return 1;
  }


// Delete word
long FXPLText::onCmdDeleteWord(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=getCursorPos();
  what[1]=rightWord(getCursorPos())-what[0];
  if (isDeletable(what)) {
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    removeText(what[0],what[1]);
    }
  makePositionVisible(getCursorPos());
  return 1;
  }


// Delete end of line
long FXPLText::onCmdDeleteEol(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=getCursorPos();
  what[1]=rowEnd(getCursorPos())-what[0];
  handle(this,MKUINT(0,SEL_DELETED),(void*)what);
  removeText(what[0],what[1]);
  return 1;
  makePositionVisible(getCursorPos());
  }


// Delete line
long FXPLText::onCmdDeleteLine(FXObject*,FXSelector,void*){
  FXint what[2];
  if(!isEditable()) return 1;
  what[0]=rowStart(getCursorPos());
  what[1]=nextRow(getCursorPos())-what[0];
  if (options & PLTEXT_PREPEND && textsrc->getChar(what[0]+what[1]) == '\n')
    what[1]--;
  if (what[1] > 0) {
    handle(this,MKUINT(0,SEL_DELETED),(void*)what);
    removeText(what[0],what[1]);
    }
  makePositionVisible(getCursorPos());
  return 1;
  }

/*******************************************************************************/

// Editable toggle
long FXPLText::onCmdToggleEscape(FXObject*,FXSelector,void*){
  setEscape(!(options & PLTEXT_ESCAPE));
  return 1;
  }

// Update escape mode
long FXPLText::onUpdToggleEscape(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&PLTEXT_ESCAPE) ? ID_UNCHECK : ID_CHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }

// Editable toggle
long FXPLText::onCmdToggleEditable(FXObject*,FXSelector,void*){
  // This is only for the widget.  Also need to set the text source editable.
  setEditable((options & PLTEXT_EDITMASK) == PLTEXT_READONLY);
  return 1;
  }


// Update editable toggle
long FXPLText::onUpdToggleEditable(FXObject* sender,FXSelector,void*){
  FXuint msg=((options&PLTEXT_EDITMASK) == PLTEXT_READONLY) ? ID_UNCHECK : ID_CHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }


// Overstrike toggle
long FXPLText::onCmdToggleOverstrike(FXObject*,FXSelector,void*){
  options^=PLTEXT_OVERSTRIKE;
  if (cur_cstyle == 0 && (options & PLTEXT_OVERSTRIKE))
    changeCursor(1);
  else if (cur_cstyle == 1 && !(options & PLTEXT_OVERSTRIKE))
    changeCursor(0);
  return 1;
  }


// Update overstrike toggle
long FXPLText::onUpdToggleOverstrike(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&PLTEXT_OVERSTRIKE) ? ID_CHECK : ID_UNCHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }

// Cursor unconditional flash toggle
long FXPLText::onCmdToggleFlash(FXObject*,FXSelector,void*){
  options^=PLTEXT_FLASH_CURSOR;
  changeCursor(cur_cstyle, cstyles[cur_cstyle].flash);
  return 1;
  }


// Update unconditional flash toggle
long FXPLText::onUpdToggleFlash(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&PLTEXT_FLASH_CURSOR) ? ID_CHECK : ID_UNCHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }

// Set preferred column to current cursor apparent column.  This implements a 3-state machine.
// State 0 if no preferred column, 1 when a row change is requested, 2 when this is actioned.
// When actioned (in onChange()), the state is altered ready for next movement.
void FXPLText::setPrefcol() {
  if (!setpref)
    getRowOfPos(getCursorPos(), prefcol, TRUE);
  setpref = 1; 
  }

// Move cursor to specified row, maintaining the same real or apparent column
void FXPLText::moveCursorToRow(FXint row, FXbool apparent /*= TRUE*/ ) {
  FXint csrcol, csrrow, col, newpos;
  if (row < 0)
    row = 0;
  if(row!=(csrrow = getRowOfPos(getCursorPos(), csrcol, apparent))) {
      col=csrcol;
      if (setpref)
        col = prefcol;
      FXuint a = cursor->getAttrs();
      if (col)
        a |= FXTextAnchor::ANCHOR_S_ATEND;
      else
        a &= ~FXTextAnchor::ANCHOR_S_ATEND;
      cursor->setAttrs(a);
      newpos=pager->rowPos((FXuint)row,(FXuint)col, FALSE); // Don't include newline char
      handle(this, MKUINT(0,SEL_CHANGED), (void*)newpos);
      }
  }

// Move cursor to specified line, maintaining the same offset
void FXPLText::moveCursorToLine(FXint line) {
  FXint csrcol;
  FXuint csrrow;
  FXint col, newpos, le;
  if (line < 0)
    line = 0;
  if((FXuint)line!=(csrrow = getLineOfPos(getCursorPos(), csrcol))) {
      col=csrcol;
      newpos = textsrc->getPosOfLine((FXuint)line);
      le = lineEnd(newpos);
      if (newpos + col < le)
        newpos += col;
      else
        newpos = le;
      handle(this, MKUINT(0,SEL_CHANGED), (void*)newpos);
      }
  }

// Move cursor to indicated row
long FXPLText::onCmdCursorRow(FXObject* sender,FXSelector,void*){
  FXint row;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&row)){
    if(row<1) row=1;
    if(row>(FXint)pager->numRows()) row=pager->numRows();
    --row;
    moveCursorToRow(row);
    }
  return 1;
  }

// Move cursor to indicated line
long FXPLText::onCmdCursorLine(FXObject* sender,FXSelector,void*){
  FXint line;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&line)){
    if(line<1) line=1;
    if(line>(FXint)textsrc->lines()) line=textsrc->lines();
    --line;
    moveCursorToLine(line);
    }
  return 1;
  }


// Being asked about current row number
long FXPLText::onUpdCursorRow(FXObject* sender,FXSelector,void*){
  FXint row=getCursorRow()+1;
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&row);
  return 1;
  }

// Being asked about current line number
long FXPLText::onUpdCursorLine(FXObject* sender,FXSelector,void*){
  FXint line=getCursorLine()+1;
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&line);
  return 1;
  }


// Move cursor to indicated column (character offset, not display)
long FXPLText::onCmdCursorColumn(FXObject* sender,FXSelector,void*){
  FXint col,newpos,rowbeg,rowend;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&col)){
    rowbeg=rowStart(getCursorPos());
    rowend=rowEnd(rowbeg);
    newpos=rowbeg+col;
    if(newpos<rowbeg) newpos=rowbeg;
    if(newpos>rowend) newpos=rowend;
    handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
    }
  return 1;  }

// Move cursor to indicated character offset within line
long FXPLText::onCmdCursorOffset(FXObject* sender,FXSelector,void*){
  FXint col,newpos,rowbeg,rowend;
  if(sender->handle(this,MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND),(void*)&col)){
    rowbeg=lineStart(getCursorPos());
    rowend=lineEnd(rowbeg);
    newpos=rowbeg+col;
    if(newpos<rowbeg) newpos=rowbeg;
    if(newpos>rowend) newpos=rowend;
    handle(this,MKUINT(0,SEL_CHANGED),(void*)newpos);
    }
  return 1;  }


// Being asked about current column
long FXPLText::onUpdCursorColumn(FXObject* sender,FXSelector,void*){
  FXint csrcol;
  getRowOfPos(getCursorPos(), csrcol, FALSE); 
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&csrcol);
  return 1;
  }

// Being asked about current offset
long FXPLText::onUpdCursorOffset(FXObject* sender,FXSelector,void*){
  FXint csrcol;
  getLineOfPos(getCursorPos(), csrcol); 
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&csrcol);
  return 1;
  }

// Update somebody who works on the selection
long FXPLText::onUpdHaveSelection(FXObject* sender,FXSelector,void* ptr){
  FXuint msg=hasSelection() ? ID_ENABLE : ID_DISABLE;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
  return 1;
  }


// Tab columns
long FXPLText::onUpdTabColumns(FXObject* sender,FXSelector,void*){
  FXint cols = pager->getTabColumns();
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&cols);
  return 1;
  }


// Change tab columns
long FXPLText::onCmdTabColumns(FXObject* sender,FXSelector,void*ptr){
  FXint cols;
  if (sender->handle(this, MKUINT(FXWindow::ID_GETINTVALUE,SEL_COMMAND), (void*)&cols)) {
    if (cols < 1) cols = 8;	// Enforce some sanity
    pager->setTabColumns(cols);
    }
  return 1;
  }


// Wrap columns
long FXPLText::onUpdWrapColumns(FXObject* sender,FXSelector,void*){
  FXint cols = pager->getCols();
  sender->handle(this,MKUINT(FXWindow::ID_SETINTVALUE,SEL_COMMAND),(void*)&cols);
  return 1;
  }


// Change wrap columns
long FXPLText::onCmdWrapColumns(FXObject* sender,FXSelector,void*){
  /* This is deprecated, and has no effect */
  return 1;
  }


long FXPLText::onUpdToggleShowtabs(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&PLTEXT_SHOW_TABS) ? ID_CHECK : ID_UNCHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }

long FXPLText::onUpdToggleAutoindent(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&PLTEXT_AUTO_INDENT) ? ID_CHECK : ID_UNCHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }

long FXPLText::onUpdToggleShowaux(FXObject* sender,FXSelector,void*){
  FXuint msg=(options&PLTEXT_SHOW_AUX) ? ID_CHECK : ID_UNCHECK;
  sender->handle(this,MKUINT(ID_SHOW,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(ID_ENABLE,SEL_COMMAND),NULL);
  sender->handle(this,MKUINT(msg,SEL_COMMAND),NULL);
  return 1;
  }

long FXPLText::onCmdToggleSecSel(FXObject* sender,FXSelector,void*){
  FXint pos = getCursorPos();
  FXint sp, ep;
  FXTextAnchor * ta;
  if (isPosSelected(pos)) {
    sp = getSelStartPos();
    ep = getSelEndPos();
    killSelection();
    setSecSel(sp, ep-sp);
    }
  else if ((ta = getPosSecSel(pos))) {
    sp = ta->getPos();
    ep = ta->getEpos();
    delete ta;
    setSelection(sp, ep-sp);
    }
  else
    getApp()->beep();
  return 1;
  }

long FXPLText::onCmdClearSecSel(FXObject* sender,FXSelector,void*){
  FXint pos = getCursorPos();
  FXTextAnchor * ta;
  if ((ta = getPosSecSel(pos)))
    delete ta;
  else
    getApp()->beep();
  return 1;
  }

long FXPLText::onCmdClearAllSecSel(FXObject* sender,FXSelector,void*){
  textsrc->removeMatchingTextAnchors(FXPFATTR_SECONDARY, FXPFATTR_SECONDARY);
  return 1;
  }

long FXPLText::onCmdToggleShowtabs(FXObject* sender,FXSelector,void*){
  options ^= PLTEXT_SHOW_TABS;
  update();
  return 1;
  }

long FXPLText::onCmdToggleAutoindent(FXObject* sender,FXSelector,void*){
  options ^= PLTEXT_AUTO_INDENT;
  return 1;
  }

long FXPLText::onCmdToggleShowaux(FXObject* sender,FXSelector,void*){
  options ^= PLTEXT_SHOW_AUX;
  if (options & PLTEXT_GREEK) return 1;
  if (options & PLTEXT_SHOW_AUX)
    auxwidth = set_auxwidth;
  else
    auxwidth = 0;
  recalc();
  update();
  return 1;
  }

void FXPLText::setAuxwidth(FXint auxw) { 
  if (auxw < 0) auxw = 0;
  set_auxwidth = auxw;
  if (auxwidth) {
    auxwidth = auxw;
    recalc();
    update();
    }
  }

void FXPLText::setSepwidth(FXint sepw) { 
  if (sepw < 0) sepw = 0;
  sepwidth = sepw;
  if (auxwidth) {
    recalc();
    update();
    }
  }

/*******************************************************************************/
/* These routines handle notification of various state changes in the page format.
   This is the primary means by which the text display is changed based on alterations
   to the underlying text source.
     pagerScroll: The top row to display has changed [pager->setTopRow()]
     pagerResize: The display row/column dimension has changed [pager->resize()]
     pagerUpdate: Visible content has changed [pager->change()]
     pagerLineno: Line numbers of visible content has changed [pager->change()]
     pagerNrows:  Number of rows changed, but not visible [pager->change()]
   Note that pager->change() is not called directly by a PLText object.  It is invoked
   in response to a replace() call to the underlying TextSrc, which is how changes to
   TextSrc contents get propagated to all display (widget) objects.
*/


void FXPLText::pagerScroll(FXint dr) {
  FXEvent ev;
  FXRectangle & r = ev.rect;
  //printf("FXPLText::pagerScroll(%d)\n", dr);
  setClean();	// Forget dirty areas, since basically do complete redraw.

  // This does the main scroll of the text area not counting margins.
  scroll(0, margintop, viewport_w_pix, viewport_h_pix - margintop - marginbottom, 
         0, -charheight2pix(dr));
  // We also repaint the top and bottom margins, because anchors can overhang into the
  // margins.
  r.x = 0;
  r.w = viewport_w_pix;
  if (margintop) {
    r.y = 0;
    r.h = margintop;
    renderRect(&ev);
    }
  if (marginbottom) {
    r.y = viewport_h_pix - marginbottom;
    r.h = marginbottom;
    renderRect(&ev);
    }
  
  }

void FXPLText::pagerResize() {
  pos_y = -pager->getTopRow();
  setDirty();	// Set all dirty 
  //printf("FXPLText::pagerResize\n");
  updateDirty();
  }


void FXPLText::pagerUpdate() {
  if (!id()) return;
  pos_y = -pager->getTopRow();
  //printf("FXPLText::pagerUpdate\n");
  updateDirty();
  
  // It's important to do this after updateDirty(), since it can cause a scroll if the last
  // lines of the text were deleted.  If this were done before updateDirty(), then the
  // screen would not be updated properly.
  layout();
  if (want_visible) makePositionVisible(getCursorPos());
  }


void FXPLText::pagerLineno() {
  pos_y = -pager->getTopRow();
  layout();
  //printf("FXPLText::pagerLineno\n");
  for (FXint i = 0; i < nvisrows+partialRow; i++)
    rds[i].aux = TRUE;	// Set all aux dirty.
  anydirty = TRUE;
  if (auxwidth) {
    updateDirty();
    }
  if (want_visible) makePositionVisible(getCursorPos());
  }


void FXPLText::pagerNrows() {
  // No major repaint required, but do layout to update scrollbar and updateDirty in case
  // any anchors changed.
  layout();
  //printf("FXPLText::pagerNrows\n");
  updateDirty();
  if (want_visible) makePositionVisible(getCursorPos());

  }

long FXPLText::onTrackTop(FXObject*,FXSelector,void*) {
  if (options & PLTEXT_GREEK) {
    // We are greek window tracking position of main widget.  Move our highlighting
    // bracket to match the extent of the other display.
    //printf("greek: top=%d, bot=%d\n", track->pager->getFirstPos(), track->pager->getLastPos());
    //greekBrk->move(track->pager->getFirstPos(), track->pager->getLastPos(), FALSE);
    FXint top = getRowOfPos(track->pager->getFirstPos()) - (FXint)pager->getTopRow();
    FXint bot = getRowOfPos(track->pager->getLastPos()) - (FXint)pager->getTopRow();
    FXint maxw = charwidth2pix(pager->getCols())+auxwidth;
    FXint x1 = auxwidth + marginleft;
    FXint x2 = maxw - 1;
    FXint y1 = matrixRow2pix(top);
    FXint y2 = matrixRow2pix(bot);
    FXEvent ev;
    ev.rect.x = x1;
    ev.rect.w = 1;
    ev.rect.y = FXMIN(gy1, y1);
    ev.rect.h = FXMAX(gy2, y2) - ev.rect.y;
    renderRect(&ev);
    ev.rect.x = x2;
    renderRect(&ev);
    ev.rect.y = gy1;
    ev.rect.h = 1;
    ev.rect.x = x1;
    ev.rect.w = x2 - x1 + 1;
    renderRect(&ev);
    ev.rect.y = y1;
    renderRect(&ev);
    ev.rect.y = gy2;
    renderRect(&ev);
    ev.rect.y = y2;
    renderRect(&ev);
    gx1 = x1; gx2 = x2; gy1 = y1; gy2 = y2;
    }
  return 1;
  }


/*******************************************************************************/

void FXPLText::setClean() {
  for (FXint i = 0; i < nvisrows+partialRow; i++) {
    rds[i].aux = FALSE;
    rds[i].start = 32767;
    rds[i].end = 0;
    }
  anydirty = FALSE;
  }

void FXPLText::setDirty() {
  for (FXint i = 0; i < nvisrows+partialRow; i++) {
    rds[i].aux = TRUE;
    rds[i].start = 0;
    rds[i].end = pager->getCols();
    }
  anydirty = TRUE;
  }

void FXPLText::setRowDirty(FXuint visrow, FXbool aux, FXushort s, FXushort e) {
  if (visrow >= (FXuint)(nvisrows+partialRow))
    return;
  if (e > pager->getCols())
    e = pager->getCols();
  rds[visrow].aux = rds[visrow].aux || aux;
  rds[visrow].start = FXMIN(s, rds[visrow].start);
  rds[visrow].end = FXMAX(e, rds[visrow].end);
  anydirty = TRUE;
  }

//#define TECHNIPAINT	// Define this to redraw in technicolor (this is for debugging!)

#ifdef TECHNIPAINT
  static FXColor techniback;		// Background
  static FXColor technitext;		// Text foreground
  static FXColor techniaux;		// Aux text foreground
  static FXint rg = 255;	// Cyclic counter for new colours
  
static void nextTechnipaint() {
  techniback = FXRGB(rg, rg, rg);	// Shades of light grey
  technitext = FXRGB(255-rg, rg-92, rg-92);	// various
  techniaux = FXRGB(rg-92, rg-92, 255-rg);	// various
  rg = rg - 16;
  if (rg < 191) rg = 255;
  }
#endif

void FXPLText::updateDirty() {
  if (!anydirty)
    return;
  
  // Call repaint routine for each dirty row.  To optimise this, if we notice that a
  // contiguous group of rows is (almost) completely dirty, then we update them in a single
  // block operation.  Note that if a dirty char is on the edge of the character matrix,
  // the margin is also included in the event rectangle.
  FXEvent ev;
  FXushort cc = pager->getCols();
  FXushort sc, ec;
  FXint nvis = nvisrows+partialRow;
  //printf("FXPLText::updateDirty\n");
#ifdef TECHNIPAINT
  nextTechnipaint(); 
#endif
  // First do the aux area if any.  Complete rows are done (it's not worth doing finer
  // horizontal granularity).
  ev.rect.x = 0;
  ev.rect.w = auxwidth;
  if (auxwidth) for (FXint i = 0; i < nvis; ) {
    if (rds[i].aux) {
      ev.rect.y = charheight2pix(i) + margintop;
      ev.rect.h = 0;
      if (!i)
        ev.rect.y -= margintop, ev.rect.h += margintop;
      do {
        ev.rect.h += charHeight;
        } while (++i < nvis && rds[i].aux);
      if (i == nvis)
        ev.rect.h = height - ev.rect.y;
      renderRect(&ev);
      }
    else
      i++;
    }
  
  // Now do the main character display
  for (FXint i = 0; i < nvis; ) {
    if ((sc = rds[i].start) < (ec = rds[i].end)) {
      FXbool verydirty = sc <= 1 && ec > cc-1;  // Dirty except perhaps the last chars
      ev.rect.y = charheight2pix(i) + margintop;
      ev.rect.h = 0;
      FXushort minsc = cc;
      FXushort maxec = 0;
      if (!i)
        ev.rect.y -= margintop, ev.rect.h += margintop;
      do {
        minsc = FXMIN(sc, minsc);
        maxec = FXMAX(ec, maxec);
        ev.rect.h += charHeight;
        } while (++i < nvis && verydirty && (sc = rds[i].start) <= 1 && (ec = rds[i].end) > cc-1);
      if (i == nvis)
        ev.rect.h = height - ev.rect.y;
      ev.rect.x = auxwidth + charwidth2pix(minsc) + marginleft;
      ev.rect.w = charwidth2pix(maxec-minsc);
      if (!minsc)
        ev.rect.x -= marginleft, ev.rect.w += marginleft;
      if (maxec >= cc)
        ev.rect.w = width - ev.rect.x;
      renderRect(&ev);
      }
    else
      i++;
    }
  
  
  // Now we're clean
  setClean();
  }

/*******************************************************************************/

// Draw all renderable markers within specified rectangle.  Parameter specifies whether
// just underlays are to be drawn, or just overlays.
void FXPLText::drawAnchors(FXRectangle & rect, FXDCWindow & dc, FXbool underlay) {
  FXint x, y;
  FXuint c;
  FXint r, i;
  FXRectangle bb;
  
  for (i = 0; i < pager->getNumRenderables(); i++) {
    const FXAnchorInfo * ai;
    FXTextAnchor * ta = (ai = pager->getRenderable(i))->ta;
    if (!(ta->getAttrs() & FXTextAnchor::ANCHOR_UNDERLAY) ^ !underlay)
       continue;

    r = ai->s.r;
    c = ai->s.c; 
    
    y = matrixRow2pix(r);
    if (c >= 0) {
      x = matrixCol2pix(c);
      bb = ta->bbox(0, this, charWidth, charHeight, charBase);
      bb.x += x;
      bb.y += y;
      if (!(bb.x + bb.w < rect.x ||
          bb.x > rect.x + rect.w ||
          bb.y + bb.h < rect.y ||
          bb.y > rect.y + rect.h))
        ta->render(0, this, dc, x, y, x, y, 0, charWidth, charHeight, charBase);
      }
    // Now do aux area
    if (auxwidth && ai->s.aux >= 0) {
      bb = ta->bbox(1, this, charWidth, charHeight, charBase);
      // This is our policy for spreading aux area glyphs over the available space.  Needs work.
      x = 2 + (auxwidth/4-1)*ai->s.aux; //FIXME
      bb.x += x;
      bb.y += y;
      if (!(bb.x + bb.w < rect.x ||
            bb.x > rect.x + rect.w ||
            bb.y + bb.h < rect.y ||
            bb.y > rect.y + rect.h))
        ta->render(1, this, dc, x, y, x, y, 0, charWidth, charHeight, charBase);
      }
    }
  }

static FXColor prev_fg;

FXColor FXPLText::setFg(FXDCWindow & dc, FXattr a, FXbool force) {
  FXattr fg = fg_mask & a;
  FXColor newc;
  if (!fg)
#ifdef TECHNIPAINT
    newc = technitext;
#else
    newc = colors[PLCOLOR_TEXT_FG];
#endif
  else {
    FXint cx = PLCOLOR_TEXT_SEL_FG;
    while (!(fg & 0x01)) fg >>= 1, cx += 2;
    newc = colors[cx];
    }
  if (newc != prev_fg || force)
    dc.setForeground(prev_fg = newc);
  return newc;
  }

FXColor FXPLText::setBgToFg(FXDCWindow & dc, FXattr a, FXbool force) {
  FXattr bg = bg_mask & a;
  FXColor newc;
  if (!bg)
#ifdef TECHNIPAINT
    newc = techniback;
#else
    newc = backColor;
#endif
  else {
    FXint cx = PLCOLOR_TEXT_SEL_BG;
    while (!(bg & 0x01)) bg >>= 1, cx += 2;
    newc = colors[cx]; 
    }
  if (newc != prev_fg || force)
    dc.setForeground(prev_fg = newc);
  return newc;
  }

FXColor FXPLText::getFg(FXattr a) {
  FXattr fg = fg_mask & a;
  FXColor newc;
  if (!fg)
#ifdef TECHNIPAINT
    newc = technitext;
#else
    newc = colors[PLCOLOR_TEXT_FG];
#endif
  else {
    FXint cx = PLCOLOR_TEXT_SEL_FG;
    while (!(fg & 0x01)) fg >>= 1, cx += 2;
    newc = colors[cx];
    }
  return newc;
  }

FXColor FXPLText::getBg(FXattr a) {
  FXattr bg = bg_mask & a;
  FXColor newc;
  if (!bg)
#ifdef TECHNIPAINT
    newc = techniback;
#else
    newc = backColor;
#endif
  else {
    FXint cx = PLCOLOR_TEXT_SEL_BG;
    while (!(bg & 0x01)) bg >>= 1, cx += 2;
    newc = colors[cx]; 
    }
  return newc;
  }

/* Draw everything within the specified pixel rectangle (of the event).
   This is done in the following sequence:
     - Aux area background
     - aux area separator line
     - aux area line numbers
     - aux area underlay markers
     - aux area overlay markers
     - text area margins (if non-default initial background colour)
     - text area backgrounds
     - text area underlay markers
     - text area text and special characters
     - text area overlay markers.
   In the main text area, every position in the character and atribute matrices completely
   describes the text and background rendition, thus the redraw process can be optimised to
   look at only the sections that were marked 'dirty'.
   Text backgrounds are drawn separately, since drawImageText() doesn't necessarily draw
   enough background i.e. if the aux text is larger.
   When drawing text and text backgrounds, the clip rectangle is reduced to be within the
   margins.  Otherwise, the margins can be drawn into by anchors etc.
   When initially drawing the background of the exposed area, the colour in which to do
   this is selected from the majority of the first and last char position.  If there is no
   majority then the normal background is used.  This speeds up scrolling through large
   areas of highlit text.
 */
void FXPLText::renderRect(FXEvent * event) {
  FXDCWindow dc(this,event);
  FXRectangle & rect = event->rect;
  FXRectangle mrect = rect;	// Rectangle reduced to within text margin at bottom.
  FXint xc1, xc2, yc1, yc2, yc3;// Char co-ords in text area
  FXint x, y, rr, nr;
  FXint i, xx;
  FXColor ibg;			// Initial background colour
  FXint maxw = charwidth2pix(pager->getCols())+auxwidth;

  // Convert repaint co-ordinates to character positions in the text area.
  pix2char(rect.x - auxwidth - marginleft - 1, rect.y - margintop, xc1, yc1);
  pix2charUp(rect.x - auxwidth - marginleft + rect.w, 
             rect.y - margintop + rect.h - 1, xc2, yc2);
  
  xc2++;
  yc2++;
  if (xc1 < 0) xc1 = 0;
  if (xc2 > (FXint)pager->getCols()) xc2 = pager->getCols();
  if (yc1 < 0) yc1 = 0;
  if (yc2 > (FXint)pager->getRows()) yc2 = pager->getRows();
  //printf("renderRect() rect=(%d,%d,%d,%d) = (%d,%d,%d,%d)\n",
  //  rect.x, rect.y, rect.w, rect.h,
  //  xc1, yc1, xc2, yc2);
  nr = (FXint)pager->numRows() - (FXint)pager->getTopRow();
  yc3 = yc2;
  if (yc2 > nr) yc2 = nr;	// No rows visible after this point, because don't exist
  
  // If aux area shown, and repaint area includes it, draw it and the separator.
  if (auxwidth && rect.x < auxwidth) {
    dc.setTextFont(auxFont);
    
    // Draw separator line
    if (sepwidth > 0) {
      dc.setForeground(colors[PLCOLOR_AUX_SEPARATOR]);
      dc.fillRectangle(auxwidth-sepwidth, 0, sepwidth, viewport_h_pix);
      }
    // Erase background 
#ifdef TECHNIPAINT
    dc.setForeground(techniback);
#else
    dc.setForeground(colors[PLCOLOR_AUX_BG]);
#endif    
    dc.fillRectangle(0, 0, auxwidth-sepwidth, viewport_h_pix);
    
    // Draw line numbers
    if (options & PLTEXT_SHOW_LINE_NUMBERS) {
#ifdef TECHNIPAINT
      dc.setForeground(techniaux);
#else
      dc.setForeground(colors[PLCOLOR_AUX_LINE_NUMS]);
#endif
      for (rr = yc1; rr < yc2; rr++) {	// rr iterates row numbers relative to 1st displayed.
        FXint linerow;
        FXint line = pager->getLineOfRow(rr + pager->getTopRow(), linerow);
        line++;		// Convert to start at '1'.
        if (/*rr &&*/ linerow)// If not top row, and not the first row of line, don't display
          // Note: rr condition commented out so we can do optimised scrolling.
          continue;
        FXchar buf[20];
        sprintf(buf, "%d", line);
        y = charheight2pix(rr) + margintop + auxCharBase;
        // Back the end of the line number text away from the separator line, by amount
        // equal to left margin in main text area.
        x = auxwidth - sepwidth - marginleft - auxFont->getTextWidth(buf, strlen(buf));
        dc.drawText(x, y, buf, strlen(buf));
        }
      }
    }
  
  // Skip the rest if only aux area (except draw anchors, since they can overlap into
  // aux area, or indeed can only appear in aux area).
  if (rect.x + rect.w <= auxwidth) {
    drawAnchors(rect, dc, TRUE);	// Draw underlays
    drawAnchors(rect, dc, FALSE);	// Draw overlays
    return;
    }
 
  dc.setTextFont(font);
#ifdef TECHNIPAINT
  ibg = setBgToFg(dc, 0, TRUE);	// Set technicolor background.
#else
  // Decide on initial background fill based on 'majority colour' of first and last
  // redrawn char positions.
  FXattr maja = 0;		// Majority attribute
  if (xc1 < xc2 && yc1 < yc2) {
    maja = pager->getAttrs()[yc1*pager->getCols() + xc1];
    FXattr lasta = pager->getAttrs()[(yc2-1)*pager->getCols() + xc2-1];
    //printf(" maja=%02X lasta=%02X\n", maja, lasta);
    if (getBg(maja) != getBg(lasta))
      maja = 0;
    }
  ibg = setBgToFg(dc, maja, TRUE);	// Set colour according to majority attribute.
  
#endif
  dc.fillRectangle(auxwidth, 0, viewport_w_pix - auxwidth, viewport_h_pix);
  
  // Redraw blank margins if above majority colour was not default
  if (ibg != getBg(0)) {
    //printf("  ibg not default\n");
    setBgToFg(dc, 0);
    if (margintop && rect.y < margintop)
      dc.fillRectangle(auxwidth, 0, viewport_w_pix - auxwidth, margintop);
    if (marginleft && rect.x < auxwidth+marginleft)
      dc.fillRectangle(auxwidth, margintop, marginleft, viewport_h_pix - margintop);
    if (rect.x + rect.w > maxw)
      dc.fillRectangle(maxw, margintop, viewport_w_pix - maxw, 
                       viewport_h_pix - margintop);
    if (marginbottom && rect.y + rect.h > viewport_h_pix - marginbottom)
      dc.fillRectangle(auxwidth+marginleft, viewport_h_pix - marginbottom,
                       viewport_w_pix - auxwidth - marginleft - marginright, marginbottom);  
    }

  // Adjust text clipping rectangle if bottom margin included in redraw.
  if (rect.y + rect.h > viewport_h_pix - marginbottom)
    mrect.h = viewport_h_pix - marginbottom - mrect.y;
  // Adjust text clipping rectangle if right margin included in redraw.
  if (rect.x + rect.w > maxw)
    mrect.w = maxw - mrect.x;
    
    
  // Draw the text area backgrounds, if not default (in which case they are already
  // drawn by the big fill rectangle above).
  
  dc.setClipRectangle(mrect.x, mrect.y, mrect.w, mrect.h);
  if (xc1 < xc2) for (rr = yc1; rr < yc3; rr++) {
    x = matrixCol2pix(xc1);
    FXint len = xc2 - xc1;
    const FXchar * t = pager->getMatrix() + rr*pager->getCols() + xc1;
    const FXattr * a = pager->getAttrs() + rr*pager->getCols() + xc1;
    FXint al;
    FXattr prev = a[0];
    y = matrixRow2pix(rr);
    while (len > 0) {
      for (al = 0; al < len; al++)
        if (a[al] != prev)
          break;
      if (setBgToFg(dc, prev) != ibg)
        dc.fillRectangle(x, y-charBase, charwidth2pix(al), charHeight); 
      len -= al;
      if (len > 0) {
        prev = a[al]; 
        a += al;
        t += al; 
        x += charwidth2pix(al);
        }
      }
    }

  dc.setClipRectangle(rect.x, rect.y, rect.w, rect.h);
  drawAnchors(rect, dc, TRUE);	// Draw underlays
  
  // Draw main text
  FXbool first = TRUE;
  
  dc.setClipRectangle(mrect.x, mrect.y, mrect.w, mrect.h);
  if (xc1 < xc2) for (rr = yc1; rr < yc2; rr++) {
    x = matrixCol2pix(xc1);
    FXint len = xc2 - xc1;
    const FXchar * t = pager->getMatrix() + rr*pager->getCols() + xc1;
    const FXattr * a = pager->getAttrs() + rr*pager->getCols() + xc1;
    FXint al;
    FXattr prev = a[0];
    y = matrixRow2pix(rr);
    while (len > 0) {
      for (al = 0; al < len; al++)
        if (a[al] != prev)
          break;
      setFg(dc, prev, first);
      first = FALSE;
      if (options & PLTEXT_GREEK) {
        /* Greek rendering depends on pixel width and height:
            1x1: all except space/tab/nl represented by lit pixel.
            1x(2,3): Uppercase 2 pix high, else 1 pix, or none for space
            2x(2,3): Uppercase 2x2, Lowercase 2x1, punct 1x1.
        */
        for (i = 0, xx = x; i < al; i++, xx += charWidth) {
          FXint ww, hh;
          if (t[i] <= ' ') continue;
          if (isupper(t[i])||isdigit(t[i])) ww = charWidth, hh = 2;
          else if (islower(t[i])) ww = charWidth, hh = 1;
          else ww = 1, hh = 1;
          for (; ww; ww--)
            for (FXint uu = 0; uu < hh; uu++)
              dc.drawPoint(xx+ww, y-uu+1);
          }
        }
      else if (prev & FXPFATTR_SPECIAL) {
        for (i = 0, xx = x; i < al; i++, xx += charWidth) {
          switch (t[i]) {
            case '\n':
              if (options & PLTEXT_SHOW_TABS) {
                dc.drawLine(xx, y-charBase/2, xx+charWidth-1, y-charBase/2);
                dc.drawLine(xx+charWidth-1, y-charBase/2, xx+2, y);
                }
              break;
            case '\t':
              if (options & PLTEXT_SHOW_TABS)
                dc.drawLine(xx, y-charBase/2, xx+charWidth-1, y-charBase/2);
              break;
            case 0x0B:	// tab arrow head
              if (options & PLTEXT_SHOW_TABS) {
                dc.drawLine(xx, y-charBase/2, xx+charWidth-1, y-charBase/2);
                dc.drawLine(xx+charWidth-1, y-charBase/2, xx+charWidth-4, y-charBase/2-2);
                dc.drawLine(xx+charWidth-1, y-charBase/2, xx+charWidth-4, y-charBase/2+2);
                }
              break;
            case ' ':
              // Don't generate X traffic for spaces!
              break;
            default:
              dc.drawText(xx, y, t+i, 1);
            }
          }
        }
      else
        dc.drawText(x, y, t, al);
      len -= al;
      if (len > 0) {
        prev = a[al]; 
        a += al;
        t += al;
        x += charwidth2pix(al);
        }
      }
    }
  
  if (options & PLTEXT_GREEK && track) {
    // Draw outline of the viewable area (in the track widget) in this widget.
    dc.setForeground(colors[PLCOLOR_GREEK_WINDOW]);
    FXint top = getRowOfPos(track->pager->getFirstPos()) - (FXint)pager->getTopRow();
    FXint bot = getRowOfPos(track->pager->getLastPos()) - (FXint)pager->getTopRow();
    FXint x1 = auxwidth + marginleft;
    FXint x2 = maxw - 1;
    FXint y1 = matrixRow2pix(top);
    FXint y2 = matrixRow2pix(bot);
    dc.drawLine(x1, y1, x2, y1);
    dc.drawLine(x1, y1, x1, y2);
    dc.drawLine(x1, y2, x2, y2);
    dc.drawLine(x2, y1, x2, y2);
    }

  dc.setClipRectangle(rect.x, rect.y, rect.w, rect.h);
  drawAnchors(rect, dc, FALSE);	// Draw overlays
  }

// Draw everything within the event rectangle.
long FXPLText::onPaint(FXObject*,FXSelector,void* ptr){
  FXEvent* event=(FXEvent*)ptr;
#ifdef TECHNIPAINT
  nextTechnipaint();
#endif
  renderRect(event);
  return 1;
  }

void FXPLText::dirtyRectangle(const FXRectangle & bb, FXint r, FXuint c, FXbool in_aux) {
  FXint lr, lc, hr, hc;
  FXbool auxDirty = in_aux;
  if (bb.w == 0 || bb.h == 0)
    return;	// No pixels
  if (!id()) return;	// Not created
  // Compute extended char range which may be affected.
  pix2char(bb.x, bb.y+charBase, lc, lr);
  pix2char(bb.x+bb.w-1, bb.y+charBase+bb.h-1, hc, hr);
  if (!in_aux) {
    if ((FXint)c + lc < 0)
      lc = -c, auxDirty = TRUE;
    if ((FXint)c + hc < 0)
      hc = -c;
    }
  else {
    c = 0;
    lc = 65535;
    hc = 0;
    }
  lr += r;
  hr += r;
  while (lr <= hr) {
    if (lr >= 0 && lr < nvisrows+partialRow)
      setRowDirty(lr, auxDirty, c+lc, c+hc+1);
    lr++;
    }
  }

void FXPLText::dirtyAnchor(const FXTextAnchor * ta, FXint r, FXuint c, FXint where) {
  FXRectangle bb = ta->bbox(where, this, charWidth, charHeight, charBase);
  dirtyRectangle(bb, r, c, where == 1);
  }
  
void FXPLText::dirtyBracket(const FXTextAnchor * ta, 
                            FXint r1, FXuint c1, FXint r2, FXuint c2,
                            FXbool realstart, FXbool realend) {
  FXRectangle bb = ta->bbox(0, this, charWidth, charHeight, charBase);
  if (realstart)
    dirtyRectangle(bb, r1, c1, FALSE);
  if (realend)
    dirtyRectangle(bb, r2, c2, FALSE);
  }
  
 


/*******************************************************************************/


// Move the cursor
void FXPLText::setCursorPos(FXint pos){
  pos = validPos(pos);
  cursor->move(pos,pos,FALSE);
  }


// Set anchor position
void FXPLText::setAnchorPos(FXint pos){
  pos = validPos(pos);
  anchorpos = pos;
  textsrc->setSelection(pos, pos);
  }

// Get anchor position
FXint FXPLText::getAnchorPos() const {
  return anchorpos;
  }


// Kill the selection
FXbool FXPLText::killSelection(){
  if(hasSelection()){
    releaseSelection();
    textsrc->killSelection();
    return TRUE;
    }
  return FALSE;
  }


// Set selection
FXbool FXPLText::setSelection(FXint pos,FXint len){
  setAnchorPos(pos);
  setCursorPos(pos+len);
  extendSelection(getCursorPos(),PLTEXT_SELECT_CHARS);
  return TRUE;
  }

FXTextAnchor * FXPLText::setSecSel(FXint pos,FXint len){
  FXTextBracket * tb = new FXTextBracket(textsrc, FXPFATTR_SECONDARY, this, pos, pos+len);
  tb->setVisible();
  return tb;
  }

// Test whether have selection (or sharing it).
FXbool FXPLText::hasSelection(void) const {
  return textsrc->isSelection();
  }

// Extend selection
FXbool FXPLText::extendSelection(FXint pos,FXPLTextSelectionMode mode){
  FXbool changes=FALSE;
  FXDragType types[1];
  FXint sp,ep;
  FXbool hasSel = hasSelection();
  
  // Validate position
  pos = validPos(pos);
  
  // Did position change?
  switch(mode){
 
    // Selecting words
    case PLTEXT_SELECT_WORDS:
      if(pos<=getAnchorPos()){
        sp=wordStart(pos);
        ep=wordEnd(getAnchorPos());
        }
      else{
        sp=wordStart(getAnchorPos());
        ep=wordEnd(pos);
        }
      break;
 
    // Selecting lines
    case PLTEXT_SELECT_LINES:
      if (options & PLTEXT_SEL_BY_ROW) {
        if(pos<=getAnchorPos()){
          sp=rowStart(pos);
          ep=nextRow(getAnchorPos());
          }
        else{
          sp=rowStart(getAnchorPos());
          ep=nextRow(pos);
          }
        }
      else {
        if(pos<=getAnchorPos()){
          sp=lineStart(pos);
          ep=nextLine(getAnchorPos());
          }
        else{
          sp=lineStart(getAnchorPos());
          ep=nextLine(pos);
          }
        }
      break;
 
    // Selecting characters
    default:
      if(pos<=getAnchorPos()){
        sp=pos;
        ep=getAnchorPos();
        }
      else{
        sp=getAnchorPos();
        ep=pos;
        }
      break;
    }
  
  // Got a selection at all?
  if(sp!=ep){
    types[0]=stringType;
    if (!hasSel)
      acquireSelection(types,1);
    }
  else{
    if (hasSel)
      releaseSelection();
    }
  
  textsrc->setSelection(sp, ep);
  changes=TRUE;
  
  // Must be true
  FXASSERT(getSelEndPos()<=getLength());
  FXASSERT(getSelStartPos()<=getSelEndPos());
  return changes;
  }

/*******************************************************************************/


// Search forward [pos...len> for character c, -1 if not found
FXint FXPLText::findf(FXchar c,FXint pos) const {
  /*FIXME - deprecated*/
  return -1; 
  }


// Search backward [0...pos> for character c, -1 if not found
FXint FXPLText::findb(FXchar c,FXint pos) const {
  /*FIXME - deprecated*/
  return -1;
  }


// Change top margin
void FXPLText::setMarginTop(FXint mt){
  if(margintop!=mt){
    margintop=mt;
    recalc();
    update();
    }
  }

// Change bottom margin
void FXPLText::setMarginBottom(FXint mb){
  if(marginbottom!=mb){
    marginbottom=mb;
    recalc();
    update();
    }
  }


// Change left margin
void FXPLText::setMarginLeft(FXint ml){
  if(marginleft!=ml){
    marginleft=ml;
    recalc();
    update();
    }
  }


// Change right margin
void FXPLText::setMarginRight(FXint mr){
  if(marginright!=mr){
    marginright=mr;
    recalc();
    update();
    }
  }


// Get wrap columns
FXint FXPLText::getWrapColumns() const {
  return pager->getCols();
  }

// Get tab columns
FXint FXPLText::getTabColumns() const {
  return pager->getTabColumns();
  }


// Set tab columns
void FXPLText::setTabColumns(FXint cols) {
  pager->setTabColumns(cols);
  }

FXbool FXPLText::isModified() const {
  return textsrc->isModified();
  }

void FXPLText::setModified(FXbool mod) {
  textsrc->setModified(mod);
  }


// Set text color
void FXPLText::setTextColor(FXColor clr){
  if(clr!=colors[PLCOLOR_TEXT_FG]){
    colors[PLCOLOR_TEXT_FG]=clr;
    update();
    }
  }

// Set select background color
void FXPLText::setSelBackColor(FXColor clr){
  if(clr!=colors[PLCOLOR_TEXT_SEL_BG]){
    colors[PLCOLOR_TEXT_SEL_BG]=clr;
    update();
    }
  }


// Set selected text color
void FXPLText::setSelTextColor(FXColor clr){
  if(clr!=colors[PLCOLOR_TEXT_SEL_FG]){
    colors[PLCOLOR_TEXT_SEL_FG]=clr;
    update();
    }
  }


// Set cursor color
void FXPLText::setCursorColor(FXColor clr){
  cursor->setColor(clr);
  colors[PLCOLOR_CURSOR] = clr;
  }


// Get cursor color
FXColor FXPLText::getCursorColor() const{
  return cursor->getColor();
  }

// Set cursor style(s).  which is 0 for insert mode, 1 for overstrike mode.  Default
// style is caret with no flash.
void FXPLText::setCursorStyle(FXint which, FXuint onstyle, FXuint offstyle, FXuint ontime, FXuint offtime) {
  if (which < 0 || which >= FXPLTEXT_MAX_CURSOR_STYLES)
    return;
  cstyles[which].onstyle = onstyle;
  cstyles[which].offstyle = offstyle;
  cstyles[which].ontime = ontime;
  cstyles[which].offtime = offtime;
  if (which == cur_cstyle)
    changeCursor(which, cstyles[which].flash);
  }

// Change the current cursor to the specified style set (i.e. the which value on setCursorStyle()).
// Style sets range from 0 through to FXPLTEXT_MAX_CURSOR_STYLES-1.
void FXPLText::changeCursor(FXint which, FXbool flash) {
  if (which < 0 || which >= FXPLTEXT_MAX_CURSOR_STYLES)
    return;
  cur_cstyle = which;
  cstyles[which].flash = flash;
  if (options & PLTEXT_FLASH_CURSOR)
    flash = TRUE;
  cursor->setStyle(cstyles[which].ontime, cstyles[which].offtime, cstyles[which].onstyle, cstyles[which].offstyle);
  cursor->flash(flash);
  }



// Set general color
void FXPLText::setColor(FXint whichcolor, FXColor clr, FXbool opaque) {
  if (whichcolor < 0 || whichcolor >= PLCOLOR_LAST)
    return;
  if (whichcolor == PLCOLOR_CURSOR)
    setCursorColor(clr);
  else {
    FXbool req_upd = FALSE;
    if (clr != colors[whichcolor]) {
      colors[whichcolor] = clr;
      req_upd = TRUE;
      }
    if (whichcolor >= PLCOLOR_TEXT_SEL_BG && whichcolor <= PLCOLOR_TEXT_SPECIAL_FG) {
      FXuint x = whichcolor - PLCOLOR_TEXT_SEL_BG;
      FXattr om;
      if (x & 1) {
        om = fg_mask;
        if (opaque)
          fg_mask |= 1<<(x>>1);
        else
          fg_mask &= ~(1<<(x>>1));
        if (fg_mask != om)
          req_upd = TRUE;
        }
      else {
        om = bg_mask;
        if (opaque)
          bg_mask |= 1<<(x>>1);
        else
          bg_mask &= ~(1<<(x>>1));
        if (bg_mask != om)
          req_upd = TRUE;
        }
      }
    if (req_upd)
      update();
    }
  }


// Get general color
FXColor FXPLText::getColor(FXint whichcolor) const {
  if (whichcolor < 0 || whichcolor >= PLCOLOR_LAST)
    return FXRGB(0,0,0);
  return colors[whichcolor]; 
  }

// Set or clear background colour mask bits.  The bitmask has a bit set for each attribute
// to be set or cleared.  The previous mask contents are returned.
FXattr FXPLText::setBgMask(FXattr mask, FXbool set /*= TRUE*/ ) {
  FXattr om = bg_mask;
  if (set)
    bg_mask |= mask;
  else
    bg_mask &= ~mask;
  if (bg_mask != om)
    update();
  return om;
  }
  
// Set or clear foreground colour mask bits.  The bitmask has a bit set for each attribute
// to be set or cleared.  The previous mask contents are returned.
FXattr FXPLText::setFgMask(FXattr mask, FXbool set /*= TRUE*/ ) {
  FXattr om = fg_mask;
  if (set)
    fg_mask |= mask;
  else
    fg_mask &= ~mask;
  if (fg_mask != om)
    update();
  return om;
  }


// Change help text
void FXPLText::setHelpText(const FXString& text){
  help=text;
  }


// Change text style
void FXPLText::setTextStyle(FXuint style){
  FXuint opts = (options&~TEXT_MASK) | (style&TEXT_MASK);
  if ((options^opts)&REQ_RECALC) {
    options = opts;
    if (options & PLTEXT_SHOW_AUX)
      auxwidth = set_auxwidth;
    else
      auxwidth = 0;
    recalc();
    update();
    }
  else
    options = opts;
  }
  

// Get text style
FXuint FXPLText::getTextStyle() const {
  return (options&TEXT_MASK);
  }

FXPageFormat * FXPLText::setPageFormat(FXPageFormat * pf) {
  FXASSERT(pf);
  FXint tl;
  FXTextSrc * nts = pf->getTextSrc();	// The new text source
  FXbool diff = nts != textsrc;
  if (!diff)
    tl = pager->getFirstLine();	// Remember first line displayed
  FXPageFormat * opf = pager;
  pager->setPLText(NULL);	// Maintain invariant of 1 pager per widget.
  pager = pf;
  pager->setPLText(this);
  textsrc = nts;
  recompute();		// Resize and redisplay
  FXint dummy;
  if (!diff)
    pager->setTopRow(pager->getRowOfLine(tl, dummy));
  layout();
  update();
  return opf;
  }
  

// Return TRUE if widget is globally editable and text source is modifiable
FXbool FXPLText::isEditable() const { 
  return (options & PLTEXT_EDITMASK) != PLTEXT_READONLY && 
         textsrc->isModifiable() &&
         !(options & PLTEXT_GREEK); 
  }

// Return TRUE if widget is appendable and text source is modifiable
FXbool FXPLText::isAppendable() const { 
  return ((options&PLTEXT_EDITMASK)==PLTEXT_APPEND) && textsrc->isModifiable(); 
  }

// Return TRUE if widget is prependable and text source is modifiable
FXbool FXPLText::isPrependable() const { 
  return ((options&PLTEXT_EDITMASK)==PLTEXT_PREPEND) && textsrc->isModifiable(); 
  }

void FXPLText::setDelimiters(const FXCharset& delims) const {
  textsrc->setDelimiters(delims);
  }

// Return word delimiters  
FXCharset FXPLText::getDelimiters() const {
  return textsrc->getDelimiters();
  }


// Set widget escape state.  In escape mode, some user actions which are normally handled
// by this widget are handed over to the text source for an initial attempt at processing.
// This includes cursor movement.  The user toggles escape mode by pressing the ESC key.
// Escape mode is automatically reset after a successful action.  The old state is returned.
FXbool FXPLText::setEscape(FXbool esc){
  FXbool oldstate = !!(options & PLTEXT_ESCAPE);
  if (esc) options |= PLTEXT_ESCAPE;
  else options &= ~PLTEXT_ESCAPE;
  return oldstate;
  }


// Set widget editable or not.  This only applies to the widget: the underlying
// text source must also be editable.  This enables/disables global editing (i.e.
// not append or prepend mode).
void FXPLText::setEditable(FXbool edit){
  FXuint oldmode = options;
  options &= ~PLTEXT_EDITMASK;
  if (!edit) options |= PLTEXT_READONLY;
  if ((oldmode ^ options) & PLTEXT_EDITMASK)
    cursor->setVisible(edit);
  rolines = 0;
  }

void FXPLText::setAppendable(FXint nedit, FXbool movecsr) {
  FXbool wasRO = (options & PLTEXT_EDITMASK) == PLTEXT_READONLY;
  options &= ~PLTEXT_EDITMASK;
  options |= PLTEXT_APPEND;
  if (wasRO)
    cursor->setVisible(TRUE);
  rolines = FXMAX((FXint)textsrc->lines() - nedit, 0);
  if (movecsr) {
    FXint pos = textsrc->getPosOfLine(rolines);
    handle(this, MKUINT(0,SEL_CHANGED), (void*)pos);
    }
  }

void FXPLText::setPrependable(FXint nedit, FXbool movecsr) {
  FXbool wasRO = (options & PLTEXT_EDITMASK) == PLTEXT_READONLY;
  options &= ~PLTEXT_EDITMASK;
  options |= PLTEXT_PREPEND;
  if (wasRO)
    cursor->setVisible(TRUE);
  rolines = FXMAX((FXint)textsrc->lines() - nedit, 0);
  if (movecsr) {
    handle(this, MKUINT(0,SEL_CHANGED),(void*)0);
    }
  }


// Change all of the text
void FXPLText::setText(const FXString& text){
  setText(text.text(),text.length());
  }



// We return a copy of the buffer.  Even with the old text widget this was very
// inefficient since at one point there would be 3 copies of the data!  With the new
// text sources it is much better to make use of the direct getText() function, or
// even better to use getBuf() which doesn't need to copy anything.
// This function is deprecated, and definitely should not be used if large text sources
// are to be used.
// Most importantly, we don't check for malloc returning null.  There is no point in
// doing this, since FOX FXString type does not check either!  Another good reason not
// to use this function.
FXString FXPLText::getText() const {
  FXchar * buf;
  FXint len = getLength();
  FXMALLOC(&buf, FXchar, len+1);
  buf[len] = 0; 
  textsrc->getText(0, buf, len); // Make first copy.
  FXString s(buf);  // Make another copy
  FXFREE(&buf);	// Free the first copy
  return s;	// Make yet another copy into the return before s destructs.
  }


// Length of text.  This includes the trailing newline added by the text source.
FXint FXPLText::getLength() const {
  return textsrc->length();
}

// Offset of cursor in text.
FXint FXPLText::getCursorPos() const {
  return cursor->getPos();
}

FXuint FXPLText::getRowOfPos(FXint pos, FXint & col, FXbool apparent /*= TRUE*/ ) const {
  FXuint r;
  FXint sp;
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND); 
  r = pager->getRowOfPos(pos, sp, atend);
  if (apparent)
    col = pager->rowWidth(r, pos-sp);
  else
    col = pos-sp;
  //printf("FXPLText::getRowOfPos(%d), r=%u col=%d\n", pos, r, col);
  return r;
  }
  
FXuint FXPLText::getRowOfPos(FXint pos) const {
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND);
  return pager->getRowOfPos(pos, atend);
  }

FXuint FXPLText::getLineOfPos(FXint pos) const {
  return textsrc->getLineOfPos(pos);
  }

FXuint FXPLText::getCursorRow() const { 
  FXint dmy; 
  FXbool atend = !!(cursor->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND); 
  return pager->getRowOfPos(getCursorPos(), dmy, atend); 
  }


FXuint FXPLText::getLineOfPos(FXint pos, FXint & offs) const {
  FXuint line = textsrc->getLineOfPos(pos);
  offs = pos - textsrc->getPosOfLine(line);
  return line;
  }


/*******************************************************************************/


// Set back color.  Special case because property of base class.
long FXPLText::onCmdBackColor(FXObject*,FXSelector,void* ptr){
  setBackColor((FXColor)(long)ptr);
  return 1;
  }


// Update back color
long FXPLText::onUpdBackColor(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(FXWindow::ID_SETVALUE,SEL_COMMAND),(void*)backColor);
  return 1;
  }


// General colour setting
long FXPLText::onCmdColor(FXObject *, FXSelector sel,void * ptr) {
  FXColor clr = (FXColor)ptr;
  FXint which = SELID(sel) - ID_FIRST_COLOR;
  if (which < PLCOLOR_FIRST || which >= PLCOLOR_LAST)
    return 0;
  if(clr!=colors[which]){
    colors[which]=clr;
    update();
    }
  return 1;
  }

// General colour update
long FXPLText::onUpdColor(FXObject * sender, FXSelector sel, void *) {
  FXint which = SELID(sel) - ID_FIRST_COLOR;
  if (which < PLCOLOR_FIRST || which >= PLCOLOR_LAST)
    return 0;
  sender->handle(this, MKUINT(FXWindow::ID_SETVALUE, SEL_COMMAND),
  		(void*)colors[which]);
  return 1;
  }


// Update value from a message
long FXPLText::onCmdSetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdSetStringValue: NULL pointer.\n",getClassName()); }
  setText(*((FXString*)ptr));
  return 1;
  }


// Obtain value from text.
long FXPLText::onCmdGetStringValue(FXObject*,FXSelector,void* ptr){
  if(ptr==NULL){ fxerror("%s::onCmdGetStringValue: NULL pointer.\n",getClassName()); }
  *((FXString*)ptr) = getText();          //// This may be expensive
  return 1;
  }


/*******************************************************************************/


// Save object to stream
void FXPLText::save(FXStream& store) const {
  FXComposite::save(store);
  FXASSERT(0);//// FIXME
  }


// Load object from stream
void FXPLText::load(FXStream& store){
  FXComposite::load(store);
  FXASSERT(0);//// FIXME
  }


// Clean up
FXPLText::~FXPLText(){
  if(scrolltimer){getApp()->removeTimeout(scrolltimer);}
  horizontal=(FXScrollbar*)-1;
  vertical=(FXScrollbar*)-1;
  corner=(FXScrollCorner*)-1;
  scrolltimer=(FXTimer*)-1;
  FXFREE(&clipbuffer);
  clipbuffer=(FXchar*)-1;
  font=(FXFont*)-1;
  auxFont=(FXFont*)-1;
  delete cursor;
  delete pager;
  if (frd_private) {
    if (frd)
      delete frd;
    frd = (FXFindReplaceDialog *)-1;
    }
  if (rds)
    delete rds;
  rds = (RowDirty *)-1;
  }
