/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                              Text Widget                                      *
*********************************************************************************
*   $Id: FXPLText.h,v 1.2 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Jeroen van der Zijp
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXPLText.h,v $
* Revision 1.2  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.1.1.1  2000/10/29 14:16:12  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXPLTEXT_H
#define FXPLTEXT_H


#ifndef FXROWINFO_H
#include "FXRowInfo.h"
#endif

#define FXPLTEXT_MAX_CURSOR_STYLES 4	// Number of cursor 'styles' in local cache.

/**
* Text widget options.  Numerically compatible with FXText/FXScrollArea.
* Horizontal scrolling is not implemented.  Options relating to horizontal scrolling
* will be ignored.
*/
enum {
  PLTEXT_SCROLLERS_NORMAL     = 0,		/// Show the scrollbars when needed 
  PLTEXT_HSCROLLER_ALWAYS     = 0x00008000,	/// Always show horizontal scrollers 
  PLTEXT_HSCROLLER_NEVER      = 0x00010000,	/// Never show horizontal scrollers
  PLTEXT_VSCROLLER_ALWAYS     = 0x00020000,	/// Always show vertical scrollers
  PLTEXT_VSCROLLER_NEVER      = 0x00040000,	/// Never show vertical scrollers
  PLTEXT_HSCROLLING_ON        = 0,		/// Horizontal scrolling turned on (default)
  PLTEXT_HSCROLLING_OFF       = PLTEXT_HSCROLLER_NEVER|PLTEXT_HSCROLLER_ALWAYS,  /// Horizontal scrolling turned off
  PLTEXT_VSCROLLING_ON        = 0,		/// Vertical scrolling turned on (default)
  PLTEXT_VSCROLLING_OFF       = PLTEXT_VSCROLLER_NEVER|PLTEXT_VSCROLLER_ALWAYS,  /// Vertical scrolling turned off
  PLTEXT_SCROLLERS_TRACK      = 0,		/// Scrollers track continuously for smooth scrolling
  PLTEXT_SCROLLERS_DONT_TRACK = 0x00080000,	/// Scrollers don't track continuously

  PLTEXT_READWRITE     = 0x00000000,		/// Read/write anywhere
  PLTEXT_READONLY      = 0x00100000,		/// Text is NOT editable
  PLTEXT_PREPEND       = 0x01000000,		/// Add lines at top only
  PLTEXT_APPEND = PLTEXT_READONLY|PLTEXT_PREPEND,/// Add lines at bottom only
  PLTEXT_EDITMASK      = PLTEXT_APPEND,		/// Mask for read/write access.
  
  PLTEXT_SEL_BY_ROW    = 0x00200000,		/// Triple-click selects by row, rather than line.
  PLTEXT_OVERSTRIKE    = 0x00400000,		/// Overstrike mode
  PLTEXT_AUTO_INDENT   = 0x00800000,		/// Auto-indent when newline
  PLTEXT_SINGLE        = 0x02000000,		/// Single line edit (like text field)
  PLTEXT_SHOW_AUX      = 0x04000000,		/// Show the auxiliary area
  PLTEXT_SHOW_LINE_NUMBERS = 0x08000000,	/// Show line numbers in aux area
  PLTEXT_FLASH_CURSOR  = 0x10000000,		/// Unconditionally flash cursor
  PLTEXT_ESCAPE        = 0x20000000,		/// Editor in escape mode
  PLTEXT_SHOW_TABS     = 0x40000000,		/// Show tabs as arrows (and newlines too).
  PLTEXT_GREEK         = 0x80000000,		/// Show as 'greek text' instead of normal
  						/// fonts.
  };

   
/// Selection modes
enum FXPLTextSelectionMode {
  PLTEXT_SELECT_CHARS,
  PLTEXT_SELECT_WORDS,
  PLTEXT_SELECT_LINES
  };

class FXPageFormat;
class FXTextSrc;
class FXTextAnchor;
class FXTextCursor;
struct FXFindData;
class FXFindReplaceDialog;
#ifdef POSIX_REGEX
struct regex_t;
#else
struct fxregexp; 
#endif
 
/**
* Enhanced multiline text widget.  Unlike the old text widget, this one is derived from
* FXComposite and implements its own scroll area functionality.
*/
class FXAPI FXPLText : public FXComposite {
  FXDECLARE(FXPLText)
protected:
  FXbool ok;				// Flag to indicate construction completed successfully
  // These fields are roughly compatible with FXScrollArea, except that the units are in
  // characters not pixels (unless otherwise noted), and horizontal scrolling is not implemented.
  FXScrollbar    *horizontal;		// Horz. scrollbar - not implemented, always null
  FXScrollbar    *vertical;		// Vert. scrollbar.  Scroll unit is line not pixel
  FXScrollCorner *corner;		// Not implemented, always null.
  FXTimer        *scrolltimer;		// Autoscroll timer.
  FXint           viewport_w;		// Viewable area width
  FXint           viewport_h;		// Viewable area height
  FXint           viewport_w_pix;	// Viewable area width in pixels (whole chars only)
  FXint           viewport_h_pix;	// Viewable area height in pixels (whole chars only)
  FXint           content_w;		// Text width - not implemented
  FXint           content_h;		// Number of _rows_ in text
  FXint           pos_x;		// Not implemented
  FXint           pos_y;		// Row number of top row, negated
  
  FXbool        want_visible;		// Cursor should remain visible
  FXint         nvisrows;		// Number of visible rows. This does not include any
  					// partially visible row at bottom.
  FXint		partialRow;		// Whether partial row visible at bottom.  Zero if
  					// no partial row, otherwise 1.  The partial row is
  					// displayed, but most of the above dimensions do not
  					// count it.
  FXint         toprow;               // Row number of first visible row
  FXint         margintop;            // Margins top
  FXint         marginbottom;         // Margin bottom
  FXint         marginleft;           // Margin left
  FXint         marginright;          // Margin right
  FXFont       *font;                 // Text font - must be monospaced
  FXint         charWidth;            // Text metric: character width
  FXint         charHeight;           // Text metric: character height.  If the aux font is
  					// larger, then charHeight and charBase will be set
  					// to the auxFont metrics.
  FXint         charBase;             // Text metric: character baseline
  
  FXint         auxwidth;		// Pixel width of auxiliary area display (incl border).
  FXint		set_auxwidth;		// Desired width (set by API call) when shown.
  FXint		sepwidth;		// Pixel width of separator between aux and text areas
  FXFont       *auxFont;              // Auxiliary font - may be monospaced or proportional
  FXint         auxCharWidth;         // Text metric: aux font character width
  FXint         auxCharHeight;        // Text metric: aux font character height
  FXint         auxCharBase;          // Text metric: aux font character baseline
  FXint         greek_x, greek_y;     // Greek text "char" size
  FXint         gx1,gx2,gy1,gy2;
  FXPLText     *track;                // If not null, is widget we are tracking.
  FXuint        dragmode;             // Drag mode
  FXint         grabx;                // Grab point x
  FXint         graby;                // Grab point y
  FXbool        cursorActive;		// Cursor is potentially visible 
  FXchar       *clipbuffer;           // Clipped text
  FXint         cliplength;           // Length of clipped text
  FXuchar      *clipcopy;		// Copy of clip, modified with \r\n for DOS etc. 
  FXString	help;
  
  FXPageFormat *pager;			// Describes which chars go where, maps rows to lines.
  FXTextSrc    *textsrc;		// Underlying text source, cached from pager->getTextSrc().
  FXTextCursor *cursor;			// The cursor
  FXint		prefcol;		// Preferred cursor column when moving up/down rows.
  FXint		setpref;		// State machine for use of prefcol (see setPrefcol())
  struct {
    FXuint onstyle;
    FXuint offstyle;
    FXuint ontime;
    FXuint offtime;
    FXbool flash;
    } cstyles[FXPLTEXT_MAX_CURSOR_STYLES];
  FXint cur_cstyle;
  
  
  FXint		anchorpos;		// Anchor position.
  FXint		rolines;		// Number of read-only lines (non-zero in append or prepend mode).

  FXbool frd_private;			// True if following object private to this widget.
  FXFindReplaceDialog * frd;		// Search/replace facility
  
  // These fields are to assist with graphics optimisation.  RowDirty struct indicates,
  // for each visible row, what range of characters needs to be redrawn.  Only a single
  // stretch of chars in any column can be set not, say, one char at the start and one at
  // the end.
  struct RowDirty {
    FXbool	aux;		// Aux data column
    FXushort	start;		// Starting column number
    FXushort	end;		// End column+1.  If end<=start, this row not dirty.
    };
  
  FXbool	anydirty;	// Set if anything dirty
  RowDirty     *rds;		// Point to array dimensioned (nvisrows+partialRow)
  
  // Colours, and my aren't there a lot of them!  So many that we put them in an array
  // and index the particular one by number.
public:
  enum {
    PLCOLOR_FIRST
   ,PLCOLOR_AUX_BG = PLCOLOR_FIRST /// Aux area background
   ,PLCOLOR_AUX_SEPARATOR	/// Aux area separator line
   ,PLCOLOR_AUX_LINE_NUMS	/// Line number text
   ,PLCOLOR_AUX_HIGHLIGHT	/// Text when highlit
   ,PLCOLOR_GREEK_WINDOW	/// Visible window border for greek text
   ,PLCOLOR_AUX_RESV2		// Reserved
   ,PLCOLOR_AUX_RESV3		//
   ,PLCOLOR_AUX_RESV4		//
   ,PLCOLOR_TEXT_FG		/// Default text colour
   ,PLCOLOR_TEXT_SEL_BG		/// Selection background (bit 0 in pager attribute matrix) -
   ,PLCOLOR_SELECTED_BG = PLCOLOR_TEXT_SEL_BG
   ,PLCOLOR_TEXT_SEL_FG		/// Selection foreground
   ,PLCOLOR_SELECTED_FG = PLCOLOR_TEXT_SEL_FG
   // defaults for following are 'not used' by bg_mask and fg_mask bits.
   ,PLCOLOR_TEXT_ATTR1_BG	/// Background bit 1 in pager attribute matrix
   ,PLCOLOR_ERRORS_BG = PLCOLOR_TEXT_ATTR1_BG
   ,PLCOLOR_TEXT_ATTR1_FG	/// Foreground bit 1 in pager attribute matrix
   ,PLCOLOR_ERRORS_FG = PLCOLOR_TEXT_ATTR1_FG
   ,PLCOLOR_TEXT_ATTR2_BG	/// bit 2
   ,PLCOLOR_SECONDARY_BG = PLCOLOR_TEXT_ATTR2_BG
   ,PLCOLOR_TEXT_ATTR2_FG	///
   ,PLCOLOR_SECONDARY_FG = PLCOLOR_TEXT_ATTR2_FG
   ,PLCOLOR_TEXT_ATTR3_BG	/// bit 3
   ,PLCOLOR_RETENTIVE_BG = PLCOLOR_TEXT_ATTR3_BG
   ,PLCOLOR_TEXT_ATTR3_FG	///
   ,PLCOLOR_RETENTIVE_FG = PLCOLOR_TEXT_ATTR3_FG
   ,PLCOLOR_TEXT_ATTR4_BG	/// bit 4
   ,PLCOLOR_MARKER_BG = PLCOLOR_TEXT_ATTR4_BG
   ,PLCOLOR_TEXT_ATTR4_FG	///
   ,PLCOLOR_MARKER_FG = PLCOLOR_TEXT_ATTR4_FG
   ,PLCOLOR_TEXT_ATTR5_BG	/// bit 5
   ,PLCOLOR_INPUT_BG = PLCOLOR_TEXT_ATTR5_BG
   ,PLCOLOR_TEXT_ATTR5_FG	///
   ,PLCOLOR_INPUT_FG = PLCOLOR_TEXT_ATTR5_FG
   ,PLCOLOR_TEXT_ATTR6_BG	/// bit 6
   ,PLCOLOR_OUTPUT_BG = PLCOLOR_TEXT_ATTR6_BG
   ,PLCOLOR_TEXT_ATTR6_FG	///
   ,PLCOLOR_OUTPUT_FG = PLCOLOR_TEXT_ATTR6_FG
   ,PLCOLOR_TEXT_SPECIAL_BG	/// bit 7 (special characters e.g. tab, newline, control chars)
   ,PLCOLOR_SPECIAL_BG = PLCOLOR_TEXT_SPECIAL_BG
   ,PLCOLOR_TEXT_SPECIAL_FG	/// special char foreground
   ,PLCOLOR_SPECIAL_FG = PLCOLOR_TEXT_SPECIAL_FG
   ,PLCOLOR_CURSOR		/// Cursor
   ,PLCOLOR_LAST	// Dummy for array size
    };
protected:
  FXColor	colors[PLCOLOR_LAST];
  
  // The colours assigned to particular attribute bits are obtained from the first 1 bit
  // (counting from bit 0) in the following masks ANDed with the current attribute matrix
  // element.  If none are set, then the defaults (TEXT_FG/BG) are used.
  FXattr	bg_mask;	// Mask of attribute bits which specify background colour
  FXattr	fg_mask;	// Mask of attribute bits which specify foreground colour

protected:
  FXPLText();
  void setDefaults();
  virtual void layout();
  void recompute();
  FXbool startAutoScroll(FXint x,FXint y,FXbool onlywheninside=FALSE);
  void stopAutoScroll();
  FXColor setFg(FXDCWindow & dc, FXattr a, FXbool force = FALSE);
  FXColor setBgToFg(FXDCWindow & dc, FXattr a, FXbool force = FALSE); 
  FXColor getFg(FXattr a);
  FXColor getBg(FXattr a); 
  void drawAnchors(FXRectangle & rect, FXDCWindow & dc, FXbool underlay); 
  void renderRect(FXEvent * event);
  
  void setClean();	// Set all rows clean
  void setDirty();	// Set everything dirty for complete redraw
  void dirtyRectangle(const FXRectangle & bb, FXint r, FXuint c, FXbool in_aux); 
  			// Set dirty rectangle.
  
  // Get the current selection (not necessarily in this widget) as a string.  Returns null
  // if there is no appropriate selection.  The caller MUST free the returned string after use.
  FXchar * getStringSel(FXint & len);
  FXint startSearchPos(FXbool fwd, FXbool restart = FALSE) const;
  FXint doSearchResults(FXint pos, FXint len); 
  
  void setPrefcol();
  void setCharMetrics();

private:
  FXPLText(const FXPLText&);
  FXPLText& operator=(const FXPLText&);
public:
  
  // Messages for scroll area functionality
  long onVScrollerChanged(FXObject*,FXSelector,void*);
  long onVScrollerDragged(FXObject*,FXSelector,void*);
  long onAutoScroll(FXObject*,FXSelector,void*);
   
  // System messages
  long onPaint(FXObject*,FXSelector,void*);
  long onFocusIn(FXObject*,FXSelector,void*);
  long onFocusOut(FXObject*,FXSelector,void*);
  long onLeftBtnPress(FXObject*,FXSelector,void*);
  long onLeftBtnRelease(FXObject*,FXSelector,void*);
  long onMiddleBtnPress(FXObject*,FXSelector,void*);
  long onMiddleBtnRelease(FXObject*,FXSelector,void*);
  long onRightBtnPress(FXObject*,FXSelector,void*);
  long onRightBtnRelease(FXObject*,FXSelector,void*);
  long onUngrabbed(FXObject*,FXSelector,void*);
  long onMotion(FXObject*,FXSelector,void*);
  long onBeginDrag(FXObject*,FXSelector,void*);
  long onEndDrag(FXObject*,FXSelector,void*);
  long onDragged(FXObject*,FXSelector,void*);
  long onDNDEnter(FXObject*,FXSelector,void*);
  long onDNDLeave(FXObject*,FXSelector,void*);
  long onDNDMotion(FXObject*,FXSelector,void*);
  long onDNDDrop(FXObject*,FXSelector,void*);
  long onDNDRequest(FXObject*,FXSelector,void*);
  long onSelectionLost(FXObject*,FXSelector,void*);
  long onSelectionGained(FXObject*,FXSelector,void*);
  long onSelectionRequest(FXObject*,FXSelector,void* ptr);
  long onClipboardLost(FXObject*,FXSelector,void*);
  long onClipboardGained(FXObject*,FXSelector,void*);
  long onClipboardRequest(FXObject*,FXSelector,void*);
  long onKeyPress(FXObject*,FXSelector,void*);
  long onKeyRelease(FXObject*,FXSelector,void*);

  long onQueryHelp(FXObject*,FXSelector,void*);
  long onChanged(FXObject*,FXSelector,void*);
  long onSelected(FXObject*,FXSelector,void*);
  long onDeselected(FXObject*,FXSelector,void*);
  long onInserted(FXObject*,FXSelector,void*);
  long onDeleted(FXObject*,FXSelector,void*);
  long onCmdToggleEscape(FXObject*,FXSelector,void*);
  long onUpdToggleEscape(FXObject*,FXSelector,void*);
  long onCmdToggleEditable(FXObject*,FXSelector,void*);
  long onUpdToggleEditable(FXObject*,FXSelector,void*);
  long onCmdToggleOverstrike(FXObject*,FXSelector,void*);
  long onUpdToggleOverstrike(FXObject*,FXSelector,void*);
  long onCmdToggleFlash(FXObject*,FXSelector,void*);
  long onUpdToggleFlash(FXObject*,FXSelector,void*);
  long onCmdCursorRow(FXObject*,FXSelector,void*);
  long onUpdCursorRow(FXObject*,FXSelector,void*);
  long onCmdCursorLine(FXObject*,FXSelector,void*);
  long onUpdCursorLine(FXObject*,FXSelector,void*);
  long onCmdCursorColumn(FXObject*,FXSelector,void*);
  long onUpdCursorColumn(FXObject*,FXSelector,void*);
  long onCmdCursorOffset(FXObject*,FXSelector,void*);
  long onUpdCursorOffset(FXObject*,FXSelector,void*);
  long onUpdHaveSelection(FXObject*,FXSelector,void*);
  long onCmdSetStringValue(FXObject*,FXSelector,void*);
  long onCmdGetStringValue(FXObject*,FXSelector,void*);
  long onCmdTabColumns(FXObject*,FXSelector,void*);
  long onUpdTabColumns(FXObject*,FXSelector,void*);
  long onCmdWrapColumns(FXObject*,FXSelector,void*);
  long onUpdWrapColumns(FXObject*,FXSelector,void*);
  
  // Cursor movement
  long onCmdCursorTop(FXObject*,FXSelector,void*);
  long onCmdCursorBottom(FXObject*,FXSelector,void*);
  long onCmdCursorHome(FXObject*,FXSelector,void*);
  long onCmdCursorEnd(FXObject*,FXSelector,void*);
  long onCmdCursorRight(FXObject*,FXSelector,void*);
  long onCmdCursorLeft(FXObject*,FXSelector,void*);
  long onCmdCursorUp(FXObject*,FXSelector,void*);
  long onCmdCursorDown(FXObject*,FXSelector,void*);
  long onCmdCursorWordLeft(FXObject*,FXSelector,void*);
  long onCmdCursorWordRight(FXObject*,FXSelector,void*);
  long onCmdCursorPageDown(FXObject*,FXSelector,void*);
  long onCmdCursorPageUp(FXObject*,FXSelector,void*);
  long onCmdCursorScreenTop(FXObject*,FXSelector,void*);
  long onCmdCursorScreenBottom(FXObject*,FXSelector,void*);
  long onCmdCursorScreenCenter(FXObject*,FXSelector,void*);
  long onCmdCursorParHome(FXObject*,FXSelector,void*);
  long onCmdCursorParEnd(FXObject*,FXSelector,void*);
  
  // Mark and extend
  long onCmdMark(FXObject*,FXSelector,void*);
  long onCmdExtend(FXObject*,FXSelector,void*);
  
  // Inserting
  long onCmdInsertChar(FXObject*,FXSelector,void*);
  long onCmdInsertString(FXObject*,FXSelector,void*);
  long onCmdInsertNewline(FXObject*,FXSelector,void*);
  
  // Manipulation Selection
  long onCmdCutSel(FXObject*,FXSelector,void*);
  long onCmdCopySel(FXObject*,FXSelector,void*);
  long onCmdPasteSel(FXObject*,FXSelector,void*);
  long onCmdDeleteSel(FXObject*,FXSelector,void*);
  
  // Changing Selection
  long onCmdSelectChar(FXObject*,FXSelector,void*);
  long onCmdSelectWord(FXObject*,FXSelector,void*);
  long onCmdSelectLine(FXObject*,FXSelector,void*);
  long onCmdSelectAll(FXObject*,FXSelector,void*);
  long onCmdDeselectAll(FXObject*,FXSelector,void*);
  
  // Deletion
  long onCmdBackspace(FXObject*,FXSelector,void*);
  long onCmdBackspaceWord(FXObject*,FXSelector,void*);
  long onCmdBackspaceBol(FXObject*,FXSelector,void*);
  long onCmdDelete(FXObject*,FXSelector,void*);
  long onCmdDeleteWord(FXObject*,FXSelector,void*);
  long onCmdDeleteEol(FXObject*,FXSelector,void*);
  long onCmdDeleteLine(FXObject*,FXSelector,void*);
  
  // Colors
  long onCmdBackColor(FXObject*,FXSelector,void*);
  long onUpdBackColor(FXObject*,FXSelector,void*);
  long onCmdColor(FXObject*,FXSelector,void*);
  long onUpdColor(FXObject*,FXSelector,void*);
  
  // Find/replace functions.
  long onCmdFind(FXObject*,FXSelector,void*);
  long onCmdFindNext(FXObject*,FXSelector,void*);
  long onCmdFindStr(FXObject*,FXSelector,void*);
  long onCmdFindAll(FXObject*,FXSelector,void*);
  long onCmdReplStr(FXObject*,FXSelector,void*);
  long onCmdReplAllSel(FXObject*,FXSelector,void*);
  long onCmdReplAll(FXObject*,FXSelector,void*);
  
  long onCmdGotoSelLine(FXObject*,FXSelector,void*);
  long onCmdIndent(FXObject*,FXSelector,void*);
  long onCmdRoll(FXObject*,FXSelector,void*);
  long onCmdCursorToRoll(FXObject*,FXSelector,void*);
  
  long onCmdToggleShowtabs(FXObject*,FXSelector,void*);
  long onCmdToggleAutoindent(FXObject*,FXSelector,void*);
  long onCmdToggleShowaux(FXObject*,FXSelector,void*);
  long onUpdToggleShowtabs(FXObject*,FXSelector,void*);
  long onUpdToggleAutoindent(FXObject*,FXSelector,void*);
  long onUpdToggleShowaux(FXObject*,FXSelector,void*);
  long onCmdToggleSecSel(FXObject*,FXSelector,void*);
  long onCmdClearSecSel(FXObject*,FXSelector,void*);
  long onCmdClearAllSecSel(FXObject*,FXSelector,void*);
  
  long onTrackTop(FXObject*,FXSelector,void*);
  
public:
  
  enum {
    ID_CURSOR_TOP=FXComposite::ID_LAST,
    ID_CURSOR_BOTTOM,
    ID_CURSOR_HOME,
    ID_CURSOR_END,
    ID_CURSOR_RIGHT,
    ID_CURSOR_LEFT,
    ID_CURSOR_UP,
    ID_CURSOR_DOWN,
    ID_CURSOR_WORD_LEFT,
    ID_CURSOR_WORD_RIGHT,
    ID_CURSOR_PAGEDOWN,
    ID_CURSOR_PAGEUP,
    ID_CURSOR_SCRNTOP,
    ID_CURSOR_SCRNBTM,
    ID_CURSOR_SCRNCTR,
    ID_CURSOR_PAR_HOME,
    ID_CURSOR_PAR_END,
    ID_MARK,
    ID_EXTEND,
    ID_INSERT_CHAR,
    ID_INSERT_STRING,
    ID_INSERT_NEWLINE,
    ID_CUT_SEL,
    ID_COPY_SEL,
    ID_PASTE_SEL,
    ID_DELETE_SEL,
    ID_SELECT_CHAR,
    ID_SELECT_WORD,
    ID_SELECT_LINE,
    ID_SELECT_ALL,
    ID_DESELECT_ALL,
    ID_BACKSPACE,
    ID_BACKSPACE_WORD,
    ID_BACKSPACE_BOL,
    ID_DELETE,
    ID_DELETE_WORD,
    ID_DELETE_EOL,
    ID_DELETE_LINE,
    ID_TOGGLE_ESCAPE,
    ID_TOGGLE_EDITABLE,
    ID_TOGGLE_OVERSTRIKE,
    ID_TOGGLE_FLASH,
    ID_CURSOR_ROW,
    ID_CURSOR_COLUMN,
    ID_WRAP_COLUMNS,
    ID_TAB_COLUMNS,
    ID_BACK_COLOR,
    ID_FIRST_COLOR,
    ID_TEXT_COLOR = ID_FIRST_COLOR + PLCOLOR_TEXT_FG - PLCOLOR_FIRST,
    ID_SELTEXT_COLOR = ID_FIRST_COLOR + PLCOLOR_TEXT_SEL_FG - PLCOLOR_FIRST,
    ID_SELBACK_COLOR = ID_FIRST_COLOR + PLCOLOR_TEXT_SEL_BG - PLCOLOR_FIRST,
    ID_CURSOR_COLOR = ID_FIRST_COLOR + PLCOLOR_CURSOR - PLCOLOR_FIRST,
    ID_LAST_COLOR = ID_FIRST_COLOR + PLCOLOR_LAST - 1 - PLCOLOR_FIRST,

    ID_CURSOR_LINE,
    ID_CURSOR_OFFSET,
    
    ID_FIND,			// Pop up find/replace dialogue
    ID_FINDNEXT,		// Find next matching current selection
    ID_FINDPREV,		// Find previous matching current selection
    ID_FINDSTR,			// Find specific string and select it (FindData is data)
    ID_FINDALL,			// Find all matching strings and secondary-select them
    ID_REPLSTR,			// Replace current selection with specific string
    ID_REPLALLSEL,		// Replace all secondary selections with specific string
    ID_REPLALL,			// Replace all occurrences with specific string
    
    ID_GOTO_SEL_LINE,		// Goto selected line (under X, selection may be anywhere)
    ID_INDENT,			// Indent/outdent selection
    ID_ROLLBACK,		// Undo
    ID_ROLLFORWARD,		// Redo
    ID_CURSOR_TO_ROLLBACK,	// Move cursor to site of next undo
    ID_CURSOR_TO_ROLLFORWARD,	// Move cursor to site of next redo
    
    ID_TOGGLE_SHOWTABS,		// Toggle whether show tabs as arrows
    ID_TOGGLE_AUTOINDENT,	// Toggle whether to auto-indent new lines
    ID_TOGGLE_SHOWAUX,		// Toggle whether to show aux area
    ID_TOGGLE_SECSEL,
    ID_CLEAR_SECSEL,
    ID_CLEAR_ALL_SECSEL,
    
    ID_TRACK_TOP,		// Tracking another widget's display position
    
    ID_LAST
    };
    
public:

  /// Construct multi-line text widget.  Old-style, compatible with FXText.
  FXPLText(FXComposite *p,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0,
           FXint x=0,FXint y=0,FXint w=0,FXint h=0);

  /// Construct multi-line text widget.  Supply a text source that may be shared between
  /// multiple widgets.
  FXPLText(FXComposite *p,FXTextSrc* ts,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0,
           FXint x=0,FXint y=0,FXint w=0,FXint h=0); 

  /// Test whether constructor completed successfully.  Since ctors may allocate memory etc.,
  /// it is necessary for the caller to check whether construction was sucessful.  If not,
  /// then the dtor should be invoked as soon as possible and an error flagged to the user.
  /// Note that the ctor is exclusively responsible for setting this flag.
  FXbool isOk() const { return ok; }
  
  /// Make this widget visually like the specified one.  (Same colours, options etc.)
  void makeLike(FXPLText * model);
  
  /* ****************** FXScrollArea compatibility functions... *******************/
  /// Return default width in units of _pixels_.
  virtual FXint getDefaultWidth();

  /// Return default height in units of _pixels_.
  virtual FXint getDefaultHeight();
  
  /// Return viewport size in units of _pixels_.
  virtual FXint getViewportHeight();
  virtual FXint getViewportWidth();
  
  /// Return content size in units of chars.
  virtual FXint getContentWidth();
  virtual FXint getContentHeight();
  
  /// Change scroll style
  void setScrollStyle(FXuint style);
  
  /// Return scroll style
  FXuint getScrollStyle() const;
  
  /// Return TRUE if horizontally scrollable.  Not implemented, always returns FALSE.
  FXbool isHorizontalScrollable() const;
  
  /// Return TRUE if vertically scrollable
  FXbool isVerticalScrollable() const;

  /// Return a pointer to the horizontal scrollbar.  Not implemented, will return null.
  FXScrollbar* horizontalScrollbar() const { return horizontal; }

  /// Return a pointer to the vertical scrollbar
  FXScrollbar* verticalScrollbar() const { return vertical; }

  /// Return the current x-position.  Units of chars.  x not implemented.
  FXint getXPosition() const { return pos_x; }

  /// Return the current y-position.  Units of chars.
  FXint getYPosition() const { return pos_y; }

  /// Set the current position.  Units of characters.  x not implemented.
  void setPosition(FXint x,FXint y);

  /// Get the current position.  Units of characters.  x not implemented.
  void getPosition(FXint& x,FXint& y) const { x=pos_x; y=pos_y; }

  /// Move contents to the specified position.  Units of characters.  x not implemented.
  virtual void moveContents(FXint x,FXint y);

  /* **************** ...end of FXScrollArea compatibility functions. **************/
  
  /// Convert pixel units to character units (rounding down).  If no font, uses 1:1.
  /// May be same vars in as out.
  void pix2char(FXint w_pix, FXint h_pix, FXint & w_char, FXint & h_char) const;
  
  /// Convert pixel units to character units (rounding up).  If no font, uses 1:1
  /// May be same vars in as out.
  void pix2charUp(FXint w_pix, FXint h_pix, FXint & w_char, FXint & h_char) const;
  
  /// Convert character height units to pixels.  If no font, uses 1:1
  FXint charheight2pix(FXint h) const { return h * charHeight; }
  
  /// Convert character width units to pixels.  If no font, uses 1:1
  FXint charwidth2pix(FXint w) const { return w * charWidth; }
  
  /// Convert character matrix column number to absolute x pixel of where the baseline of
  /// the corresponding char should be rendered.
  FXint matrixCol2pix(FXuint c) const { return auxwidth + marginleft + charwidth2pix(c); }
  
  /// Convert character matrix relative row number to absolute y pixel of where the baseline of
  /// the corresponding char should be rendered.
  FXint matrixRow2pix(FXuint r) const { return charheight2pix(r) + margintop + charBase; }

  /// Create server-side resources
  virtual void create();
  
  /// Detach server-side resources
  virtual void detach();
  
  /// Enable the text widget
  virtual void enable();
  
  /// Disable the text widget
  virtual void disable();
  
  /// Need to recalculate size
  virtual void recalc();

  /// Resize this window to the specified width and height 
  virtual void resize(FXint w,FXint h);
  
  /// Move and resize this window in the parent's coordinates
  virtual void position(FXint x,FXint y,FXint w,FXint h);
  
  /// Returns true because a text widget can receive focus
  virtual FXbool canFocus() const;
  
  /// Find next instance of character after pos
  FXint findf(FXchar c,FXint pos) const;
  
  /// Find previous instance of character before pos
  FXint findb(FXchar c,FXint pos) const;
  
  /// Change top margin
  void setMarginTop(FXint pt);
  
  /// Return top margin
  FXint getMarginTop() const { return margintop; }

  /// Change bottom margin
  void setMarginBottom(FXint pb);
  
  /// Return bottom margin
  FXint getMarginBottom() const { return marginbottom; }

  /// Change left margin
  void setMarginLeft(FXint pl);
  
  /// Return left margin
  FXint getMarginLeft() const { return marginleft; }

  /// Change right margin
  void setMarginRight(FXint pr);
  
  /// Return right margin
  FXint getMarginRight() const { return marginright; }

  /// Return wrap columns, i.e. number of char positions visible per line.
  FXint getWrapColumns() const;

  /// Set wrap columns.  This function deprecated (doesn't do anything).
  void setWrapColumns(FXint cols) {}

  /// Return tab columns
  FXint getTabColumns() const; 

  /// Change tab columns
  void setTabColumns(FXint cols);

  /// Return TRUE if text was modified
  FXbool isModified() const;

  /// Set modified flag.  
  void setModified(FXbool mod=TRUE);
  
  /// Return TRUE if in escape mode
  FXbool isEscape() const { return !!(options & PLTEXT_ESCAPE); }
  
  /// Set escape mode T or F (usually done by user pressing ESC).  Returns previous state.
  FXbool setEscape(FXbool esc=TRUE);

  /// Return TRUE if widget is globally editable and text source is modifiable
  FXbool isEditable() const;
  
  /// Return TRUE if widget is appendable and text source is modifiable
  FXbool isAppendable() const;
  
  /// Return TRUE if widget is prependable and text source is modifiable
  FXbool isPrependable() const;
  
  /// Set editable flag of widget only.  Text source must be modifiable for
  /// this to be meaningful.
  void setEditable(FXbool edit=TRUE);
    
  /// Set append flag of widget.  nedit indicates initial number of editable lines
  /// at end of text.  movecsr indicates whether to move cursor to start of
  /// first editable line.
  void setAppendable(FXint nedit = 1, FXbool movecsr = TRUE);
    
  /// Set prepend flag of widget.  nedit indicates initial number of editable lines
  /// at start of text.  movecsr indicates whether to move cursor to start of
  /// line 1.
  void setPrependable(FXint nedit = 1, FXbool movecsr = TRUE);
  
  /// Test whether range of chars is deletable.  This is false only if in append
  /// or prepend mode, and at least one char in the range is in the read-only
  /// section or one of the chars to be deleted is the newline on the boundary
  /// between editable and read-only sections.
  FXbool isDeletable(FXint what[2]) const;
    
  /// Change delimiters of words
  void setDelimiters(const FXCharset& delims) const;

  /// Return word delimiters  
  FXCharset getDelimiters() const;

  /// Change text font(s).  If either parameter is NULL, don't change that font.  First
  /// parm is for normal font, second for auxiliary font.
  void setFont(FXFont* fnt = NULL, FXFont * auxf = NULL);
  
  /// Return text font
  FXFont* getFont() const { return font; }

  /// Return auxiliary font
  FXFont* getAuxFont() const { return auxFont; }
  
  /// Change greek settings.  xpix/ypix specify horiz/vertical size for each cell.  Zero
  /// to not change that dimension.  Valid range for xpix is 0-2, ypix is 0-3.  Note that
  /// using wide windows for small greek text is going to take a lot of memory in the pager,
  /// since 2 bytes are still allocated for each visible char position.  gtgt is a target
  /// widget which we query (in greek mode) for its window/cursor position.
  /// In greek mode, the widget is read-only.
  void setGreek(FXbool on = TRUE, FXint xpix = 2, FXint ypix = 3, FXPLText * gtgt = NULL);

  /// Change text color
  void setTextColor(FXColor clr);
  
  /// Return text color
  FXColor getTextColor() const { return colors[PLCOLOR_TEXT_FG]; }

  /// Change selected background color
  void setSelBackColor(FXColor clr);
  
  /// Return selected background color
  FXColor getSelBackColor() const { return colors[PLCOLOR_TEXT_SEL_BG]; }

  /// Change selected text color
  void setSelTextColor(FXColor clr);
  
  /// Return selected text color
  FXColor getSelTextColor() const { return colors[PLCOLOR_TEXT_SEL_FG]; }

  /// Change cursor color
  void setCursorColor(FXColor clr);
  
  /// Return cursor color
  FXColor getCursorColor() const;
  
  /// Set cursor style(s).  which is 0 for insert mode, 1 for overstrike mode.  Default
  /// style is caret with no flash.
  void setCursorStyle(FXint which, FXuint onstyle=1, FXuint offstyle=1, FXuint ontime=0, FXuint offtime=0);
  
  /// Change the current cursor to the specified style set (i.e. the which value on setCursorStyle()).
  /// Style sets range from 0 through to FXPLTEXT_MAX_CURSOR_STYLES-1.
  void changeCursor(FXint which, FXbool flash = FALSE);
  
  /// Change general color.  Ignores out-of-bounds color number.  The 'opaque' parameter
  /// is true to make the colour visible, or false to make invisible -- this affects the
  /// appropriate fg- or bgmask bits (see set[F|B]gMask()).
  void setColor(FXint whichcolor, FXColor clr, FXbool opaque = TRUE);
  
  /// Return general color.  Returns black for out-of-bounds color number.
  FXColor getColor(FXint whichcolor) const;
  
  /// Set or clear background colour mask bits.  The bitmask has a bit set for each attribute
  /// to be set or cleared.  The previous mask contents are returned.
  FXattr setBgMask(FXattr mask, FXbool set = TRUE);
  
  /// Set or clear foreground colour mask bits.  The bitmask has a bit set for each attribute
  /// to be set or cleared.  The previous mask contents are returned.
  FXattr setFgMask(FXattr mask, FXbool set = TRUE);
  
  
  /// Set help text
  void setHelpText(const FXString& text);
  
  /// Return help text
  FXString getHelpText() const { return help; }

  /// Extract n characters of text from position pos
  void extractText(FXchar *text,FXint pos,FXint n) const;
  
  /// Replace m characters at pos by n characters
  void replaceText(FXint pos,FXint m,const FXchar *text,FXint n);
  
  /// Append n characters of text at the end of the buffer
  void appendText(const FXchar *text,FXint n);
  
  /// Insert n characters of text at position pos into the buffer
  void insertText(FXint pos,const FXchar *text,FXint n);
  
  /// Remove n characters of text at position pos from the buffer
  void removeText(FXint pos,FXint n);
  
  /// Blank out n characters of text at position pos from the buffer, except whitespace and newlines preserved.
  void blankText(FXint pos,FXint n);
  
  /// Change the text in the buffer to new text
  void setText(const FXchar* text,FXint n);
  
  /// Retrieve text into buffer
  void getText(FXchar* text,FXint n) const;
  
  /// Change the text
  void setText(const FXString& text);
  
  /// Return text in the widget
  FXString getText() const;
  
  /// Return length of text.
  FXint getLength() const;

  /// Return cursor position
  FXint getCursorPos() const;
  
  /// Return the cursor object
  FXTextCursor * getCursor() const { return cursor; }

  /// Return row number given position.  Also sets col to column number in that row.  If
  /// apparent is true, then calculates apparent column number (i.e. takes into account
  /// expansion of tabs etc.)  Otherwise, returns character offset within row.
  FXuint getRowOfPos(FXint pos, FXint & col, FXbool apparent = TRUE) const;
  
  /// Return row number given position.
  FXuint getRowOfPos(FXint pos) const; 
  
  /// Return line number given position.
  FXuint getLineOfPos(FXint pos) const; 
  
  /// Return line number given position, and offset (in chars) within that line.
  FXuint getLineOfPos(FXint pos, FXint & offs) const; 
  
  /// Return row number of cursor.
  FXuint getCursorRow() const;

  /// Return line number of cursor.
  FXuint getCursorLine() const;

  /// Return window X pixel coord of position
  FXint getXOfPos(FXint pos) const;

  /// Return window Y pixel coord of position
  FXint getYOfPos(FXint pos) const;

  /// Return selection starting position
  FXint getSelStartPos() const;

  /// Return selection ending position
  FXint getSelEndPos() const;

  /// Return TRUE if position pos is selected.  This includes the position immediately
  /// following the selection.
  FXbool isPosSelected(FXint pos) const;
  
  /// Return enclosing TA if position pos is in or at the end of a secondary selection
  FXTextAnchor * getPosSecSel(FXint pos) const;
  
  /// Return text position at given visible (or invisible) x,y coordinate
  FXint getPosAt(FXint x,FXint y) const;
  
  /// Return position of begin of line containing position pos
  FXint lineStart(FXint pos) const;
  
  /// Return position of end of line containing position pos
  FXint lineEnd(FXint pos) const;
  
  /// Return start of next line
  FXint nextLine(FXint pos,FXint nl=1) const;
  
  /// Return start of previous line
  FXint prevLine(FXint pos,FXint nl=1) const;
  
  /// Return row start
  FXint rowStart(FXint pos) const;
  
  /// Return row end
  FXint rowEnd(FXint pos) const;
  
  /// Return start of next row
  FXint nextRow(FXint pos,FXint nr=1) const;
  
  /// Return start of previous row
  FXint prevRow(FXint pos,FXint nr=1) const;

  /// Determine of position is visible in window
  FXbool posVisible(FXint pos) const;
  
  /// Get character by position
  FXchar getChar(FXint pos) const;
  
  /// Replace character at position
  void setChar(FXint pos,FXchar ch);
  
  /// Return end of previous word
  FXint leftWord(FXint pos) const;
  
  /// Return begin of next word
  FXint rightWord(FXint pos) const;
  
  /// Return begin of word
  FXint wordStart(FXint pos) const;
  
  /// Return end of word
  FXint wordEnd(FXint pos) const;
  
  /// Return validated position
  FXint validPos(FXint pos) const;
  
  /// Make line containing pos the top line
  void setTopLine(FXint pos);
  
  /// Return position of top line
  FXint getTopLine() const;
  
  /// Make line containing pos the bottom line
  void setBottomLine(FXint pos);
  
  /// Return the position of the bottom line
  FXint getBottomLine() const;
  
  /// Make line containing pos the center line
  void setCenterLine(FXint pos);

  /// Set the anchor position
  void setAnchorPos(FXint pos);
  
  /// Return the anchor position
  FXint getAnchorPos() const;

  /// Set the cursor position
  void setCursorPos(FXint pos);
  
  /// Move cursor to specified row, maintaining the same real or apparent column
  void moveCursorToRow(FXint row, FXbool apparent = TRUE);
  
  /// Move cursor to specified line, maintaining the same line char offset
  void moveCursorToLine(FXint line); 
  
  /// Select len characters starting at given position pos
  FXbool setSelection(FXint pos,FXint len);
  
  /// Secondary select len characters starting at given position pos
  FXTextAnchor * setSecSel(FXint pos,FXint len);
  
  /// Determine whether this widget (or another one sharing the same text source) has
  /// the selection.  This overrides FXWindow::hasSelection().
  FXbool hasSelection(void) const;
  
  /// Extend the selection from the anchor to the given position
  FXbool extendSelection(FXint pos,FXPLTextSelectionMode mode=PLTEXT_SELECT_CHARS); 
  
  /// Unselect the text
  FXbool killSelection(); 
  
  /// Scroll text to make the given position visible
  void makePositionVisible(FXint pos);

  /// Change text style
  void setTextStyle(FXuint style);
  
  /// Return text style
  FXuint getTextStyle() const;
  
  /// Set a new page formatter (must be non-null).  The old pager is returned.
  FXPageFormat * setPageFormat(FXPageFormat * pf); 
  
  /// Return page formatter.
  FXPageFormat * getPageFormat() const { return pager; }

  /// Return the current text source.
  FXTextSrc * getTextSrc() const { return textsrc; }
  
  /// Return the widget that we are tracking
  FXPLText * getTrack() const { return track; }

  /// Set desired auxiliary area pixel width (when it is shown).
  void setAuxwidth(FXint auxw);

  /// Set desired separator pixel width (when it is shown).
  void setSepwidth(FXint sepw);

  /// Test whether we have the 'focus' out of competing text widgets
  FXbool isActive() const; 
  
  /// Get find/replace dialogue.  May be NULL.
  FXFindReplaceDialog * getFindReplaceDialog() const { return frd; }

  /// Set the find/replace dialogue to a shared dialogue.  Also sets the dialogue
  /// to point to our text source.
  void setFindReplaceDialog(FXFindReplaceDialog * dlg);

  /// Called by pager when an anchor changes.  Co-ordinates are given
  /// in terms of character matrix row and column, and are inclusive range.
  /// Should only set the appropriate dirtiness (not do any drawing).  Note that
  /// r can be out-of-range.  'where' is the target position (0=main text area, 1=aux area).
  void dirtyAnchor(const FXTextAnchor * ta, FXint r, FXuint c, FXint where);
  
  /// Called by pager when a bracket changes.  Co-ordinates are given
  /// in terms of character matrix row and column, and are inclusive range.
  /// Should only set the appropriate dirtiness (not do any drawing).  realstart
  /// and realend indicate whether the given point is the real end of the bracket,
  /// or just an intermediate row when the bracket extends over several rows.  r1
  /// and r2 can be out-of-range.
  void dirtyBracket(const FXTextAnchor * ta, 
                    FXint r1, FXuint c1, FXint r2, FXuint c2,
                    FXbool realstart, FXbool realend);

  /// Called by pager when part of the character matrix has been updated.  This marks it
  /// for later redraw by updateDirty().  aux is true if the aux area needs updating,
  /// s and e indicate the start and end+1 columns which need updating. 
  void setRowDirty(FXuint visrow, FXbool aux, FXushort s, FXushort e);

  /// Called by pager when sequence of changes is complete, and widget should redraw.
  void updateDirty();
  
  // Following routines invoked by pager when display needs updating.  Should not be
  // invoked from application code.
  void pagerScroll(FXint dr);
  void pagerResize();
  void pagerUpdate();
  void pagerLineno();
  void pagerNrows();
  
  /// Save to a stream
  virtual void save(FXStream& store) const;

  /// Load from a stream
  virtual void load(FXStream& store);

  /// Destructor
  virtual ~FXPLText();
  };



#endif
