/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                         Cached text row information                           *
*********************************************************************************
*   $Id: FXRowInfo.h,v 1.2 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXRowInfo.h,v $
* Revision 1.2  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.1.1.1  2000/10/29 14:19:49  tekno
* Initial release
*
*
*
********************************************************************************/
#ifndef FXROWINFO_H
#define FXROWINFO_H

/// Attributes of character matrix
typedef FXuchar FXattr;	// Define character attribute type for char matrix.

/// Bit settings in attribute matrix.  Colours are selected based on the highest
/// priority bit which specifies a colour.  Highest priority is 0x01, lowest 0x80.
/// The mapping of bits to colours (and other display attributes) is handled by
/// the FXPLText widget.  The names of the following are based on the expected
/// usage.  Selection is highest priority, since it is one of the most important
/// items.  Informational items are lower down in the list.  Only 8 bits are
/// currently available due to attribute buffer limitation.
/// Key: F=set foreground colour, B=sets background, U=sets underlining, E=embolden,
///   I=italicise (as currently implemented by FXPLText).
enum {
  FXPFATTR_SELECTED  = 0x01	// Current selection F,B
 ,FXPFATTR_ERRORS    = 0x02	// Error messages B
 ,FXPFATTR_SECONDARY = 0x04	// Secondary selection(s) F,B
 ,FXPFATTR_RETENTIVE = 0x08	// Cursor-retentive area (i.e. input fields) B
 ,FXPFATTR_MARKER    = 0x10	// Text markers B
 ,FXPFATTR_INPUT     = 0x20	// (Previous) input F
 ,FXPFATTR_OUTPUT    = 0x40	// (Previous) output F
 ,FXPFATTR_SPECIAL   = 0x80	// Special char e.g. newline; tab or control char expansion F
  };

// Number of extra rows outside visible area for which FXRowInfo entries are to be kept.  This
// is non-zero to allow for anchors/brackets which render outside the character row to which
// they belong.  For example, the caret-style cursor renders one row below its line.  This
// number should be large enough for the biggest reasonable rendering.  If any anchor renders
// larger than this (above or below) then it will not be redrawn correctly.
#define FXROWINFO_EXTRA 2

// Maximum number of anchors (per row) which render in the auxiliary area.  Excess ones are
// ignored.
#define FXROWINFO_MAX_AUX 4

/// Information kept for each char matrix row, plus FXROWINFO_EXTRA either side (top and bottom),
/// plus 1 extra at the end.
struct FXRowInfo {
  FXint		line;	// Text source line number (first line is zero)
  FXint		linerow;// Row within line (0, 1, 2 etc)
  FXint		index;	// Index of first column of this row, in char matrix.
  FXint		pos;	// Character position of first column of this row, with zero
  			// representing first character of first line of text source.
  			// -1 represents an unknown position.
  FXint		rowlen;	// Length of this row.
  FXbool	notlast;// True if this is not the last row of the line.
  FXshort	numaux;	// Number of used entries in following array.
  FXshort	aux[FXROWINFO_MAX_AUX]; 
  			// Anchors which render in the auxiliary area.  The entry in each element
  			// is the index in the pager's visible anchor list.
  };

class FXTextAnchor;

/// Information kept for each visible anchor or bracket.
struct FXAnchorInfo {
  FXTextAnchor * ta;	// The anchor, bracket or derivative.
  struct {
    FXshort r;		// Relative row.  In range [-FXROWINFO_EXTRA..rows+FXROWINFO_EXTRA)
    FXshort c;		// Column number in text area, or -1 if not there.
    FXshort aux;	// Entry number if FXRowInfo::aux, if in aux area, or -1 if not there.
    } s, e;		// One set for each of start and end positions of a bracket, or 'e'
    			// entry not used for anchors.
  FXbool visited;	// Used in list comparison/update.
  };
  
#endif 