/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                        Anchors, brackets and cursors                          *
*********************************************************************************
*   $Id: FXTextAnchor.cpp,v 1.4 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextAnchor.cpp,v $
* Revision 1.4  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.3  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.2  2000/10/30 12:43:24  tekno
* fixed minor problems noted in initial release
*
* Revision 1.1.1.1  2000/10/29 14:19:18  tekno
* Initial release
*
*
*
********************************************************************************/
#ifdef __INSURE__
#define __attribute__(x)
#endif

#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXCharset.h"
#include "FXDict.h"
#include "FXSettings.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXSize.h"
#include "FXPoint.h"
#include "FXRectangle.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXDCWindow.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXScrollbar.h"
#include "FXPLText.h"
#include "FXShell.h"
#include "FXTextSrc.h"
#include "FXPageFormat.h"
#include "FXTextAnchor.h"
#include "FXProcess.h"

/*----------------------------- Base Class ------------------------*/

FXDEFMAP(FXTextAnchor) FXTextAnchorMap[]={
   FXMAPFUNC(SEL_IO_READ,FXTextAnchor::ID_SOCKET,FXTextAnchor::onIORead)
  ,FXMAPFUNC(SEL_COMMAND,FXTextAnchor::ID_REPLACE,FXTextAnchor::onCmdReplace)
  };

// Object implementation
FXIMPLEMENT(FXTextAnchor,FXObject,FXTextAnchorMap,ARRAYNUMBER(FXTextAnchorMap))

FXTextAnchor::FXTextAnchor() {
  owner = NULL;
  attrs = 0;
  data = NULL;
  pos = 0;
  epos = 0;
  visible = FALSE;
  }
  
FXTextAnchor::FXTextAnchor(FXTextSrc * ts, FXuint atr /*= 0*/, void * d /*= NULL*/, FXint sp /*= 0*/) {
  FXASSERT(ts);
  owner = ts;
  attrs = atr;
  data = d;
  pos = sp;
  epos = sp;	// end pos same as start pos for anchor.
  visible = FALSE; 
  owner->addTextAnchor(this);
  }

FXTextAnchor * FXTextAnchor::clone() {
  FXTextAnchor * ta = new FXTextAnchor(owner, attrs, data, pos);
  if (isVisible()) ta->setVisible();
  return ta;
  }

FXint FXTextAnchor::query(const FXTextReplacement * tr, FXint & adj_s, FXint & adj_e) {
  /* The default behaviour for a text anchor is to maintain its relative position unless
     characters on both sides of the anchor were replaced or if the anchor was at the
     start position of the replacement, in which case the anchor
     moves to the position after the last character of the replacement.
  */
  FXint s = tr->startpos;
  FXint e = tr->endpos;
  FXint len = tr->n;
  if (s <= pos && e >= pos)
    adj_s = adj_e = s + len;
  else if (e <= pos)
    adj_s = adj_e = pos + len - (e - s);
  else
    adj_s = adj_e = pos;	// No change.
  return 0;			// Always keep.
  }
  
FXbool FXTextAnchor::isInterested(void) const {
  return FALSE;
  }

FXbool FXTextAnchor::permitsCursorMove(FXint from, FXint to, const FXPLText * pl /*= NULL*/) const {
  return TRUE;
  }

FXbool FXTextAnchor::permitsReplacement(const FXTextReplacement * tr) const {
  return TRUE;
  }

FXbool FXTextAnchor::isVisible(const FXPLText * pl) const {
  return visible;
  }
  
void FXTextAnchor::setVisible(FXbool vis) {
  if (vis != visible) {
    visible = vis;
    if (visible)
      owner->markPlaced(this, NULL, FALSE);
    else
      owner->markRemoved(this, NULL, FALSE);
    }
  }
  
FXbool FXTextAnchor::isRenderable() const { return FALSE; }

FXbool FXTextAnchor::isAuxRenderable() const { return FALSE; }

void FXTextAnchor::render(FXint /*where*/, FXPLText * /*pl*/, FXDCWindow & /*dc*/,
                      FXint /*x*/, FXint /*y*/,  
                      FXint /*x2*/, FXint /*y2*/, FXuint /*flgs*/,
                      FXint /*charwidth*/, FXint /*charheight*/, FXint /*charbase*/) const {
  /* No rendering by default */
  }

FXRectangle FXTextAnchor::bbox(FXint /*where*/, FXPLText * /*pl*/,
                      FXint /*charwidth*/, FXint /*charheight*/, FXint /*charbase*/) const {
  /* Zero dimensioned bounding box by default */
  FXRectangle r;
  r.x = r.y = r.w = r.h = 0;
  return r;
  }

void FXTextAnchor::tellAttrs(FXPageFormat * pf, FXint startpos, FXint endpos) const {
  if (epos < startpos || pos > endpos || !isVisible(pf->getPLText()))
    return;	// We are not visible
  pf->markUpdated(this, FXMAX(pos, startpos), FXMIN(epos, endpos), TRUE, FALSE); 
  }
  
void FXTextAnchor::move(FXint adj_s, FXint adj_e, FXbool temp) {
  FXASSERT(adj_s == adj_e);	// Since this is an anchor, not a bracket.
  if (adj_s != pos) {
    owner->markRemoved(this, NULL, TRUE);	// Tell owner to broadcast removal at old location
    pos = epos = adj_s;
    owner->markPlaced(this, NULL, temp);	// ...then replace at new location
    }
  }

void FXTextAnchor::setAttrs(FXuint atr) {
  FXuint chg = atr ^ attrs;
  if (!chg) return;
  FXbool move = !!(chg & (ANCHOR_S_ATEND|ANCHOR_E_ATEND));
  if (move)
    owner->markRemoved(this, NULL, TRUE);	// Tell owner to broadcast removal at old location
  attrs = atr;
  owner->markPlaced(this, NULL, FALSE);	// Tell owner updated with or without moving.
  }
  
FXuint FXTextAnchor::getMutex() const { return 0; }

FXColor FXTextAnchor::getColor() const {
  return FXRGB(0,0,0);	// Return black as a default.
  }

void FXTextAnchor::setColor(FXColor c) {
  // Do nothing in base class
  }

long FXTextAnchor::onCmdReplace(FXObject* sender,FXSelector,void* ptr){
  FXTextReplacement * tr = (FXTextReplacement *)ptr;
  if (!tr) return 1;
  // Default action is to pass this request on to the owner.  He will call back on
  // us to see if it's alright.
  return !owner->replace(pos + tr->startpos, pos + tr->endpos, 
                        tr->text, tr->n, tr->defer);
  return 1;
  }

long FXTextAnchor::onIORead(FXObject* sender,FXSelector,void* ptr){
  FXInputHandle fd = (FXInputHandle)ptr;
  // Default action is to append the available data to the end of this anchor just
  // like a normal text insertion.  We do up to 4K at a time, for want of a more
  // scientific determination.  If the last char read is NL, then this bracket clones
  // itself, then moves to end of text.  The owner's transaction log is reset if this
  // object does not allow editing, since it does not make sense to allow the user
  // to undo a change made by external sources.
  FXchar buf[4096];
  FXint rc;
  FXint ss = pos;
  FXint ee = epos;
  FXint len = 0;
  FXchar lc = 0;
  
  //printf("FXTextAnchor::onIORead fd=%d\n", fd);
  do {
    rc = read(fd, buf, sizeof(buf));
    if (rc > 0) {
      len += rc;
      owner->replace(ee, ee, buf, rc, TRUE);
      ee += rc;
      lc = buf[rc-1];
      }
  } while (rc > 0);
  owner->replace(0, 0, 0, 0, FALSE);
  if (attrs & ANCHOR_REPEL_CURSOR)
    owner->resetLog();
  move(ss, ee, FALSE);
  if (rc < 0) {
    if (errno && errno != EAGAIN) {
      // Subprocess terminated is usual cause for error.
      //perror("onIORead");
      if (owner->getProcess())
        owner->getProcess()->disconnectHandle(fd, INPUT_READ);
      }
    }
  //if (lc == '\n' && (len = owner->length()-1) > epos) {
  //  if (pos < epos) clone();
  //  move(len, len, FALSE);
  //  }
  return 1;
  }

void FXTextAnchor::save(FXStream& /*store*/) const {
  }

void FXTextAnchor::load(FXStream& /*store*/) {
  }

FXTextAnchor::~FXTextAnchor() {
  setVisible(FALSE);
  owner->removeTextAnchor(this); 
  owner = (FXTextSrc *)-1;
  data = (void *)-1;
  //printf("Anchor chain broke - ship adrift!\n");
  }



/*----------------------------- Cursor Class ------------------------*/

// Object implementation
FXDEFMAP(FXTextCursor) FXTextCursorMap[]={
  FXMAPFUNC(SEL_TIMEOUT, FXTextCursor::ID_FLASHTIMER, FXTextCursor::onTimer)
  };

FXIMPLEMENT(FXTextCursor,FXTextAnchor,FXTextCursorMap,ARRAYNUMBER(FXTextCursorMap))

FXTextCursor::FXTextCursor() : FXTextAnchor() {
  }
  
FXTextCursor::FXTextCursor(FXTextSrc * ts, FXPLText * pl, FXuint atr /*= 0*/, void * d /*= NULL*/, 
               FXint sp /*= 0*/, FXuint rstyle /*= TEXTCURSOR_IBEAM*/) :
               FXTextAnchor(ts, atr, d, sp), pltext(pl), style(rstyle) {
  FXASSERT(pl);	// Must give widget for text cursor.
  hbeam = 4;
  mutex = MUTEX_CURSOR; 
  visible = TRUE;
  attrs |= ANCHOR_UNDERLAY;
  flashtimer = NULL;
  on_style = style;
  flashon = flashoff = 0;
  }

FXuint FXTextCursor::getMutex() const { return mutex; }

FXbool FXTextCursor::isVisible(const FXPLText * pl) const {
  return visible && (!pl || pl == pltext); 
  }
  
void FXTextCursor::setVisible(FXbool vis) {
  if (vis != visible) {
    visible = vis;
    if (vis)
      owner->markPlaced(this, pltext->getPageFormat(), FALSE);
    else
      owner->markRemoved(this, pltext->getPageFormat(), FALSE); 
    }
  }
  
FXbool FXTextCursor::isRenderable() const { return TRUE; }

FXbool FXTextCursor::isAuxRenderable() const { return TRUE; }

void FXTextCursor::render(FXint where, FXPLText * pl, FXDCWindow & dc, 
                      FXint x, FXint y, 
                      FXint x2, FXint y2, FXuint flgs,
                      FXint charwidth, FXint charheight, FXint charbase) const {
  if (pl != pltext || !visible)
    return;	// Only render into our widget, and when visible
  
  dc.setForeground(color);
  if (!where) switch (style) {
    // Main text area
    case TEXTCURSOR_IBEAM:
      dc.setLineStyle();
      dc.setLineCap();
      dc.setLineWidth();
      dc.drawLine(x-hbeam+1, y-charbase, x+hbeam, y-charbase);
      dc.drawLine(x-hbeam+1, y+charheight-charbase-1, x+hbeam, y+charheight-charbase-1);
      dc.drawLine(x, y-charbase, x, y+charheight-charbase-1);
      dc.drawLine(x+1, y-charbase, x+1, y+charheight-charbase-1);
      break;
    case TEXTCURSOR_TRIANGLE:
    case TEXTCURSOR_CARET:
      {
      dc.setFillStyle();
      FXPoint pt[4];
      pt[0].x = x-charwidth; pt[0].y = y+charheight;
      pt[1].x = x+charwidth; pt[1].y = pt[0].y;
      pt[2].x = x; pt[2].y = y;
      if (style == TEXTCURSOR_TRIANGLE) {
      	pt[3].x = pt[0].x;
      	pt[3].y = pt[0].y;
        dc.drawLines(pt, 4);
        break;
        }
      dc.fillPolygon(pt, 3); 
      // Draw contrasting border to maintain visibility under all conditions
      dc.setForeground(darkerColor);
      dc.drawLines(pt, 3);
      dc.setForeground(lighterColor);
      dc.drawLine(pt[0].x, pt[0].y, pt[2].x, pt[2].y);
      }
      break;
    case TEXTCURSOR_BLOCK:
      dc.setFillStyle();
      dc.fillRectangle(x,y-charbase,charwidth,charheight);
      dc.setForeground(darkerColor);
    case TEXTCURSOR_FRAME:
      dc.drawLine(x+charwidth,y-charbase,x+charwidth,y-charbase+charheight);
      dc.drawLine(x,y-charbase+charheight,x+charwidth,y-charbase+charheight);
      if (style == TEXTCURSOR_BLOCK)
        dc.setForeground(lighterColor);
      dc.drawLine(x,y-charbase,x,y-charbase+charheight);
      dc.drawLine(x,y-charbase,x+charwidth,y-charbase);
      break;
    }
  else {
    // Aux area.  Paint right-pointing triangle.
    dc.setFillStyle();
    FXPoint pt[3];
    pt[0].x = x; pt[0].y = y-charbase;
    pt[1].x = x+charwidth; pt[1].y = y-charbase/2;
    pt[2].x = x; pt[2].y = y; 
    dc.fillPolygon(pt, 3); 
    }
  }

FXRectangle FXTextCursor::bbox(FXint where, FXPLText * pl,
                      FXint charwidth, FXint charheight, FXint charbase) const {
  FXRectangle r;
  r.x = r.y = r.w = r.h = 0;
  if (pl != pltext)
    return r;	// Only render into our widget
  if (!where) switch (style) {
    case TEXTCURSOR_IBEAM:
      r.x = -hbeam+1;
      r.w = 2*hbeam;
      r.y = -charbase;
      r.h = charheight+1;
      break;
    case TEXTCURSOR_TRIANGLE:
    case TEXTCURSOR_CARET:
      r.x = -charwidth;
      r.w = charwidth*2+1;
      r.h = charheight+1;
      break;
    case TEXTCURSOR_FRAME:
    case TEXTCURSOR_BLOCK:
      r.y = -charbase;
      r.h = charheight+1;
      r.w = charwidth+1;
      break;
    }
  else {
    r.w = charwidth+1;
    r.y = -charbase;
    r.h = charbase+1; 
    }
  return r;
  }
  
void FXTextCursor::move(FXint adj_s, FXint adj_e, FXbool temp) {
  // Cursor movement checks interested brackets to see whether move is permitted.
  FXTextAnchor * ta = owner->getEnclosingBracket(adj_s, adj_s);
  if (ta && !ta->permitsCursorMove(pos, adj_s, pltext))
    return;
  FXTextAnchor::move(adj_s, adj_e, temp);
  }


void FXTextCursor::setAttrs(FXuint atr) {
  FXuint chg = atr ^ attrs;
  if (!chg) return;
  FXbool move = !!(chg & ANCHOR_S_ATEND);
  if (move)
    owner->markRemoved(this, pltext->getPageFormat(), TRUE);
  attrs = atr;
  owner->markPlaced(this, pltext->getPageFormat(), FALSE);
  }

void FXTextCursor::setColor(FXColor c) {
  color = c; 
  lighterColor = makeHiliteColor(c);
  darkerColor = makeShadowColor(c);
  owner->markRemoved(this, pltext->getPageFormat(), FALSE);
  owner->markPlaced(this, pltext->getPageFormat(), FALSE); // Tell owner updated without moving.
  }

FXColor FXTextCursor::getColor() const {
  return color;	// Return basic rendering colour.
  }

FXApp * FXTextCursor::getApp() const {
  return owner->getApp();
  }


// Set the style.  Simple, non-flashing.
void FXTextCursor::setStyle(FXuint s) {
  flash(FALSE, FALSE);	// 2nd parm must be false to prevent recurse
  if (style != s) {
    owner->markRemoved(this, pltext->getPageFormat(), FALSE);
    style = on_style = s;
    owner->markPlaced(this, pltext->getPageFormat(), FALSE);
    }
  }
  
// Set fancy style.  On/off times are in ms.  If either is zero, then equivalent to setStyle(onstyle).
void FXTextCursor::setStyle(FXuint ontime, FXuint offtime, FXuint onstyle, FXuint offstyle /*= TEXTCURSOR_INVISIBLE*/) {
  off_style = offstyle;
  if (!ontime || !offtime) {
    setStyle(onstyle);
    return;
    }
  on_style = onstyle;
  flashon = ontime;
  flashoff = offtime;
  flash(TRUE, FALSE);
  }

// Start or stop flashing
void FXTextCursor::flash(FXbool on /*= TRUE*/, FXbool setstyle /*=TRUE*/) {
  if (!on && flashtimer) {
    if (getApp()) getApp()->removeTimeout(flashtimer);
    flashtimer = NULL;
    }
  else if (on && !flashtimer) {
    if (!flashon || !flashoff) flashon = 750, flashoff = 250;
    flashtimer = getApp() ? getApp()->addTimeout(flashon, this, ID_FLASHTIMER) : NULL;
    }
  if (!on && setstyle)
    setStyle(on_style);
}

// Flash timer
long FXTextCursor::onTimer(FXObject* sender,FXSelector,void*){
  FXuint nexttime;
  FXint nextstyle;
  if (style == on_style) {
    nexttime = flashoff;
    nextstyle = off_style;
    }
  else {
    nexttime = flashon;
    nextstyle = on_style;
    }
  owner->markRemoved(this, pltext->getPageFormat(), FALSE);
  style = nextstyle;
  owner->markPlaced(this, pltext->getPageFormat(), FALSE);
  
  flashtimer = getApp() ? getApp()->addTimeout(nexttime, this, ID_FLASHTIMER) : NULL;
  return 1;
  }
  


void FXTextCursor::save(FXStream& /*store*/) const {
  }

void FXTextCursor::load(FXStream& /*store*/) {
  }

FXTextCursor::~FXTextCursor() {
  pltext = (FXPLText *)-1;
  }


/*----------------------------- Text Bracket Base Class ------------------------*/

// Object implementation
FXIMPLEMENT(FXTextBracket,FXTextAnchor,NULL,0)

FXTextBracket::FXTextBracket() : FXTextAnchor() {
  }
  
FXTextBracket::FXTextBracket(FXTextSrc * ts, FXuint atr /*= 0*/, void * d /*= NULL*/, 
                             FXint sp /*= 0*/, FXint ep /*= 0*/ ) :
               FXTextAnchor(ts, atr, d, sp) {
  epos = ep;
  }

FXTextAnchor * FXTextBracket::clone() {
  FXTextBracket * ta = new FXTextBracket(owner, attrs, data, pos, epos);
  if (isVisible()) ta->setVisible();
  return ta;
  }

FXint FXTextBracket::query(const FXTextReplacement * tr, FXint & adj_s, FXint & adj_e) {
  /* The default behaviour for a text bracket is:
      . delete self if all chars bracketed are replaced.
      . encompass replacement if entirely within
      . withdraw if partial replacement
      . otherwise stay put (relatively).
     Also, if the bracket would collapse to zero length, it gets deleted.
  */
  FXint s = tr->startpos;
  FXint e = tr->endpos;
  FXint len = tr->n;
  FXint delta = len - (e - s);
  adj_s = pos;
  adj_e = epos;
  
  if (e <= pos) {
    // Entirely before
    adj_s += delta;
    adj_e += delta;
    }
  else if (s >= epos) {
    // Entirely after
    }
  else if (s < pos && e > epos)
    return 1;			// Delete if completely replaced
  else if (s >= pos && e <= epos) {
    // Completely within
    adj_e += delta;
    }
  else if (e <= epos) {
    // Overlapping start but not end
    adj_s = e;
    adj_e += delta;
    }
  else if (s >= pos) {
    // Overlapping end but not start
    adj_e = s;
    }
  
  if (adj_e < adj_s || adj_e == adj_s && !(attrs & ANCHOR_NO_DISAPPEAR))
    return 1;			// Collapsed to zero (or -ve) length, delete.
  return 0;			// Else keep.
  }

void FXTextBracket::move(FXint adj_s, FXint adj_e, FXbool temp) { 
  FXASSERT(adj_s <= adj_e);
  if (adj_s != pos || adj_e != epos) {
    // It's inefficient to move a bracket by removing then replacing it, since it sets
    // too many dirty chars in the pager.  Instead, use the sub-range forms of markRemoved()
    // so that only the appropriate positions have their attributes changed.
    // Compute bits which are removed (zero, one or two stretches).
    if (adj_s > pos && adj_s < epos && adj_s < adj_e)
      owner->markRemoved(this, NULL, TRUE, pos, adj_s);
    if (adj_e < epos && adj_e > pos && adj_s < adj_e)
      owner->markRemoved(this, NULL, TRUE, adj_e, epos);
    if (adj_s == adj_e || adj_e <= pos || adj_s >= epos)
      owner->markRemoved(this, NULL, TRUE);	// Complete removal.
    pos = adj_s;
    epos = adj_e;
    owner->markPlaced(this, NULL, temp);	// ...then replace at new location.
    }
  }



void FXTextBracket::save(FXStream& /*store*/) const {
  }

void FXTextBracket::load(FXStream& /*store*/) {
  }


/*----------------------------- Text Selection Bracket ------------------------*/

// Object implementation
FXIMPLEMENT(FXTextSelection,FXTextBracket,NULL,0)

FXTextSelection::FXTextSelection() : FXTextBracket() {
  }
  
FXTextSelection::FXTextSelection(FXTextSrc * ts, FXuint atr /*= 0*/, void * d /*= NULL*/, 
                             FXint sp /*= 0*/, FXint ep /*= 0*/ ) :
               FXTextBracket(ts, atr, d, sp, ep) {
  visible = TRUE;	// Defaults to visible on construct
  }



/*----------------------------- Input Bracket ------------------------*/

FXDEFMAP(FXInputBracket) FXInputBracketMap[]={
   FXMAPFUNC(SEL_IO_WRITE,FXTextAnchor::ID_SOCKET,FXInputBracket::onIOWrite)
  };

// Object implementation
FXIMPLEMENT(FXInputBracket,FXTextBracket,FXInputBracketMap,ARRAYNUMBER(FXInputBracketMap))

FXInputBracket::FXInputBracket() : FXTextBracket() {
  }
  
FXInputBracket::FXInputBracket(FXTextSrc * ts, FXint stream, FXuint atr /*= 0*/, void * d /*= NULL*/, 
          FXint sp /*= 0*/, FXint ep /*= 0*/ ) :
               FXTextBracket(ts, atr, d, sp, ep) {
  streamno = stream;
  }

long FXInputBracket::onIOWrite(FXObject* sender,FXSelector,void* ptr){
  FXInputHandle fd = (FXInputHandle)ptr;
  FXint p = pos;
  FXint s;
  FXint len, rc;
  const FXchar * buf;
  //printf("FXInputBracket::onIOWrite fd=%d\n", fd);
  do {
    s = p;
    len = owner->getBuf(s, buf);
    if (len > epos-p) len = epos-p;
    if (!len) break;
    rc = write(fd, buf, len);
    //printf("  Wrote `%*.*s', len=%d, rc=%d\n", len, len, buf, len, rc);
    p += rc;
    if (rc < 0) {
      //perror("onIOWrite");
      owner->getProcess()->disconnect(streamno, INPUT_WRITE);
      fd = -1;
      break;
      }
  } while (rc == len);
  if (p == epos)
    owner->getProcess()->hold(streamno, INPUT_WRITE);
  len = p - pos;	// # chars written
  owner->replace(pos, pos+len, NULL, 0);
  return 1;
  }

FXint FXInputBracket::query(const FXTextReplacement * tr, FXint & adj_s, FXint & adj_e) {
  /* Behaviour of input bracket is slightly modified from the default, in that it tends
     to expand more e.g. if insert at start position, start stays put encompassing the
     inserted character(s).
  */
  FXint s = tr->startpos;
  FXint e = tr->endpos;
  FXint len = tr->n;
  FXint delta = len - (e - s);
  adj_s = pos;
  adj_e = epos;
  
  //printf("FXInputBracket::query pos=%d, epos=%d, s=%d, e=%d, len=%u\n", pos,epos,s,e,len);
  
  if (e < pos || e == pos && !tr->from_widget) {
    // Entirely before
    adj_s += delta;
    adj_e += delta;
    }
  else if (s > epos || s == epos && !tr->from_widget) {
    // Entirely after
    }
  else if (s < pos && e > epos) {
    adj_s = s;
    adj_e = s + len;
    }
  else if (s >= pos && e <= epos) {
    // Completely within
    adj_e += delta;
    }
  else if (e <= epos) {
    // Overlapping start but not end
    adj_s = s;
    adj_e += delta;
    }
  else if (s >= pos) {
    // Overlapping end but not start
    adj_e = s + len;
    }
  
  //printf("... adj_s=%d, adj_e=%d\n", adj_s, adj_e);
  return 0;
  }


FXbool FXInputBracket::isInterested(void) const {
  return TRUE;	// Interested in cursor movement and text replacement.
  }
  
FXbool FXInputBracket::permitsReplacement(const FXTextReplacement * tr) const {
  // Replacement allowed.  Single char insertion of '\n' at end causes us to send
  // our entire contents (+NL) to the socket.  This is done indirectly, by setting
  // a writable input (which will cause onIOWrite() to get called).
  if (streamno < 0) return TRUE;
  if (tr->startpos == tr->endpos &&
      tr->n == 1 && tr->text[0] == '\n') {
    owner->getProcess()->resume(streamno, INPUT_WRITE);
    owner->resetOutput();
    return FALSE;	// Don't insert the \n
    }
  if (tr->startpos < pos || tr->endpos > epos)
    return FALSE;
  return TRUE;
  }

  
  
/*----------------------------- Output Bracket ------------------------*/

// Object implementation
FXIMPLEMENT(FXOutputBracket,FXTextBracket,NULL, 0)

FXOutputBracket::FXOutputBracket() : FXTextBracket() {
  }
  
FXOutputBracket::FXOutputBracket(FXTextSrc * ts, FXuint atr /*= 0*/, void * d /*= NULL*/, 
                             FXint sp /*= 0*/, FXint ep /*= 0*/ ) :
               FXTextBracket(ts, atr, d, sp, ep) {
  attrs |= ANCHOR_NO_DISAPPEAR;
  }

FXTextAnchor * FXOutputBracket::clone() {
  FXOutputBracket * ta = new FXOutputBracket(owner, attrs, data, pos, epos);
  if (isVisible()) ta->setVisible();
  return ta;
  }


FXbool FXOutputBracket::isInterested(void) const {
  return TRUE;	// Interested in cursor movement and text replacement.
  }
  

FXbool FXOutputBracket::permitsCursorMove(FXint from, FXint to, const FXPLText * pl /*= NULL*/) const {
  //printf("FXOutputBracket::permitsCursorMove from=%d to=%d pos=%d epos=%d attrs=%08X\n",
  //  from, to, pos, epos, attrs);
  return !(attrs & ANCHOR_REPEL_CURSOR) || to <= pos || to >= epos;
  }

FXbool FXOutputBracket::permitsReplacement(const FXTextReplacement * tr) const {
  return !tr->from_widget || (tr->endpos <= pos || tr->startpos >= epos);
  }



