/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                           Text metric mapping                                 *
*********************************************************************************
*   $Id: FXTextMap.cpp,v 1.2 2000/11/01 11:27:10 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextMap.cpp,v $
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:19:41  tekno
* Initial release
*
*
*
********************************************************************************/
#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"

#include "FXTextMap.h"

static FXuint numLevels(FXuint leaves) {
  FXuint v;
  FXuint m;
  
  for (m = 1, v = 1; 
       m < leaves;
       v++, m <<= 1);
  return v;
  }


FXTextMap::FXTextMap(FXuint leaves) {
  ok = TRUE;
  if (leaves < 1)
    leaves = 1;	// Not allowed to have zero leaves.
  N = leaves;
  V = numLevels(N);
  root = (1<<(V-1)) - 1;
  shift = 0;
  treealloc = 1<<V;
  dirty = FALSE;
  FXCALLOC(&tree, FXTextNode, treealloc);
  FXASSERT(tree);
  if (!tree) {
    treealloc = 0;
    N = V = 0;
    ok = FALSE;
    }
  }

FXuint FXTextMap::modulo_Ln(FXTextNode & search) {
  FXuint v = V-1;	// Current level
  FXuint pn = root;	// Current parent node
  FXuint child, hc;
  FXuint Ln = search.Ln;
  
  if (dirty)
    updateRange(lowupd, highupd);
  
  if (Ln >= tree[root].Ln)
    Ln = tree[root].Ln;
    
  search.Pn = 0;
  while (v) {
    child = lowchild(pn, v);
    hc = highchild(pn, v--);
    if (Ln < tree[child].Ln ||
        // always go to low child if high child is zero and not at leaf level...
        (!tree[hc].Ln /*&& v*/))
      pn = child;
    else {
      pn = hc;
      search.Pn += tree[child].Pn;
      Ln -= tree[child].Ln;
      }
    }
  search.Ln = Ln;
  return pn;
  }
  
FXuint FXTextMap::modulo_Pn(FXTextNode & search) {
  FXuint v = V-1;	// Current level
  FXuint pn = root;	// Current parent node
  FXuint child, hc;
  FXuint Pn = search.Pn;
  
  if (dirty)
    updateRange(lowupd, highupd);

  if (Pn > tree[root].Pn)
    Pn = tree[root].Pn;
    
  search.Ln = 0;
  while (v) {
    child = lowchild(pn, v);
    hc = highchild(pn,v--);
    if (Pn < tree[child].Pn ||
        // always go to low child if high child is zero and not at leaf level...
        (!tree[hc].Pn /*&& v*/))
      pn = child;
    else {
      pn = hc;
      search.Ln += tree[child].Ln;
      Pn -= tree[child].Pn;
      }
    }
  search.Pn = Pn;
  return pn;
  }
  
FXuint FXTextMap::root_Ln() {
  if (dirty)
    updateRange(lowupd, highupd);
  return tree[root].Ln;
  }

FXuint FXTextMap::root_Pn() {
  if (dirty)
    updateRange(lowupd, highupd);
  return tree[root].Pn;
  }

FXuint FXTextMap::node_Ln(FXuint node) {
  if (dirty)
    updateRange(lowupd, highupd);
  FXASSERT(node < treealloc);
  return tree[node].Ln;
  }

FXuint FXTextMap::node_Pn(FXuint node) {
  if (dirty)
    updateRange(lowupd, highupd);
  FXASSERT(node < treealloc);
  return tree[node].Pn;
  }

void FXTextMap::updateRange(FXuint lowleaf, FXuint highleaf) {
  FXuint v;
  FXuint pl = lowleaf, ph = highleaf;	// Lowest and highest parent nodes at current level
  FXuint pn;
  FXuint cl, ch;
  
  FXASSERT(lowleaf <= highleaf);
  if (lowleaf > highleaf)
    printf("updateRange: %u %u\n", lowleaf, highleaf);
  FXASSERT(!(lowleaf & 1));
  FXASSERT(!(highleaf & 1));
  FXASSERT(lowleaf < treealloc);
  FXASSERT(highleaf < treealloc);
  
  if (dirty) {
    lowleaf = FXMIN(lowleaf, lowupd);
    highleaf = FXMAX(highleaf, highupd);
    }
  
  if (lowleaf == highleaf) {
    pn = lowleaf;
    for (v = 0; v+1 < V;) {
      pn = parent(pn, v++);
      cl = lowchild(pn, v);
      ch = highchild(pn, v);
      tree[pn].Pn = tree[cl].Pn + tree[ch].Pn;
      tree[pn].Ln = tree[cl].Ln + tree[ch].Ln;
      }
    }
  else for (v = 0; v+1 < V;) {
    pl = parent(pl, v);
    ph = parent(ph, v++);
    for (pn = pl; pn <= ph; pn = highpeer(pn, v)) {
      cl = lowchild(pn, v);
      ch = highchild(pn, v);
      tree[pn].Pn = tree[cl].Pn + tree[ch].Pn;
      tree[pn].Ln = tree[cl].Ln + tree[ch].Ln;
      }
    }
  dirty = FALSE;
  
  // Testing
#if (0)
  printf("Map %08X now: V=%u N=%u root=%u\n", (FXuint)this, V, N, root); 
  for (pn = 0; pn < treealloc-1; pn++) {
    v = pn;
    while (v & 1) printf("  "), v >>= 1;
    printf("%u:%u%s\n", tree[pn].Pn, tree[pn].Ln, pn == root ? "  <----" : ""); 
    }
#endif
  }
  

FXbool FXTextMap::insertLeaves(FXuint beforeLeaf, FXuint num /*= 1*/, 
                               const FXTextNode * newLeaves /*= 0*/) {
  FXuint newV;	// New number of levels
  FXuint v;
  FXuint pn, p;
  FXuint d = num<<1;
  
  FXASSERT(!(beforeLeaf & 1));
  FXASSERT(beforeLeaf <= treealloc);
  FXASSERT(beforeLeaf <= (N<<1));
  FXASSERT((FXint)beforeLeaf >= 0);
  FXASSERT((FXint)num >= 0);
  
  if (!num)
    return TRUE;
  
  newV = numLevels(N + num);
  if (newV > V) {
    if (!FXRESIZE(&tree, FXTextNode, 1<<newV))
      // Failed resize
      return FALSE;
    memset(tree + treealloc, 0, ((1<<newV) - treealloc)*sizeof(FXTextNode));
    treealloc = 1<<newV;
    
    // Propagate old root up to new root
    for (v = V-1, p = root; v < newV-1; v++) {
      pn = parent(p, v);
      memcpy(tree + pn, tree + p, sizeof(*tree));
      p = pn;
      }
    root = pn;
    V = newV;
    }
  
  /* Do the insertion.  Be careful, we are going backwards using unsigned arithmetic.
     (Always a trap for the unwary!) */
  for (p = N<<1; p > beforeLeaf; p = pn) {
    pn = lowpeer(p);
    memcpy(tree+pn+d, tree+pn, sizeof(*tree));
    }
  if (newLeaves)
    for (p = beforeLeaf, pn = 0; pn < num; pn++, p = highpeer(p))
      memcpy(tree+p, newLeaves+pn, sizeof(*tree));
  else
    for (p = beforeLeaf, pn = 0; pn < num; pn++, p = highpeer(p))
      memset(tree+p, 0, sizeof(*tree));
  
  N += num;
  //updateRange(beforeLeaf, lowpeer(N<<1)); - defer it
  if (dirty) {
    lowupd = FXMIN(lowupd, beforeLeaf);
    highupd = FXMAX(highupd, lowpeer(N<<1)); 
    }
  else {
    dirty = TRUE;
    lowupd = beforeLeaf;
    highupd = lowpeer(N<<1);
    }
  return TRUE;
  }


void FXTextMap::deleteLeaves(FXuint fromLeaf, FXuint num /*= 1*/ ) {
  FXuint newV;	// New number of levels
  FXuint p;
  FXuint d = num<<1;
  
  FXASSERT(!(fromLeaf & 1));
  FXASSERT(fromLeaf < treealloc);
  FXASSERT(fromLeaf < (N<<1));
  FXASSERT(num <= N);
  FXASSERT((FXint)fromLeaf >= 0);
  FXASSERT((FXint)num >= 0);
  
  if (!num) return;
  
  if ((N<<1)>d) {
    /* Do the deletion.  Note that the memset zeros leaves plus their ancestors. */
    for (p = fromLeaf; p <= lowpeer(N<<1)-d; p = highpeer(p))
      memcpy(tree+p, tree+p+d, sizeof(*tree));
    memset(tree+(N<<1)-d, 0, sizeof(*tree)*d);
    // Include the last+1 node in the update range, unless it is the last leaf in the tree.
    p = N<<1 == treealloc ? lowpeer(N<<1) : N<<1;
    // updateRange(fromLeaf, p); - defer it
    if (dirty) {
      lowupd = FXMIN(lowupd, fromLeaf);
      highupd = FXMAX(highupd, p); 
      }
    else {
      dirty = TRUE;
      lowupd = fromLeaf;
      highupd = p;
      }
    N -= num;
    }
  else {
    // Everything deleted.  One node must remain, and it is zerod.  3 nodes are zerod since
    // the new root may be left one level up (at level 2).
    N = 1;
    memset(tree, 0, FXMIN(3, treealloc)*sizeof(*tree));
    }
  
  // Prune tree if new number of levels is 2 or more less than old, but we keep one empty
  // level.  This provides some hysteresis against thrashing the tree levels.
  newV = numLevels(N);
  if (newV+1 < V) {
    FXRESIZE(&tree, FXTextNode, 1<<newV+1);
    treealloc = 1<<newV+1;
    V = newV+1;
    root = (1<<(V-1)) - 1;
    highupd = FXMIN(highupd, lowpeer(treealloc));
    }
  
  }

void FXTextMap::updateLeaf(FXuint leaf, FXuint Pn, FXuint Ln) {
  FXASSERT(!(leaf & 1));
  FXASSERT(leaf < treealloc);
  FXASSERT(leaf < (N<<1));
  
  // Must not create nodes with zeros.
  FXASSERT((int)Pn > 0);
  FXASSERT((int)Ln > 0);
  
  tree[leaf].Pn = Pn;
  tree[leaf].Ln = Ln;
  if (dirty) {
    lowupd = FXMIN(lowupd, leaf);
    highupd = FXMAX(highupd, leaf); 
    }
  else {
    dirty = TRUE;
    lowupd = highupd = leaf;
    }
  }


FXTextMap::~FXTextMap() {
  if (tree) FXFREE(&tree);
  tree = (FXTextNode *)-1;
  }

