/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                              Text Source                                      *
*********************************************************************************
*   $Id: FXTextSrc.cpp,v 1.3 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXTextSrc.cpp,v $
* Revision 1.3  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:19:02  tekno
* Initial release
*
*
*
********************************************************************************/
#ifdef __INSURE__
#define __attribute__(x)
#endif

#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXObjectList.h"
#include "FXCharset.h"
#include "FXDict.h"
#include "FXFile.h"
#include "FXSettings.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXSize.h"
#include "FXPoint.h"
#include "FXRectangle.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXDCWindow.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXGIFIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXTextField.h"
#include "FXButton.h"
#include "FXToggleButton.h"
#include "FXMenuButton.h"
#include "FXComposite.h"
#include "FXPacker.h"
#include "FXHorizontalFrame.h"
#include "FXMatrix.h"
#include "FXCanvas.h"
#include "FXScrollbar.h"
#include "FXPLText.h"
#include "FXShell.h"
#include "FXTextAnchor.h"
#include "FXTextSrc.h"
#include "FXPageFormat.h"
#include "FXTextBuf.h"
#include "FXTextMap.h"
#include "FXElement.h"
#include "FXTopWindow.h"
#include "FXDialogBox.h"
#include "FXFindDialog.h"
#include "FXMessageBox.h"
#include "FXScrollArea.h"
#include "FXList.h"
#include "FXTreeList.h"
#include "FXComboBox.h"
#include "FXTreeListBox.h"
#include "FXDirBox.h"
#include "FXHeader.h"
#include "FXIconList.h"
#include "FXFileList.h"
#include "FXRecentFiles.h"
#include "FXFileSelector.h"
#include "FXFileDialog.h"
#include "FXProcess.h"
#ifdef POSIX_REGEX
#include <regex.h>
#else
#include "fxregex.h"
#endif

#ifndef WIN32
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#endif

/*-------------------- Base Class -------------------------*/
FXDEFMAP(FXTextSrc) FXTextSrcMap[]={
   FXMAPFUNC(SEL_CHANGED,FXTextSrc::ID_TRACK_TOP,FXTextSrc::onTrackTop)
  };
  

// Object implementation
FXIMPLEMENT(FXTextSrc,FXObject,FXTextSrcMap,ARRAYNUMBER(FXTextSrcMap))

FXTextSrc::FXTextSrc() {
  ok = TRUE;
  modified_since_create = FALSE;
  modified_since_write = FALSE;
  modified_since_autosave = FALSE;
  selection = NULL;
  last_sel_len = 0;
  FXMALLOC(&last_sel, FXchar, FXTEXTSRC_MAX_LAST_SEL_LEN);
  }

FXTextSrc::FXTextSrc(FXApp * a, FXuint opts) :
    delimiters("~.,/\\`'!@#$%^&*()-=+{}|[]\":;<>?") {
  ok = TRUE;
  app = a;
  modified_since_create = FALSE;
  modified_since_write = FALSE;
  modified_since_autosave = FALSE;
  selection = NULL;
  last_sel_len = 0;
  FXMALLOC(&last_sel, FXchar, FXTEXTSRC_MAX_LAST_SEL_LEN);
  }

void FXTextSrc::addPageFormat(FXPageFormat * pf) {
  FXASSERT(pf);
  if (pfs.findf(pf) >= 0) return;	// Ignore if already in list
  pfs.append(pf);
  }

void FXTextSrc::removePageFormat(const FXPageFormat * pf) {
  FXASSERT(pf);
  pfs.remove(pf);
  }

long FXTextSrc::onTrackTop(FXObject*,FXSelector,void*ptr) {
  FXPLText * widget = (FXPLText *)ptr;
  for (FXint i = 0; i < pfs.no(); i++) {
    FXPLText * pl = pfs[i]->getPLText();
    if (pl && pl->getTrack() == widget)
      pl->handle(widget, MKUINT(FXPLText::ID_TRACK_TOP, SEL_CHANGED), NULL);
    }
  return 1;
  }


void FXTextSrc::addTextAnchor(FXTextAnchor * ta) {
  FXASSERT(ta);
  if (anchs.findf(ta) >= 0) return;	// Ignore if already in list
  anchs.append(ta);
  enforceMutex(ta);
  }

void FXTextSrc::removeTextAnchor(const FXTextAnchor * ta) {
  FXASSERT(ta);
  FXASSERT(!ta->getTextSrc() || ta->getTextSrc() == this);
  anchs.remove(ta);
  if (ta == selection)
    selection = NULL;	// Just deleted selection.
  }

void FXTextSrc::removeTextAnchors(FXint n, const FXTextAnchor ** tas) {
  FXASSERT(tas);
  for (FXint i = 0; i < n; i++)
    removeTextAnchor(tas[i]);
  }

FXint FXTextSrc::removeMatchingTextAnchors(FXuint mask, FXuint match) {
  FXint ret;
  for (FXint i = 0; i < getNumTextAnchors(); i++) {
    FXTextAnchor * ta = getTextAnchor(i);
    if ((ta->getAttrs() & mask) == match) {
      delete ta;
      i--;
      ret++;
      }
    }
  return ret;
  }

FXint FXTextSrc::findTextAnchors(FXint sp, FXint ep, FXint n, FXTextAnchor ** tas, 
        FXuint mask, FXuint match, FXbool containing) const {
  FXint nfound = 0;
  for (FXint i = 0; nfound < n && i < getNumTextAnchors(); i++) {
    FXTextAnchor * ta = getTextAnchor(i);
    if ((sp < 0 || (containing ? ta->getPos() <= sp : ta->getPos() == sp)) &&
        (ep < 0 || (containing ? ta->getEpos() >= ep : ta->getEpos() == ep)) &&
        (ta->getAttrs() & mask) == match)
      tas[nfound++] = ta;
    }
  return nfound;
  }

FXTextAnchor * FXTextSrc::getEnclosingBracket(FXint startpos, FXint endpos) {
  FXint minlen = 0x7FFFFFFF;
  FXTextAnchor * minta = 0;
  for (FXint i = 0; i < getNumTextAnchors(); i++) {
    FXTextAnchor * ta = getTextAnchor(i);
    if (ta->isInterested() && ta->getPos() <= startpos && 
        ta->getEpos() >= endpos && ta->getLength() && ta->getLength() < minlen) {
      minlen = ta->getLength();
      minta = ta;
      }
    }
  return minta;
  }


// The effect of the following is to have a single blank line.
#define BASE_CHARS 1	// Chars in base class constant 'line' which is following string...
#define BASE_TEXT "\n"
#define BASE_LINES 1	// Number of 'lines'

const FXString & FXTextSrc::getLineText(FXuint lineno, FXint & pos) const {
  static FXString zero(BASE_TEXT);
  pos = lineno * BASE_CHARS;
  return zero;
  }

FXint FXTextSrc::getText(FXint pos, FXchar * ptr, FXint len) const {
  if (len)
    ptr[0] = 0;
  FXint tlen = length();
  if (pos < 0 || pos >= tlen) return 0;
  FXint i;
  for (i = 0; i < len && pos <= tlen; i++)
    *ptr++ = getChar(pos++);
  return i;
  }

FXint FXTextSrc::getBuf(FXint & pos, const FXchar * & ptr) const {
  ptr = 0;
  if (pos < 0 || pos >= length()) return 0;
  ptr = BASE_TEXT + pos % BASE_CHARS;
  FXint len = BASE_CHARS - pos % BASE_CHARS;
  pos += len;
  return len;
  }

FXint FXTextSrc::getBackwardsBuf(FXint & pos, const FXchar * & ptr) const {
  ptr = 0;
  if (pos <= 0 || pos > length()) return 0;
  ptr = BASE_TEXT;
  pos -= pos % BASE_CHARS;
  return BASE_CHARS;
  }


FXint FXTextSrc::getPosOfLine(FXuint lineno) const {
  if (lineno >= lines())
    return length();
  return lineno*BASE_CHARS;
  }

FXuint FXTextSrc::getLineOfPos(FXint pos) const {
  if (pos < 0) return 0;
  if (pos >= length()) return lines()-1; 
  return pos/BASE_CHARS;
  }

FXuint FXTextSrc::lines() const {
  return BASE_LINES;
  }

FXint FXTextSrc::length() const {
  return BASE_CHARS*lines();   
  }

FXchar FXTextSrc::getChar(FXint pos) const {
  if (pos < 0 || pos >= length()) return 0;
  return BASE_TEXT[pos % BASE_CHARS];
  }
  
FXint FXTextSrc::leftWord(FXint pos) const {
  register FXchar ch;
  if(pos>length()) pos=length();
  if(0<pos){
    ch=getChar(pos-1);
    if(delimiters.has(ch)) return pos-1;
    }
  while(0<pos){
    ch=getChar(pos-1);
    if(delimiters.has(ch)) return pos;
    if(isspace(ch)) break;
    pos--;
    }
  while(0<pos){
    ch=getChar(pos-1);
    if(!isspace(ch)) return pos;
    pos--;
    }
  return 0;
  }
  
FXint FXTextSrc::rightWord(FXint pos) const {
  register FXchar ch;
  if(pos<0) pos=0;
  if(pos<length()){
    ch=getChar(pos);
    if(delimiters.has(ch)) return pos+1;
    }
  while(pos<length()){
    ch=getChar(pos);
    if(delimiters.has(ch)) return pos;
    if(isspace(ch)) break;
    pos++;
    }
  while(pos<length()){
    ch=getChar(pos);
    if(!isspace(ch)) return pos;
    pos++;
    }
  return length();
  }

FXint FXTextSrc::wordStart(FXint pos) const {
  register FXchar c=' ';
  if(pos<=0) return 0; 
  if(pos<length()) c=getChar(pos);
  if(c==' ' || c=='\t'){
    while(0<pos){
      c=getChar(pos-1);
      if(c!=' ' && c!='\t') return pos;
      pos--;
      }
    }
  else if(delimiters.has(c)){
    while(0<pos){
      c=getChar(pos-1);
      if(!delimiters.has(c)) return pos;
      pos--;
      }
    }
  else{
    while(0<pos){
      c=getChar(pos-1);
      if(delimiters.has(c) || isspace(c)) return pos;
      pos--;
      }
    }
  return 0;
  }

FXint FXTextSrc::wordEnd(FXint pos) const {
  register FXchar c=' ';
  if(pos>=length()) return length();
  c=getChar(pos); 
  if(c==' ' || c=='\t'){
    while(pos<length()){
      c=getChar(pos);
      if(c!=' ' && c!='\t') return pos;
      pos++;
      }
    }
  else if(delimiters.has(c)){
    while(pos<length()){
      c=getChar(pos);
      if(!delimiters.has(c)) return pos;
      pos++;
      }
    }
  else{
    while(pos<length()){
      c=getChar(pos);
      if(delimiters.has(c) || isspace(c)) return pos;
      pos++;
      }
    }
  return length(); 
  }



FXint FXTextSrc::replace(FXint startpos, FXint endpos, const FXString & text, FXbool defer /*= FALSE*/, const FXPLText * pltext) {
  return 1; // Operation not supported
  }

FXint FXTextSrc::replace(FXint startpos, FXint endpos, const FXchar * text, FXint n, FXbool defer /*= FALSE*/, const FXPLText * pltext) {
  return 1; // Operation not supported
  }

FXint FXTextSrc::rollback(FXint num) {
  return 0; // No rollback available
  }

FXint FXTextSrc::rollforward(FXint num) {
  return 0; // No rollforward available
  }

FXint FXTextSrc::rollbackPos() const {
  return -1; // No rollback available
  }

FXint FXTextSrc::rollforwardPos() const {
  return -1; // No rollforward available
  }

FXint FXTextSrc::resetLog() { return 0; }

FXint FXTextSrc::write() {
  return -1; // Operation not supported
  }

FXint FXTextSrc::autosave() {
  return -1; // Operation not supported
  }

FXint FXTextSrc::revert() {
  return -1; // Operation not supported
  }

FXbool FXTextSrc::isModifiable() const {
  return FALSE;
  }

static inline FXint memcasecmp(const FXchar * a, const FXchar * b, FXint len) {
  // Compare, converting a to uppercase (the caller is assumed to have set b to uppercase).
  // Returns 0 if same, else 1.
  for (register FXint i = 0; i < len; i++)
    if (toupper(a[i]) != b[i])
      return 1;
  return 0;
  }
  

FXint FXTextSrc::findLiteralForward(FXFindData * fd, FXint pos, FXint epos, 
          const FXchar * ss, FXint sslen, FXbool case_sens, FXbool whole_word) const {
  /* Forward literal search, with epos > pos */
  FXASSERT(epos > pos);
  FXint sp = pos;
  const FXchar * p, * q;
  FXint np = pos;
  FXint bl;
  FXchar c1 = ss[0];	// First char to look for
  FXint ctm;		// Available chars to compare in buffer
  
  // This looks frightening because of the non-contiguous nature of the buffer.  Most reasonable
  // buffers have large contiguous chunks, so efficiency is still high.  Also, we almost double
  // the code by cracking out the case sensitivity test to near the outermost level (rather than
  // inside the inner loop).  Should I have trusted the optimiser to do this?
  for (;;) {
    bl = getBuf(np, p);
    if (!bl) break;
    ctm = FXMIN(sslen, bl);
    if (case_sens) for (q = p; ctm; q++) {
      if (*q == c1 && !memcmp(q, ss, ctm)) {
        // 1st part matches.  Check remainder.
        FXint rem = sslen - ctm;
        const FXchar * s = ss + ctm;
        FXint xp = np;
        const FXchar * r;
        while (rem) {
          FXint xl = getBuf(xp, r);
          if (!xl) return -1;
          FXint xtm = FXMIN(rem, xl);
          if (!memcmp(r, s, xtm)) {
            rem -= xtm;
            s += xtm;
            }
          else
            break;
          }
        if (!rem) {
          // Eureka
          return sp + (q-p);
          }
        }
      bl--;
      if (bl < ctm) ctm = bl;
      if (++pos == epos)
        return -1;
      }
    else for (q = p; ctm; q++) {
      if (toupper(*q) == c1 && !memcasecmp(q, ss, ctm)) {
        // 1st part matches.  Check remainder.
        FXint rem = sslen - ctm;
        const FXchar * s = ss + ctm;
        FXint xp = np;
        const FXchar * r;
        while (rem) {
          FXint xl = getBuf(xp, r);
          if (!xl) return -1;
          FXint xtm = FXMIN(rem, xl);
          if (!memcasecmp(r, s, xtm)) {
            rem -= xtm;
            s += xtm;
            }
          else
            break;
          }
        if (!rem) {
          // Eureka
          return sp + (q-p);
          }
        }
      bl--;
      if (bl < ctm) ctm = bl;
      if (++pos == epos)
        return -1;
      }
    sp = np;
    }
  return -1;
  }
  

FXint FXTextSrc::findLiteralBackward(FXFindData * fd, FXint pos, FXint epos, 
          const FXchar * ss, FXint sslen, FXbool case_sens, FXbool whole_word) const {
  /* Backward literal search, with epos < pos */
  FXASSERT(epos < pos);
  const FXchar * p, * q;
  FXint np = pos;
  FXint op;
  FXint bl;
  FXchar c1 = ss[0];	// First char to look for
  FXint ctm;		// Available chars to compare in buffer
  
  for (;;) {
    op = np;
    bl = getBackwardsBuf(np, p);
    if (!bl) break;
    ctm = FXMIN(sslen, np+bl-op+1);
    if (case_sens) for (q = p+op-np-1; q >= p; q--) {
      if (*q == c1 && !memcmp(q, ss, ctm)) {
        // 1st part matches.  Check remainder.
        FXint rem = sslen - ctm;
        const FXchar * s = ss + ctm;
        FXint xp = np+bl;
        const FXchar * r;
        while (rem) {
          FXint xl = getBuf(xp, r);
          if (!xl) break;
          FXint xtm = FXMIN(rem, xl);
          if (!memcmp(r, s, xtm)) {
            rem -= xtm;
            s += xtm;
            }
          else
            break;
          }
        if (!rem) {
          // Eureka
          return np + (q-p);
          }
        }
      if (ctm < sslen) ctm++;
      if (--pos == epos)
        return -1;
      }
    else for (q = p+op-np-1; q >= p; q--) {
      if (toupper(*q) == c1 && !memcasecmp(q, ss, ctm)) {
        // 1st part matches.  Check remainder.
        FXint rem = sslen - ctm;
        const FXchar * s = ss + ctm;
        FXint xp = np+bl;
        const FXchar * r;
        while (rem) {
          FXint xl = getBuf(xp, r);
          if (!xl) break;
          FXint xtm = FXMIN(rem, xl);
          if (!memcasecmp(r, s, xtm)) {
            rem -= xtm;
            s += xtm;
            }
          else
            break;
          }
        if (!rem) {
          // Eureka
          return np + (q-p);
          }
        }
      if (ctm < sslen) ctm++;
      if (--pos == epos)
        return -1;
      }
    }
  return -1;
  }
  

  
FXint FXTextSrc::find(FXFindData * fd, FXint pos, FXint epos, 
                    FXint & len, FXWindow * widget) const {
  /* Base class text search facility.  Supports all search modes, however derived classes
     may re-implement those where they can do the job more efficiently.  For some search modes,
     the corresponding specialised function is called.
     
     For regular expression searches, case insensitivity is not always implemented by
     the underlying library so we retrieve one line at a time and convert it to all
     uppercase.  The regex expression is also converted to uppercase.
  */
  FXchar * ss = NULL;	// Search string
  FXint    sslen;	// Search string length
  FXint    sp = -1;	// Search result
  FXint    tlen;	// Cached length of buffer
  FXint    linew = 1;	// Line window: number of lines buffered.  For regex this is always 1,
  			// for others it is the number of newlines in the find string, plus 1.
  FXchar * lbuf = NULL;	// Line window buffer: either points directly to text buffer, or
  			// to allocated buffer if text buffer is not contiguous or doing
  			// case-insensitive or regex search - regex needs null termination :-(
  FXint    bufsz = 0;	// Current allocated lbuf dimension.
  FXuint   line;	// Current 1st line of line buffer
  FXint    ilen;        // Length of 1st line
  FXint    pos1; 	// Corresponding position
  FXbool   wrap;	// True if search can wrap around
  FXbool   first;	// True if testing first search line
  FXint    delta;	// Initial search position on first line
#ifdef POSIX_REGEX
  regex_t * re;		// Compiled regex.
  regex_t   rei;	// An instance.
  regmatch_t pmatch[1];
#else
  fxregexp * re;
#endif
  int       rc;		// Result from compiling regex.
  FXbool    do_rx = fd->search_type == FXFindData::FD_REGEX;
  
  //printf("FXTextSrc::find `%*.*s' from %d to %d\n", fd->flen, fd->flen, fd->find, pos, epos);
  
  // Pre-process find string into suitable form
  FXint i;
  switch (fd->search_type) {
    case FXFindData::FD_LITERAL:
    case FXFindData::FD_C_ESC:
      if (!fd->case_sens) {
        FXMALLOC(&ss, FXchar, fd->flen);
        for (i = 0; i < fd->flen; i++)
          ss[i] = toupper(fd->find[i]);
        }
      else
        ss = fd->find;
      sslen = fd->flen;
      for (i = 0; i < sslen; i++)
        if (ss[i] == '\n')
          linew++;
      break;
    case FXFindData::FD_REGEX:
      FXMALLOC(&ss, FXchar, fd->flen+1);
      if (!fd->case_sens) {
        for (i = 0; i < fd->flen; i++)
          ss[i] = toupper(fd->find[i]);
        }
      else
        memcpy(ss, fd->find, fd->flen);
      ss[fd->flen] = 0;	// Null terminate for regex compilers.
      sslen = fd->flen;
      // Compile regex.
#ifdef POSIX_REGEX
      re = &rei;
      if (rc = regcomp(re, ss, 0)) {
        char str[80];
        regerror(rc, re, str, sizeof(str));
        regfree(re);
        //printf("regex: regcomp rc=%d, %s\n", rc, str);
        if (widget)
          FXMessageBox::error(widget, MBOX_OK,
            "Regular expression error", "%s", str);
        sp = -2;
        goto bye;
        }
#else
      re = fxregcomp(ss);
      if (!re) {
        if (widget)
          FXMessageBox::error(widget, MBOX_OK,
            "Regular expression error", "Syntax error");
        sp = -2;
        goto bye;
        }
#endif
      break;
    default:
      FXASSERT(0);	// Bad search args
      return -1;
    }
  
  /* Perform search.  For non-multi-line search, a single line is retrieved at a time.
     Line-by-line is always done for regex search.  Only the first matching regex
     on a line is returned for backward regex searches, since native backward search is
     not implemented by the underlying regex library.
     
     For non-regex searches, we search directly in the text buffer which is much more
     efficient than retrieving lines into a buffer.  However, this is complicated by the
     necessity of handling non-contiguous buffers (as well as case-insensitivity).
  */
  
  tlen = length();
  if (pos < 0) pos = tlen-1;
  else if (pos >= tlen) pos = 0;
  if (epos < 0) epos = tlen-1;
  else if (epos >= tlen) epos = 0;
  wrap = fd->fwd ? epos <= pos : epos >= pos;
  
  if (!fd->whole_word && !do_rx) {
    len = sslen;
    if (fd->fwd) {
      if (wrap) {
        if ((sp = findLiteralForward(fd, pos, tlen, ss, sslen, fd->case_sens, fd->whole_word)) >= 0)
          return sp;
        return epos ? findLiteralForward(fd, 0, epos, ss, sslen, fd->case_sens, fd->whole_word) : -1;
        }
      return findLiteralForward(fd, pos, epos, ss, sslen, fd->case_sens, fd->whole_word);
      }
    else {
      if (wrap) {
        if ((sp = findLiteralBackward(fd, pos, 0, ss, sslen, fd->case_sens, fd->whole_word)) >= 0)
          return sp;
        return findLiteralBackward(fd, tlen, epos, ss, sslen, fd->case_sens, fd->whole_word);
        }
      return findLiteralBackward(fd, pos, epos, ss, sslen, fd->case_sens, fd->whole_word);
      }
    }

  first = TRUE;
  
  sp = pos;
  line = getLineOfPos(sp);
  pos1 = getPosOfLine(line);
  ilen = getPosOfLine(line+1) - pos1;
  delta = do_rx && !fd->fwd ? 0 : sp-pos1;
  
  for (;;) {
    FXuint last = line + linew; 
    FXint reqbuf = linew == 1 ? ilen : getPosOfLine(last) - pos1;
    FXint cbuf;
    FXbool maybe = FALSE;
    
    if (!do_rx && reqbuf < sslen) goto nextline; // Can't possibly match (too small)
    
    if (reqbuf+1 > bufsz) {
      FXRESIZE(&lbuf, FXchar, reqbuf+1);
      bufsz = reqbuf;
      }
    cbuf = getText(pos1, lbuf, reqbuf);
    lbuf[reqbuf] = 0;
    if (do_rx && lbuf[cbuf-1] == '\n') lbuf[cbuf-1] = 0; // Don't include newlines for rx.
    if (!fd->case_sens)
      for (FXint i = 0; i < cbuf; i++)
        lbuf[i] = toupper(lbuf[i]);
    
    while (!maybe && delta < ilen && delta >= 0) {
    if (do_rx) {
#ifdef POSIX_REGEX
      rc = regexec(re, lbuf+delta, 1, pmatch, delta ? REG_NOTBOL : 0);
      if (!rc) {
        sp = pos1+delta+pmatch[0].rm_so;
        len = pmatch[0].rm_eo - pmatch[0].rm_so;
#else
      rc = fxregexec(re, lbuf+delta);
      if (rc) {
        sp = pos1+(re->startp[0]-lbuf);
        len = re->endp[0] - re->startp[0];
#endif
	maybe = TRUE;
        }
      }
    else {
      FXint i;
      if (fd->fwd) {
        for (i = delta; i < ilen && i+sslen <= cbuf; i++)
          if (!memcmp(lbuf+i, ss, sslen)) {
            sp = pos1+i;
            len = sslen;
            maybe = TRUE;
            break;
            }
        }
      else {
        delta = FXMIN(delta, cbuf-sslen);
        for (i = delta; i >= 0; i--)
          if (!memcmp(lbuf+i, ss, sslen)) {
            sp = pos1+i;
            len = sslen;
            maybe = TRUE;
            break;
            }
        }
      }
    if (maybe) {
      FXchar c;
      if (do_rx && first && !fd->fwd && sp > pos) {
        // Backward search and first line, but found rx at or after pos.  Discard.
        maybe = FALSE;
        break;
        }
      if (fd->whole_word &&
          (!(sp==pos1 || isspace(c = getChar(sp-1)) || delimiters.has(c)) || 
           !(sp+len>=pos1+ilen-1 || isspace(c = getChar(sp+len)) || delimiters.has(c)))) {
        // Keep searching, wasn't required whole word.
        if (do_rx || fd->fwd)
          delta = sp-pos1+1;
        else
          delta = sp-pos1-1;
        maybe = FALSE;
        }
      }
    else
      break;	// Didn't find 
    } //while
    
    if (maybe)
      break; // Found it. 
    
nextline:
    first = FALSE;
    if (fd->fwd) {
      line += 1;
      if (line >= lines()) {
        if (wrap) {
          wrap = FALSE;
          line = 0;
          }
        else {
          sp = -1;
          break;
          }
        }
      }
    else {
      line -= 1;
      if (line >= lines()) {  // Really less than 0, but unsigned
        if (wrap) {
          wrap = FALSE;
          line = lines()-1;
          }
        else {
          sp = -1;
          break;
          }
        }
      }
    sp = pos1 = getPosOfLine(line);
    if (fd->fwd) {
      if (!wrap && pos1 >= epos) {
        sp = -1;
        break;
        }
      }
    ilen = getPosOfLine(line+1) - pos1; // Length of 1st line 
    delta = do_rx || fd->fwd ? 0 : ilen-1;
    if (!fd->fwd) {
      if (!wrap && pos1+ilen-1 <= epos) {
        sp = -1;
        break;
        }
      }
    } // for(;;)
  
  if (fd->fwd) {
    if (!wrap && sp >= epos)
      sp = -1;
    }
  else {
    if (!wrap && sp < epos)
      sp = -1;
    }

  // Clean up regex structures if any.
  if (do_rx) {
#ifdef POSIX_REGEX
    if (re = &rei)
      regfree(re);
#else
    fxregfree(re);
#endif
    }
  FXFREE(&lbuf);
  
bye:  
  if (ss && ss != fd->find)
    FXFREE(&ss);
  return sp;
  }

FXbool FXTextSrc::isModified(FXuint which /*= 1*/) const {
  switch (which) {
    case 0: return modified_since_create;
    case 1: return modified_since_write;
    default: return modified_since_autosave;
    }
  return FALSE; // Never executed
  }

void FXTextSrc::setModified(FXbool mod, FXuint which /*= 1*/) {
  switch (which) {
    case 0: modified_since_create = mod; break;
    case 1: modified_since_write = mod; break;
    default: modified_since_autosave = mod; break; 
    }
  }

void FXTextSrc::enumerateMarkers(FXPageFormat * pf, FXint startpos, FXint endpos) const {
  FXint i, j;
  //printf("FXTextSrc::enumerateMarkers() %08X sp=%d ep=%d\n", pf, startpos, endpos);
  for (i = 0; i < anchs.no(); i++)
    if (!pf) 
      for (j = 0; j < pfs.no(); j++)
        anchs[i]->tellAttrs(pfs[j], pfs[j]->getFirstPos(), pfs[j]->getLastPos()+1);
    else
      anchs[i]->tellAttrs(pf, startpos, endpos);
  }

void FXTextSrc::markRemoved(FXTextAnchor * ta, FXPageFormat * pf, FXbool temp,
                            FXint sp, FXint ep) {
  FXint i, j;
  if (ignoreMarkers) return;
  // Removal is complicated by the fact that overlapping brackets which use the same
  // attribute bits may exist.  Any brackets overlapping ta have an update call so that
  // the pager can set the bits back on that were turned off by removal of this bracket.
  // First we need to determine how many brackets overlap.
  if (sp < 0) sp = ta->getPos();
  if (ep < 0) ep = ta->getEpos();
  FXint overlap = 0;
  FXTextAnchor * other;
  if (ep > sp) {
    for (i = 0; i < anchs.no(); i++) {
      if ((other = anchs[i]) == ta)
        continue;
      if (!(other->getAttrs() & ta->getAttrs()))
        continue;
      if (other->getPos() >= ep ||
          other->getEpos() <= sp)
        continue;
      if (!other->isVisible(pf ? pf->getPLText() : NULL))
        continue;
      overlap++;
      }
    }
  
  // Turn off attrib bits for this object.
  if (!pf) 
    for (j = 0; j < pfs.no(); j++)
      pfs[j]->markUpdated(ta, sp, ep, FALSE, temp || overlap);
  else
    pf->markUpdated(ta, sp, ep, FALSE, temp || overlap);
  
  // Turn overlapping ones back on.
  if (overlap) {
    FXint maxsp, minep;
    for (i = 0; i < anchs.no(); i++) {
      if ((other = anchs[i]) == ta)
        continue;
      if (!(other->getAttrs() & ta->getAttrs()))
        continue;
      if (other->getPos() >= ep ||
          other->getEpos() <= sp)
        continue;
      if (!other->isVisible(pf ? pf->getPLText() : NULL))
        continue;
      overlap--;
      maxsp = FXMAX(sp, other->getPos());
      minep = FXMIN(ep, other->getEpos());
      if (!pf) 
        for (j = 0; j < pfs.no(); j++)
          pfs[j]->markUpdated(other, maxsp, minep, TRUE, temp || overlap);
      else
        pf->markUpdated(ta, maxsp, minep, TRUE, temp || overlap); 
      }
    }
  
  }
  
void FXTextSrc::markPlaced(FXTextAnchor * ta, FXPageFormat * pf, FXbool temp,
                           FXint sp, FXint ep) {
  FXint j;
  if (ignoreMarkers) return;
  if (sp < 0) sp = ta->getPos();
  if (ep < 0) ep = ta->getEpos();
  if (!pf) 
    for (j = 0; j < pfs.no(); j++)
      pfs[j]->markUpdated(ta, sp, ep, TRUE, temp);
  else
    pf->markUpdated(ta, sp, ep, TRUE, temp); 
  enforceMutex(ta);
  } 

void FXTextSrc::enforceMutex(FXTextAnchor * ta) {
  if (ta->isVisible() && ta->getMutex()) {
    for (FXint i = 0; i < anchs.no(); i++) {
      if (anchs[i] == ta)
        continue;
      if (anchs[i]->getMutex() & ta->getMutex())
        anchs[i]->setVisible(FALSE); 
      }
    }
  }

/*----------- FXTextSrc selection handling ----------------*/

void FXTextSrc::createSelection(FXint sp, FXint ep) {
  //printf("create %d %d\n",sp,ep);
  selection = new FXTextSelection(this, FXPFATTR_SELECTED, NULL, sp, ep);
  markPlaced(selection, NULL, FALSE);
  
  // Now remember the selection until next time.
  if (last_sel) {
    last_sel_len = ep - sp;
    if (last_sel_len > FXTEXTSRC_MAX_LAST_SEL_LEN)
      last_sel_len = FXTEXTSRC_MAX_LAST_SEL_LEN;
    last_sel_len = getText(sp, last_sel, last_sel_len);
    }
  }

void FXTextSrc::killSelection() {
  //printf("kill\n");
  if (selection) {
    selection->setVisible(FALSE); 
    delete selection;
    }
  }

void FXTextSrc::setSelection(FXint sp, FXint ep) {
  //printf("set %d %d\n",sp,ep);
  if (!selection) { if (sp < ep && sp >= 0) createSelection(sp, ep); } 
  else if (sp >= ep || sp < 0 || ep < 0)
    killSelection();
  else if (sp != selection->getPos() || ep != selection->getEpos()) {
    selection->move(sp, ep, FALSE);
  
    // Now remember the selection until next time.
    if (last_sel) {
      last_sel_len = ep - sp;
      if (last_sel_len > FXTEXTSRC_MAX_LAST_SEL_LEN)
        last_sel_len = FXTEXTSRC_MAX_LAST_SEL_LEN;
      last_sel_len = getText(sp, last_sel, last_sel_len);
      }
    }
  }

FXint FXTextSrc::getSelectionStart() const {
  return isSelection() ? selection->getPos() : -1;
  }

FXint FXTextSrc::getSelectionEnd() const {
  return isSelection() ? selection->getEpos() : -1;
  }

FXint FXTextSrc::getSelection(FXint & sp, FXint & ep) const {
  if (isSelection()) {
    ep = selection->getEpos();
    sp = selection->getPos();
    return 1;
    }
  ep = -1;
  sp = -1;
  return 0;
  }

FXbool FXTextSrc::isSelection() const {
  return selection && selection->getPos() < selection->getEpos() && selection->getPos() >= 0;
  }

void FXTextSrc::resetOutput() {}

/*---------------------------------------------------------*/

void FXTextSrc::save(FXStream& /*store*/) const {
  }

void FXTextSrc::load(FXStream& /*store*/) {
  }

FXTextSrc::~FXTextSrc() {
  // Delete anchors (but not pagers, since they are responsible for deleting this).
  while (anchs.no())
    removeTextAnchor(anchs[0]);
  selection = (FXTextSelection *)-1;
  FXFREE(&last_sel);
  }

 

/*-------------------- Text Source Buffered in Memory -------------------------*/


inline void constructElms(FXMemTextSrc::Transaction*t, unsigned int num) {
  for (unsigned int i = 0; i < num; i++)
    t[i].before=NULL; 
  }
  
inline void destructElms(FXMemTextSrc::Transaction*, unsigned int) { 
  } 

// Object implementation
FXIMPLEMENT(FXMemTextSrc,FXTextSrc,NULL,0)

FXMemTextSrc::FXMemTextSrc() : FXTextSrc() {
  rolling = FALSE;
  init = TRUE;
  logent = 0;
  textbuf = NULL;
  map = NULL;
  }

// Create an 'empty' source.
FXMemTextSrc::FXMemTextSrc(FXApp * a, FXuint opts) :
    FXTextSrc(a, opts) {
  rolling = FALSE;
  logent = 0;
  textbuf = new FXTextBufS((FXString)"");
  map = new FXTextMap(1);
  map->updateLeaf(0, 1, 1);	// Must have 1 for virtual trailing newline.  This newline cannot
  				// be deleted (enforced by replace()).
  init = FALSE;
  if (!textbuf || !textbuf->isOk() || !map || !map->isOk()) {
    if (textbuf) delete textbuf;
    if (map) delete map;
    ok = FALSE;			// Set error condition if necessary.
    }
  }

// Create and initialise to given string (text) of given length.
FXMemTextSrc::FXMemTextSrc(FXApp * a, FXuint opts, const FXchar * text, FXint textlen) :
    FXTextSrc(a, opts) {
  logent = 0;
  textbuf = new FXTextBufS((FXString)text);
  map = new FXTextMap(1);
  map->updateLeaf(0, 1, 1);
  init = TRUE;
  rolling = TRUE;
  replace(0, 0, text, textlen);
  rolling = FALSE;
  init = FALSE;
  if (!textbuf || !textbuf->isOk() || !map || !map->isOk()) {
    if (map) delete map;
    if (textbuf) delete textbuf;
    ok = FALSE;			// Set error condition if necessary.
    }
  }

const FXString & FXMemTextSrc::getLineText(FXuint lineno, FXint & pos) const {
  static FXString s;
  pos = getPosOfLine(lineno);
  FXuint len = textbuf->find(pos);
  FXuint clen;
  FXchar * ptr;
  s.size(0);
  FXint p = pos;
  while (len && (clen = textbuf->getContig(p, ptr))) {
    s.append(ptr, FXMIN(len, clen));
    if (clen >= len)
      break;
    len -= clen;
    p += clen;  
    }
  return s;
  }
  
FXint FXMemTextSrc::getText(FXint pos, FXchar * ptr, FXint len) const {
  FXint c, i;
  FXchar * p;
  FXint tl = (FXint)map->root_Pn();
  if (pos < 0 || pos >= tl) { 
    if (len) ptr[0] = 0;
    return 0;
    }
  if (pos + len > tl)
    c = tl - pos;
  else
    c = len;
  
  for (i=0; i < c; i+=tl) {
    tl = textbuf->getContig(pos+i, p); 
    tl = FXMIN(tl, c-i);
    memcpy(ptr+i, p, tl);
    }
  if (i < len)
    ptr[i] = 0;
  return i;
  }

FXint FXMemTextSrc::getBuf(FXint & pos, const FXchar * & ptr) const {
  FXchar * p;
  FXint len = textbuf->getContig(pos, p); // pos not updated
  ptr = p;
  pos += len;
  return len;
  }

FXint FXMemTextSrc::getBackwardsBuf(FXint & pos, const FXchar * & ptr) const {
  FXchar * p;
  FXuint pp = (FXuint)pos;
  FXint len = textbuf->getBackwardsContig(pp, p); // pp _is_ updated
  pos = (FXint)pp;
  ptr = p;
  return len;
  }

FXint FXMemTextSrc::getPosOfLine(FXuint lineno) const {
  if (lineno >= lines())
    return length();
  FXTextNode search(0, lineno);
  map->modulo_Ln(search);
  FXuint pos = search.Pn;	// Pos of first line of node
  FXuint ln = search.Ln;	// Line number within this node
  while (ln) {
    pos += textbuf->find(pos);
    ln--;
    }
  return (FXint)pos;
  }

// Public
FXuint FXMemTextSrc::getLineOfPos(FXint pos) const {
  if (pos < 0)
    return 0;
  if (pos >= length())
    return lines()-1; 
  FXTextNode search((FXuint)pos, 0);
  FXuint node = map->modulo_Pn(search);
  FXuint spos = (FXuint)pos - search.Pn; // Position of 1st line of this node
  FXuint ln = search.Ln;	// Line number of 1st line in this node
  FXuint len;
  FXuint lin = map->node_Ln(node);
  // We achieve a dramatic optimisation by only searching if more than one line in this node
  while (lin > 1)
    if ((len = textbuf->find(spos)) + spos <= (FXuint)pos && len) {
      spos += len;
      ln++;
      lin--;
      }
    else
      break;
  return ln;
  }

// Protected.  Same as above, except returns more information.
FXuint FXMemTextSrc::getLineOfPos(FXint pos, FXuint & node, FXuint & Ln, FXuint & Pn) const {
  Ln = Pn = node = 0;
  if (pos < 0)
    pos = 0;
  if (pos > length()-1)
    pos = length()-1;
  FXTextNode search((FXuint)pos, 0);
  node = map->modulo_Pn(search);
  Ln = search.Ln;
  Pn = search.Pn;
  FXuint spos = (FXuint)pos - search.Pn; // Position of 1st line of this node
  FXuint ln = search.Ln;	// Line number of 1st line in this node
  FXuint len;
  while ((len = textbuf->find(spos)) + spos <= (FXuint)pos && len) {
    spos += len;
    ln++;
    }
  return ln;
  }

FXuint FXMemTextSrc::lines() const {
  // Even in an empty text source, there must be a virtual newline, hence there is always
  // at least one line.  This is consistent with the fact that the cursor must be able to
  // move somewhere even in an empty widget.
  return map->root_Ln();
  }

FXint FXMemTextSrc::length() const {
  FXASSERT(map->root_Pn() == textbuf->length());
#ifndef NDEBUG
  if (map->root_Pn() != textbuf->length())
    printf("  Pn=%u  len=%u\n", map->root_Pn(), textbuf->length());
#endif
  return map->root_Pn();
  }

FXchar FXMemTextSrc::getChar(FXint pos) const {
  return textbuf->get(pos);
  }

FXint FXMemTextSrc::replace(FXint startpos, FXint endpos, const FXString & text, FXbool defer /*= FALSE*/, const FXPLText * pltext) {
  return replace(startpos, endpos, text.text(), text.length(), defer, pltext);
  }

FXint FXMemTextSrc::replace(FXint startpos, FXint endpos, const FXchar * text, FXint n, FXbool defer /*= FALSE*/, const FXPLText * pltext) {
  /* endpos is one past the last position to be replaced.  endpos==startpos for pure
     insertion immediately before startpos.  The trailing virtual newline cannot be
     replaced, thus startpos and endpos are adjusted back if they would overlap it.
   */
#ifndef NDEBUG
  FXuint ilines = lines();
#endif
  FXint ilength = length();
  if (startpos >= ilength)
    startpos = ilength-1;
  if (endpos >= ilength)
    endpos = ilength-1;
  FXint i; 
  FXuint u;
  FXuint len = (FXuint)(endpos-startpos); // Length of old data to be replaced
  FXint delta = n - (FXint)len;		// Nett insertion (-ve for deletion)
  FXTextReplacement tr;
  
  // Ignore if nothing to be done
  if (defer && !len && !n) return 0;

  tr.startpos = startpos;
  tr.endpos = endpos;
  tr.text = text;
  tr.n = n;
  tr.defer = defer;
  tr.from_widget = pltext;
  
  // Query enclosing bracket for permission
  FXTextAnchor * ta = getEnclosingBracket(startpos, endpos);
  if (ta && !ta->permitsReplacement(&tr))
    return 4;	// indicate permission denied by bracket
  
  // node1 is containing node for start position, node 2 that for end.  For positions at
  // end (i.e. appending), the containing node is actually the last non-zero one.  We have
  // to be aware of this.
  // Ln1 is first line number in node1, Pn1 is length of stretch before replacement start (B).
  FXuint node1, Ln1, Pn1, node2, Ln2, Pn2;
  FXuint line1 = getLineOfPos(startpos, node1, Ln1, Pn1);
  FXuint line2 = getLineOfPos(endpos, node2, Ln2, Pn2);
  FXuint olines = line2 - line1;	// Number of end-of-lines replaced
  FXuint onodes = (node2 - node1 + 2)>>1;	// Nodes to be replaced

  FXuint nlines = 0;
  for (i = 0; i < n; i++)
    if (text[i] == '\n') nlines++;	// Number of new end-of-lines
  FXuint posB = startpos - Pn1;	// Position of start of section B.
  FXuint posA = endpos - Pn2 + map->node_Pn(node2); // Pos of end of section A (+1).

  if (!init) {
    // Inform page formatters that we are about to make a change.  This gives them the
    // opportunity to save any necessary state that depends on the current buffer
    // contents.
    for (i = 0; i < pfs.no(); i++)
      pfs[i]->preChange(line1, line2, (FXint)(nlines-olines), 
                        startpos, endpos, delta);
    }
  
//#ifndef NDEBUG
#if (0)
  printf("replace: startpos=%d endpos=%d n=%d\n"
         "  repllen=%u delta=%d\n"
         "  line1=%u node1=%u Ln1=%u Pn1=%u\n"
         "  line2=%u node2=%u Ln2=%u Pn2=%u\n"
         "  olines=%u onodes=%u nlines=%u\n"
         "  posB=%u posA=%u\n"
         "  lines()=%u  length()=%d\n"
         , startpos, endpos, n
         , len, delta
         , line1, node1, Ln1, Pn1
         , line2, node2, Ln2, Pn2
         , olines, onodes, nlines
         , posB, posA
         , ilines, ilength
         );
#endif

  // Ask anchors how they want to be positioned, and remove them permanently or temporarily
  // as desired.  Removal of anchors must be done in 'old' buffer co-ordinates.  Ones which
  // are temporarily removed are turned back on after the buffer replacement.
  if (!init) for (i = 0; i < anchs.no(); i++) {
    FXTextAnchor * ta = anchs[i];
    FXint news, newe;
    if (ta->query(&tr, news, newe)) {
      // It requested deletion
      markRemoved(ta, NULL, FALSE);
      removeTextAnchor(ta);
      }
    }

  // Log the change if not currently rolling.  Note that a sequence of single-character
  // insertions (except for control characters) at increasing positions is condensed into
  // one transaction.  A sequence of single character deletions at the same or decreasing
  // positions is also condensed.
  if (!rolling) {
    Transaction * pt;
    FXbool icontig = FALSE;	// Contiguous insertion after previous
    FXbool dcontig = FALSE;	// Contiguous deletion at same position as previous
    FXbool bcontig = FALSE;	// Contiguous backspace deletion
    if ((len == 1 && n == 0 || len == 0 && n == 1) &&
        (pt = logent ? &log[logent-1] : NULL)) {	// Most recent transaction
      icontig = len == 0 && n == 1 && startpos == pt->pos + pt->nlen && text[0] >= ' ' &&
                pt->olen == 0;
      dcontig = len == 1 && n == 0 && startpos == pt->pos && pt->nlen == 0;
      bcontig = len == 1 && n == 0 && startpos == pt->pos - 1 && pt->nlen == 0;
      }
    if (icontig) {
      FXRESIZE(&pt->before, FXchar, pt->nlen+1);
      if (!pt->before)
        ok = FALSE;
      else {
        pt->after = pt->before;
        pt->after[pt->nlen++] = text[0];
        }
      }
    else if (dcontig) {
      FXRESIZE(&pt->before, FXchar, pt->olen+1);
      if (!pt->before)
        ok = FALSE;
      else {
        getText(startpos, pt->before + pt->olen++, 1);
        pt->after = pt->before + pt->olen;
        }
      }
    else if (bcontig) {
      FXRESIZE(&pt->before, FXchar, pt->olen+1);
      if (!pt->before)
        ok = FALSE;
      else {
        memmove(pt->before+1, pt->before, pt->olen);
        getText(startpos, pt->before, 1);
        pt->after = pt->before + ++pt->olen;
        pt->pos--;
        }
      }
    else {
      Transaction * t = truncateLog(++logent);
      t->pos = startpos;
      t->olen = len;
      t->nlen = n;
      t->notlast = defer;
      FXFREE(&t->before);
      FXMALLOC(&t->before, FXchar, t->olen+t->nlen);
      if (t->before) {
        t->after = t->before + t->olen;
        if (t->nlen)
          memcpy(t->after, text, t->nlen);
        if (t->olen)
          getText(startpos, t->before, t->olen);
        }
      else {
        logent--;
        ok = FALSE;
        }
      }
    if (!ok) return 2;	// Indicate error in transaction record.
    }
  
  // Replace in the buffer...
  textbuf->replace((FXuint)startpos, len, text, (FXuint)n);
  ok = textbuf->isOk();
  if (!ok)
    return 1;	// Indicate error in buffer replacement
  posA = (FXint)posA + delta;	// Adjust end position.
  // Now count how many new nodes will be required for stretch B-R-A [posB..posA]
  FXuint p = posB;	// Last position where node starts
  FXuint nn = 1;	// # nodes
  for (u = posB + textbuf->find(posB); u < posA; u += textbuf->find(u))
    if (u > p + 80) p = u, nn++;
  
//#ifndef NDEBUG
#if (0)
  printf("  nn=%u\n", nn);
#endif
  
  // Insert or delete nodes in map
  if (nn > onodes) {
    ok = map->insertLeaves(node1, nn-onodes, 0);
    if (!ok)
      return 3;		// error in line mapping update.
    }
  else if (nn < onodes)
    map->deleteLeaves(node1, onodes-nn);
  // Update nodes
  nn = node1;
  p = posB;
  FXuint ll = 1;
  for (u = posB + textbuf->find(posB); u < posA; u += textbuf->find(u))
      if (u > p + 80) {
//#ifndef NDEBUG
#if (0)
        printf("  upd leaf %u  Pn=%u Ln=%u\n", nn, u-p, ll);
#endif
        map->updateLeaf(nn, u-p, ll);
        p = u;
        nn+=2;
        ll = 1;
        }
      else
        ll++; 

  if (u-p) {
    // Update the leaf only if there is non-zero data.  If not, leave the leaf zero.
//#ifndef NDEBUG
#if (0)
    printf("  upd leaf %u  Pn=%u Ln=%u\n", nn, u-p, ll);
#endif
    map->updateLeaf(nn, u-p, ll);
  }
    
  FXASSERT(ilines-olines+nlines == lines());
  FXASSERT(ilength+delta == (FXint)length());
  
  if (!init) {
    // Ask anchors how they want to be positioned, and make them display again.  We know
    // that none will ask for deletion, since we asked above and any that wanted to be deleted
    // will have been removed already.  ignoreMarkers is set so that the call to ta->move()
    // does not cause inappropriate changes in the pager's attribute bits (via markRemoved()
    // or markPlaced(), which are called by move()).  The change() call refreshes the attribute
    // bits properly, so doing it again would only cause unnecessary redrawing.  In any case,
    // the placement would be wrong because the row matrix would not have been updated yet. 
    ignoreMarkers = TRUE;
    for (i = 0; i < anchs.no(); i++) {
      FXTextAnchor * ta = anchs[i];
      FXint news, newe; 
      ta->query(&tr, news, newe);
      ta->move(news, newe, TRUE);	// Deferred graphics, since change() below does it all at once.
      }
    ignoreMarkers = FALSE;
    modified_since_create = TRUE;
    modified_since_write = TRUE;
    modified_since_autosave = TRUE;
    
    // Inform page formatters of the change.  We pass the first and last line numbers (in the
    // original) which were affected, the difference in total number of lines, the position
    // range of the change [startpos,endpos) and nett number of chars inserted.
    for (i = 0; i < pfs.no(); i++)
      pfs[i]->change(line1, line2, (FXint)(nlines-olines), 
                     startpos, endpos, delta, defer);
    }
  return 0; 
  }

FXMemTextSrc::Transaction * FXMemTextSrc::truncateLog(FXint num) {
  FXint i;
  for (i = num; i < log.no(); i++)
    FXFREE(&log[i].before);
  log.size(num);
  logent = num;
  if (num)
    return &log[num-1];
  return NULL;
  }

FXint FXMemTextSrc::resetLog() {
  FXint ret = log.no();
  truncateLog(0);
  return ret;
  }

FXint FXMemTextSrc::rollback(FXint num) {
  FXint i;
  Transaction * t;
  FXbool more;
  FXASSERT(num >= 0);
  rolling = TRUE;
  for (i = 0; i < num && logent; i++) {
    do {
      t = &log[--logent];
      more = logent > 0 && log[logent-1].notlast;
      replace(t->pos, t->pos + t->nlen, t->before, t->olen, more);
      } while (more);
    }
  rolling = FALSE;
  if (!logent)
    modified_since_write = FALSE;
  return i;
  }

FXint FXMemTextSrc::rollforward(FXint num) {
  FXint i;
  Transaction * t;
  FXbool more;
  FXASSERT(num >= 0);
  rolling = TRUE;
  for (i = 0; i < num && logent < log.no(); i++) {
    do {
      t = &log[logent++];
      more = t->notlast;
      replace(t->pos, t->pos + t->olen, t->after, t->nlen, more); 
      } while (more);
    }
  rolling = FALSE;
  return i;
  }

FXint FXMemTextSrc::rollbackPos() const {
  if (!logent)
    return -1;
  return log[logent-1].pos;
  }

FXint FXMemTextSrc::rollforwardPos() const {
  if (logent >= log.no())
    return -1;
  return log[logent].pos;
  }


FXint FXMemTextSrc::write() {
  modified_since_write = FALSE;
  modified_since_autosave = FALSE;
  truncateLog(0);
  return -1; // Operation not supported
  }

FXint FXMemTextSrc::autosave() {
  modified_since_autosave = FALSE;
  return -1; // Operation not supported
  }

FXint FXMemTextSrc::revert() {
  rollback(2000000000);
  return 0;
  }

FXbool FXMemTextSrc::isModifiable() const {
  return TRUE;
  }



void FXMemTextSrc::save(FXStream& /*store*/) const {
  }

void FXMemTextSrc::load(FXStream& /*store*/) {
  }

FXMemTextSrc::~FXMemTextSrc() {
  if (textbuf)
    delete textbuf;
  textbuf = (FXTextBuf *)-1;
  if (map)
    delete map;
  map = (FXTextMap *)-1;
  }

 

/*-------------------- Text Source Buffered in File System -------------------------*/
FXDEFMAP(FXFileTextSrc) FXFileTextSrcMap[]={
   FXMAPFUNC(SEL_SIGNAL,FXFileTextSrc::ID_SIGCHLD,FXFileTextSrc::onSigSIGCHLD)
  };
  
// Object implementation
FXIMPLEMENT(FXFileTextSrc,FXMemTextSrc,FXFileTextSrcMap,ARRAYNUMBER(FXFileTextSrcMap))



FXFileTextSrc::FXFileTextSrc() : 
  FXMemTextSrc(), mode(AUTO), lock(FALSE), anon(TRUE), as_hold(FALSE) {
  autosavefile = makeAutosaveFileName(file);
  logfile = makeLogFileName(file);
  file = "untitled-" +  FXStringVal(anon_num);
  is_shell = FALSE;
  }


FXFileTextSrc::FXFileTextSrc(FXApp * a, FXuint opts) : 
  FXMemTextSrc(a, opts), mode(AUTO), lock(FALSE), anon(TRUE), as_hold(FALSE) {
  autosavefile = makeAutosaveFileName(file);
  logfile = makeLogFileName(file);
  file = "untitled-" +  FXStringVal(anon_num);
  is_shell = FALSE;
  }
  

FXFileTextSrc::FXFileTextSrc(FXApp * a, FXuint opts, const FXString & name,
  		             le_mode m /*= AUTO*/, FXbool lck /*= TRUE*/ ) :
  FXMemTextSrc(a, opts), file(name), mode(m), lock(lck), as_hold(FALSE) {
  getProtocol(name);
  anon = !file.text()[0];
  autosavefile = makeAutosaveFileName(file);
  logfile = makeLogFileName(file);
  if (anon)
    file = "untitled-" +  FXStringVal(anon_num);
#ifndef WIN32
  if (is_shell) {
    file = protocol + resource;
    openShell(NULL, false, resource.text());
    }
#endif  
  }
  
void FXFileTextSrc::getProtocol(const FXString & name) {
#ifndef WIN32
  // Look for URL in form xxx://
  const FXchar * rsrc = name.text();		// Resource
  const FXchar * url = rsrc;
  FXint len;
  is_shell = FALSE;
  while (isalpha(*rsrc)) rsrc++;
  if (*rsrc == ':' && rsrc[1] == '/' && rsrc[2] == '/') {
    // Looks like URL.  The only ones we understand are shell (==tty) or pipe.
    // tty is used for terminal emulation, pipe is non-interactive stream.
    len = rsrc - url;
    use_pipe = !strncmp(url, "pipe", len);
    is_shell = !strncmp(url, "shell", len) || !strncmp(url, "tty", len) || use_pipe;
    }
  if (is_shell) {
    protocol = name.left(len);
    rsrc += 3;
    if (!*rsrc) {
      // No command specified.  Get $SHELL env. var and use that, or failing that,
      // just use "sh".
      rsrc = getenv("SHELL");
      if (!rsrc)
        rsrc = "sh";
      }
    //printf("using shell %s\n", rsrc);
    resource = rsrc;
    }
  else {
    protocol = "";
    resource = name;
    }
#else
  protocol = "";
  resource = name;
  is_shell = FALSE;
#endif
  }
  
#ifndef WIN32
  #define PLATFORM_MODE LF
#else
  #define PLATFORM_MODE CRLF
#endif

FXint FXFileTextSrc::uniq = 0;

FXString FXFileTextSrc::makeAutosaveFileName(const FXString & name) {
  if (anon) {
    anon_num = ++uniq;
#ifdef WIN32
    return name + "_" + FXStringVal(anon_num) + ".jic";
#else
    return name + "_" + FXStringVal(getpid()) + "_" + FXStringVal(anon_num) + ".jic";
#endif
    }
  //return "/tmp/" + FXFile::title(name) + ".jic";
  return name + ".jic";	// FIXME: make configurable
  }

FXString FXFileTextSrc::makeLogFileName(const FXString & name) {
  if (anon) {
#ifdef WIN32
    return name + "_" + FXStringVal(anon_num) + ".log";
#else
    return name + "_" + FXStringVal(getpid()) + "_" + FXStringVal(anon_num) + ".log";
#endif
    }
  return name + ".log";
  }


FXint FXFileTextSrc::writeTo(const FXString & fname) {
  FILE *fp;
  FXint n = 0;
  FXint w;
  FXint len;
  FXint tlen = length()-1;	// Length without virtual newline
  FXint pos = 0;
  const FXchar * ptr;
  
  if (mode == BINARY)
    return 9998;
  
  if (mode == AUTO)
    mode = PLATFORM_MODE;	// Default to platform native line end convention
    
  
  if (mode == CRLF_ATEOF || mode == LF_ATEOF) {
    // Want to ensure newline is written at end, so add the virtual newline back in if
    // the last real char is not a newline.
    if (!tlen || getChar(tlen-1) != '\n')
      tlen++;
    }
  
  fp = fopen(fname.text(), "wb");
  if(!fp) return errno;
  
  if (mode == CRLF || mode == CRLF_ATEOF)
    while (n < tlen && (len = getBuf(pos, ptr))) {
      if (n+len > tlen) len = tlen-n;
      FXint i;
      for (i = 0; i < len; i+=w) {
        FXint j;
        for (j = i; j < len && ptr[j] != '\n'; j++);
        w = fwrite(ptr+i, 1, j-i, fp);
        if (w < j-i) break;
        if (j < len) {
          fwrite("\r\n", 1, 2, fp); // Yes, the old teletypewriter lives on under the guise of DOS.
          w++;	// Add one for the \n in the text source
          }
        n += w;
        }
      if (i < len)
        break;
      }
  else
    while (n < tlen && (len = getBuf(pos, ptr))) {
      if (n+len > tlen) len = tlen-n;
      w = fwrite(ptr, 1, len, fp);
      if (w < len)
        break;
      n += w;
      }
  fclose(fp);
  if (n != tlen) {
    //printf("n=%d tlen=%d\n", n, tlen);
    return 9999;
    }
  
  return 0;
  }

FXint FXFileTextSrc::write() {
  //printf("Writing to `%s', autosave is `%s'\n", file.text(), autosavefile.text());
  FXint rc = writeTo(file);
  if (rc)
    return rc;
  modified_since_write = FALSE;
  modified_since_autosave = FALSE;
  // On successful write, we delete the autosave file if it exists
  if (autosavefile.text()[0] &&
      FXFile::exists(autosavefile))
    remove(autosavefile.text());
    
  return 0;
  }
  

FXint FXFileTextSrc::autosave() {
  if (as_hold)
    return 0;
  //printf("Autosave to %s\n", autosavefile.text());
  if (autosavefile.text()[0] && !is_shell) {
    FXint rc = writeTo(autosavefile);
    if (rc)
      return rc;
    modified_since_autosave = FALSE;
    return 0;
    }
  return -1;	// No filename set.
  }
  

FXint FXFileTextSrc::revert() {
  return -1;
  }
  

FXbool FXFileTextSrc::isModifiable() const {
  return mode != BINARY; 
  }

FXbool FXFileTextSrc::isModified(FXuint which /*= 1*/) const {
  if (!is_shell) return FXTextSrc::isModified(which);
  return FALSE; // Shell looks like never modified, so that "save" button is
                // not enabled.  User would have to use "save as".
  }

FXint FXFileTextSrc::maybeSave(FXWindow * widget, FXbool prompt) {
  // Possibly save modified text.  Return -1 to cancel operation in progress, or 0 to continue,
  // or +ve for error from save.
  if (prompt && !is_shell && modified_since_write) {
    FXuint answer;
    answer = FXMessageBox::question(widget, MBOX_YES_NO_CANCEL,
        "Unsaved Document", "Save %s to file?", file.text());
    if (answer == MBOX_CLICKED_CANCEL) return -1;
    if (answer == MBOX_CLICKED_YES) {
      if (anon) {
        FXFileDialog savedialog(widget, "Save Document");
        savedialog.setFilename(file);
        if (!savedialog.execute()) return -1;
        if (fxexists(savedialog.getFilename().text())) {
          if (MBOX_CLICKED_NO == 
              FXMessageBox::question(widget, MBOX_YES_NO,
                "Overwrite Document", "Overwrite existing document?")) return -1;
          }
        file = savedialog.getFilename(); 
        }
      return save(widget, prompt); 
      }
    }
  // We delete the autosave file if it exists, because the user didn't want to save the
  // current file (pressed NO), or not prompting.
  if (autosavefile.text()[0] &&
      FXFile::exists(autosavefile))
    remove(autosavefile.text());
  
  return 0;
  }

FXint FXFileTextSrc::open(FXWindow * widget, FXbool prompt, const FXString & name,
  	                  le_mode m /*= AUTO*/, FXbool lck /*= TRUE*/ ) {
  FXbool do_dlg = !(name.text()[0]);
  FXString nfile;	// File name to open (tentatively)
  
  as_hold = TRUE;	// Temporarily inhibit autosave during dialogue
  FXint notsaved = maybeSave(widget, prompt);
  as_hold = FALSE;
  if (notsaved) return notsaved;

  if (do_dlg) {
    FXFileDialog open(widget, "Open Document"); 
    open.setFilename(file);
    open.setPatternList(patterns);
    if (open.execute())
      nfile = open.getFilename();
    else return -1;
    }
  else
    nfile = name;
  
  // Delete all anchors other than cursor

#ifndef WIN32
  // If shell, send HUP signal
  if (is_shell && proc) {
    proc->disconnect();
    //FIXME: this is unsatisfactory since the user has to retry the open dialogue.
    return -1;	// Cannot continue until child dies (wait for SIGCHLD).
    }
#endif
  getProtocol(nfile);

  if (!nfile.text()[0])
    return neo(widget, prompt);

  FXbool oldanon = anon;
  anon = FALSE;
  // temporarily turn off anon, since we are determining tentative names. 
  FXString autosave = makeAutosaveFileName(nfile);
  FXString log = makeLogFileName(nfile);
  anon = oldanon;
  FXbool useAutosave = FALSE;

#ifndef WIN32
  if (is_shell) {
    openShell(widget, prompt, resource.text());
    anon = FALSE;
    file = protocol + resource;
    mode = m;
    lock = lck;
    autosavefile = autosave;
    logfile = log;
    
    return 0;
    }
#endif

  int err = 0;
  struct stat info;
  FILE *fp = NULL;
  char *text = NULL;
  FXuint n;
  if (stat(nfile.text(), &info)) {
    err = errno;
    if (err == ENOENT) {
      // For file not found, we set empty contents but retain the file name in case the
      // user wants to save as that (new) file.  fp is set to null to signify this,
      // unless the autosave file is available.
      if (FXFile::exists(autosave)) {
        FXuint answer;
        answer = FXMessageBox::question(widget, MBOX_YES_NO_CANCEL,
          "Existing autosave file found for new file", "Use the autosave file contents?");
        if (answer == MBOX_CLICKED_CANCEL) return -1;
        useAutosave = (answer == MBOX_CLICKED_YES);
        err = 0;	// Forget the original file-not-found error.
        if (useAutosave && stat(autosave.text(), &info))
          err = errno;
        if (!err) {
          fp = fopen(useAutosave ? autosave.text() : nfile.text(), "rb");
          if (!fp)
            err = errno;
          }
        }
      }
    }
  else {
    // File exists.  If autosave file name, check if exists and modification time is
    // more recent than file.  If so, then prompt user for recovery of possible crash
    // file.
    if (FXFile::exists(autosave)) {
      if (FXFile::modified(autosave) > FXFile::modified(nfile)) {
        FXuint answer;
        answer = FXMessageBox::question(widget, MBOX_YES_NO_CANCEL,
          "Autosave file found", "Use the autosave file contents?");
        if (answer == MBOX_CLICKED_CANCEL) return -1;
        useAutosave = (answer == MBOX_CLICKED_YES);
        if (useAutosave && stat(autosave.text(), &info))
          err = errno;
        }
      }
    fp = fopen(useAutosave ? autosave.text() : nfile.text(), "rb");
    if (!fp)
      err = errno;
    }
  
  n = 0;	// Chars read.
  
  if (err) {
handle_error:
    const char * emsg;
    if (text)
      FXFREE(&text);
    if (err > 0 && err < sys_nerr)
      emsg = sys_errlist[err];
    else if (err == 9997)
      emsg = "File too large";
    else if (!err)
      emsg = "Internal error";
    else
      emsg = "Unknown cause";
    FXMessageBox::error(widget, MBOX_OK,
                  "Error opening document",
                  "Opening `%s'\nError number %d\n%d byte%s read\n%s", 
                  nfile.text(), 
                  err,
                  n, n==1?"":"s",
                  emsg);
    if (err == ENOENT) {
      // Set empty (but writable) text
      rolling = TRUE;
      replace(0, length(), "", 0);
      truncateLog(0);
      rolling = FALSE;
      modified_since_create = FALSE;
      // If we used a more recent autosave (crash) file, set mod-since-write to remind user
      // to save the file.
      modified_since_write = useAutosave;
      modified_since_autosave = FALSE;
  
      // Commit open to this object.
      anon = FALSE;
      mode = m;
      lock = lck;
      file = nfile;
      autosavefile = autosave;
      logfile = log;
      }
    return err;
    }
  
  
  
  widget->getApp()->beginWaitCursor();
  FXMALLOC(&text, FXchar, info.st_size);	// Initial buffer
  if (!text) {
    err = 9997;
    goto handle_error;
    }

  if (m == AUTO) {
    /* In auto mode, the heuristic for determining the line-ending conventions is:
         Read up to 16k, or entire file -> ISIZE.
         Count \n chars -> NL
         Count number of \r\n pairs -> CRLF
         Count number of control chars excluding \t, \f, \n and \r -> CTL
         If ISIZE==0, assume platform-dependent mode.
         If NL==0, if isize <= 256 assume platform-dependent else assume BINARY
         Compute DP=CRLF/NL, CC=CTL/ISIZE
         If CC > 0.02, assume BINARY
         If DP < 0.5, assume LF
         assume CRLF
    */
    FXint isize = FXMIN(16384, info.st_size);
    n = fread(text, 1, isize, fp);
    if ((FXint)n < isize) {
      err = errno;
      goto handle_error;
      }
    FXint nl = 0;
    FXint crlf = 0;
    FXint ctl = 0;
    for (FXint i = 0; i < isize; i++)
      if (text[i] == '\n') {
        nl++;
        if (i && text[i-1] == '\r')
          crlf++;
        }
      else if (text[i] < ' ' || text[i] == 0x7F) {
        if (text[i] != '\t' && text[i] != '\r' && text[i] != '\f')
          ctl++;
        }
    if (!isize)
      m = PLATFORM_MODE;
    else if (!nl)
      m = isize <= 256 ? PLATFORM_MODE : BINARY;
    else if ((float)ctl / isize > 0.02)
      m = BINARY;
    else if ((float)crlf/nl < 0.5)
      m = LF;
    else
      m = CRLF;
    }

  // Read everything
  n += fread(text+n, 1, info.st_size-n, fp);
  if (errno) {
    err = errno;
    goto handle_error;
    }
  
  // Convert line endings
  if (m == LF || m == LF_ATEOF) {
    // No problem for unix style.
    }
  else if (m == CRLF || m == CRLF_ATEOF) {
    // DOS style.  CRLF pairs are converted to LF.  If there were 'bare' LFs in the original
    // document, these will be converted to CRLFs when written out.  Thus the file might
    // grow slightly, even if not edited.
    FXchar * p, * q;
    FXint i;
    for (p = text, q = text, i = 1; i < (FXint)n; i++, p++)
      if (p[0] != '\r' || p[1] != '\n')
        *q++ = *p;
    *q++ = *p;
    n = q - text;
    }
  else {
    // No problem with binary unless want to insert newlines...
#ifdef INSERT_NEWLINES
    // Binary.  Insert LF if none has occurred for 64 chars.  This is irreversible, so
    // files with binary mode cannot be written out.  The buffer must be increased by 65/64
    // to account for the worst-case insertion of LFs.
    FXint newsize = info.st_size + (info.st_size>>6) + 63;
    FXRESIZE(&text, FXchar, newsize);
    if (!text) {
      err = 9997;
      goto handle_error;
      }
    memmove(text+(newsize-n), text, n);
    FXchar * p, * q;
    FXint cc;
    FXint i;
    for (p = text+(newsize-n), q = text, i = 0, cc = 0; i < (FXint)n; i++, p++, q++) {
      if (*p == '\n')
        cc = 0;
      else if (cc == 64) {
        cc = 0;
        *q++ = '\n';
        }
      else
        cc++;
      *q = *p ? *p : '.';
      }
    n = q - text;
#endif
    }

  
  rolling = TRUE;
  replace(0, length(), text, n);
  truncateLog(0);
  rolling = FALSE;
  modified_since_create = FALSE;
  // If we used a more recent autosave (crash) file, set mod-since-write to remind user
  // to save the file.
  modified_since_write = useAutosave;
  modified_since_autosave = FALSE;
  FXFREE(&text);
  fclose(fp);
  widget->getApp()->endWaitCursor();
  
  // Commit open to this object.
  anon = FALSE;
  mode = m;
  lock = lck;
  file = nfile;
  autosavefile = autosave;
  logfile = log;
    
  return 0;
  }

FXint FXFileTextSrc::open(FXWindow * widget, FXbool prompt) {
  return open(widget, prompt, file, mode, lock);
  }


FXint FXFileTextSrc::neo(FXWindow * widget, FXbool prompt) {
  as_hold = TRUE;	// Temporarily inhibit autosave during dialogue
  FXint notsaved = maybeSave(widget, prompt);
  as_hold = FALSE;
  if (notsaved) return notsaved;

  rolling = TRUE;
  replace(0, length()-1, "", 0); 
  truncateLog(0);
  rolling = FALSE;
  modified_since_create = FALSE;
  modified_since_write = FALSE;
  modified_since_autosave = FALSE;
  anon = TRUE;
  mode = AUTO;
  autosavefile = makeAutosaveFileName(file);
  logfile = makeLogFileName(file);
  file = "untitled-" + FXStringVal(anon_num);
  return 0;
  }
  

FXint FXFileTextSrc::save(FXWindow * widget, FXbool prompt) {
  return saveAs(widget, prompt, file, mode, lock);
  }
  

FXint FXFileTextSrc::saveAs(FXWindow * widget, FXbool prompt, const FXString & name,
  	                    le_mode m /*= AUTO*/, FXbool lck /*= TRUE*/) {
  FXString oldfile = file;
  le_mode oldmode = mode;
  FXint err;
   
  if (anon || !name.text()[0]) {
    FXFileDialog savedialog(widget, "Save Document");
    savedialog.setFilename(file);
    if (savedialog.execute()) {
      if (fxexists(savedialog.getFilename().text())) {
        if (MBOX_CLICKED_NO ==
            FXMessageBox::question(widget, MBOX_YES_NO,
              "Overwrite Document", "Overwrite existing document?")) return -1;
        }
      file = savedialog.getFilename();
      }
    else
      return -1;
    }
  else
    file = name;
  
  // Set line-ending mode (defaults to current mode).
  if (m != AUTO)
    mode = m;
    
  err = write();
  if (err) {
    showError(widget, err, "Error saving document", file.text());
    file = oldfile;
    mode = oldmode;
    return err;
    }
  anon = FALSE;
  logfile = makeLogFileName(file); 
  autosavefile = makeAutosaveFileName(file);
  lock = lck;
  return 0;
  }
  
void FXFileTextSrc::showError(FXWindow * widget, FXint err, 
                    const FXchar * heading, const FXchar * context) {
  if (err) {
    const char * emsg;
    if (err > 0 && err < sys_nerr)
      emsg = sys_errlist[err];
    else if (err == 9999)
      emsg = "Could not complete writing file";
    else if (err == 9998)
      emsg = "Cannot write binary file";
    else if (err == -1)
      emsg = "Function not supported";
    else
      emsg = "Unknown cause";
    FXMessageBox::error(widget, MBOX_OK,
                  heading,
                  "%s\nError number %d\n%s", context, err, emsg);
    }
  }

void FXFileTextSrc::setMode(le_mode m) {
  mode = m;
  }
  

FXbool FXFileTextSrc::setLock(FXbool lck /*= TRUE*/) {
  lock = lck;
  return TRUE;
  }
  

void FXFileTextSrc::save(FXStream& store) const {
  }
  

void FXFileTextSrc::load(FXStream& store) {
  }
  

FXFileTextSrc::~FXFileTextSrc() {
#ifndef WIN32
  if (is_shell && proc) {
     FXProcess::removeTarget(this);
     if (output) {
       FXProcess::removeTarget(output);
       delete output;
       }
     if (output2) {
       FXProcess::removeTarget(output2);
       delete output2;
       }
     if (input) {
       FXProcess::removeTarget(input);
       delete input;
       }
     proc->disconnect();
     delete proc;
     }
#endif
  }
  

#ifndef WIN32
  /* Stuff that I admit I don't know how to do for Windows...  Help me somebody!
  */
  
  // Handle Unix warts.  This code stolen from xterm, xxgdb and other random sources.
  // God this is ugly!

// Whether to use separate socket for stderr.  This garbles the display since
// one does not then know the order in which the subprocess wrote things to
// stdout and stderr.  Probably best not to define this.
//#define SEP_STDERR

  
// Open Unix shell process.  Turns the text source into a command-line tool.
// widget and prompt as for open().  rsrc is a shell:// resource (after the //).  We
// ignore any part after a '.' (since this is a dummy extension so the profiling works - FIXME).
// The part before the extension is the name of an executable.  /bin/sh is used as a default.
FXint FXFileTextSrc::openShell(FXWindow * widget, FXbool prompt, const FXchar * rsrc) {
  char * argv[2];
   
  proc = new FXProcess(getApp(), this, ID_SIGCHLD, (use_pipe ? PROCESS_PIPE : 0)
#ifdef SEP_STDERR
     | PROCESS_STDERR
#endif  
  );

  output = new FXOutputBracket(this, 0);
  output->setAttrs(output->getAttrs() | FXTextAnchor::ANCHOR_REPEL_CURSOR);
  output->setVisible();
  proc->setTargetFor(0, output, MKUINT(FXTextAnchor::ID_SOCKET, SEL_IO_READ));
#ifdef SEP_STDERR
  output2 = new FXOutputBracket(this, FXPFATTR_ERRORS);
  output2->setAttrs(output2->getAttrs() | FXTextAnchor::ANCHOR_REPEL_CURSOR);
  output2->setVisible();
  proc->setTargetFor(1, output2, MKUINT(FXTextAnchor::ID_SOCKET, SEL_IO_READ));
#else
  output2 = NULL;
#endif
  input = new FXInputBracket(this, 0, FXPFATTR_RETENTIVE, NULL, 0, 1);
  input->setVisible();
  proc->setTargetFor(0, input, MKUINT(FXTextAnchor::ID_SOCKET, SEL_IO_WRITE));
  proc->hold(0, INPUT_WRITE);
  argv[0] = (char *)rsrc;
  argv[1] = NULL;
  proc->connect(FXString(rsrc), 1, argv);
  return 0;
  }
#endif  // !WIN32

void FXFileTextSrc::resetOutput() {
  if (!isShell()) return;
  FXint len = length() - 1;
  if (output && len > output->getEpos()) {
    if (output->getPos() < output->getEpos()) output->clone();
    output->move(len, len, FALSE);
    }
  if (output2 && len > output2->getEpos()) {
    if (output2->getPos() < output2->getEpos()) output2->clone();
    output2->move(len, len, FALSE);
    }
  }

long FXFileTextSrc::onSigSIGCHLD(FXObject*,FXSelector,void*ptr) {
  int status = (int)ptr; // because comes via FXProcess
  char buf[128];
  //printf("Child succumbed\n");
  delete input;
  input = NULL;
  if (WIFEXITED(status))
    sprintf(buf, "\nShell terminated code %d\n", WEXITSTATUS(status));
  else
    sprintf(buf, "\nShell killed: signal %d\n", WTERMSIG(status));
  replace(length()-1, length()-1, buf, strlen(buf));
  return 1;
  }



